"""
An interface between Python and q.
"""

import os

# Attempt to import PyArrow early to get ahead of others (e.g. Pandas) who would try to import it
# without guarding against segfaults. Skip this if we're just doing a qinit check.
if os.environ.get('PYKX_QINIT_CHECK') is None:
    try:
        from ._pyarrow import pyarrow
    except ImportError: # nocov
        pass

# Importing core initializes q if in licensed mode, and loads the q C API symbols. This should
# happen early on so that if the qinit check is currently happening then no time is wasted.
# The current process will exit while the core module is loading if the qinit check is happening.
from .core import licensed

from abc import ABCMeta, abstractmethod
import itertools as it
from pathlib import Path
import platform
import shutil
from typing import List, Optional, Union
from warnings import warn

from . import util

if platform.system() == 'Windows': # nocov
    os.environ['PATH'] += f';{util.pykx_platlib_dir}'
    if platform.python_version_tuple()[:2] >= ('3', '8'):
        os.add_dll_directory(util.pykx_platlib_dir)


def _first_resolved_path(possible_paths: List[Union[str, Path]]) -> Path:
    """Returns the resolved version of the first path that exists."""
    for path in possible_paths:
        try:
            return Path(path).resolve(strict=True)
        except FileNotFoundError:
            pass
    unfound_paths = '\n'.join(str(path) for path in possible_paths)
    raise FileNotFoundError(f'Could not find any of the following files:\n{unfound_paths}')


class Q(metaclass=ABCMeta):
    reserved_words = {
        'abs', 'acos', 'asin', 'atan', 'avg', 'bin', 'binr', 'cor', 'cos', 'cov', 'delete', 'dev',
        'div', 'do', 'enlist', 'exec', 'exit', 'exp', 'getenv', 'hopen', 'if', 'in', 'insert',
        'last', 'like', 'log', 'max', 'min', 'prd', 'select', 'setenv', 'sin', 'sqrt', 'ss', 'sum',
        'tan', 'update', 'var', 'wavg', 'while', 'within', 'wsum', 'xexp',
    }

    def __init__(self):
        self.paths = default_paths
        # Start with an empty set to avoid errors during initialization.
        object.__setattr__(self, '_q_ctx_keys', set())
        # HACK: `'_connection_info' in self.__dict__` is used as a proxy for a type-check to avoid a
        # cyclic import.
        # NOTE: 'and ... not in' is being used instead of 'or ... in' temporarily to disable the
        # context interface over IPC. See the comment at the end of `ipc.pyx` for details.
        # if licensed or '_connection_info' in self.__dict__:
        if licensed and '_connection_info' not in self.__dict__:
            object.__setattr__(self, 'ctx', QContext(self, '', None))
            object.__setattr__(self, '_q_ctx_keys', {
                *self.ctx.q._context_keys,
                *self.reserved_words,
            })
        object.__setattr__(self, 'console', QConsole(self))
        object.__setattr__(self, 'sql', None) # placeholder
        object.__setattr__(self, 'qsql', QSQL(self))
        object.__setattr__(self, 'query', None) # placeholder
        object.__setattr__(self, 'read', QReader(self))
        object.__setattr__(self, 'write', QWriter(self))

    @abstractmethod
    def __call__(self, query: str, *args, sync: Optional[bool] = None):
        pass # nocov

    def __getattr__(self, key):
        # Elevate the q context to the global context, as is done in q normally
        if key in self._q_ctx_keys: # if-statement used instead of try-block for performance
            return self.ctx.q.__getattr__(key)
        try:
            return self.ctx.__getattr__(key)
        except AttributeError as attribute_error:
            try:
                self._register(name=key)
                self.ctx._invalidate_cache()
                return self.ctx.__getattr__(key)
            except Exception as inner_error:
                raise attribute_error from inner_error

    # __setattr__ takes precedence over data descriptors, so we implement a custom one that uses
    # the default behaviour for select keys
    def __setattr__(self, key, value):
        if key in self.__dict__ or key in Q.__dict__:
            object.__setattr__(self, key, value)
        else:
            if key in self.reserved_words or key in self.q:
                raise exceptions.PyKXException(
                    'Cannot assign to reserved word or overwrite q namespace.'
                )
            self('set', f'.{key}', value, sync=True)

    def __delattr__(self, key):
        self.ctx.__delattr__(key)

    def __getitem__(self, key):
        return self('get', key, sync=True)

    def __setitem__(self, key, value):
        # TODO: remove this first if-statement after the context interface works over IPC again.
        if '_connection_info' not in self.__dict__:
            if key in self.reserved_words or key in self.q:
                raise exceptions.PyKXException(
                    'Cannot assign to reserved word or overwrite q namespace.'
                )
        self('set', key, value, sync=True)

    def __delitem__(self, key):
        key = str(key)
        if key.startswith('.'):
            raise exceptions.PyKXException('Cannot delete from the global context.')
        elif key not in self('key `.', sync=True).py():
            raise KeyError(key)
        self('{![`.;();0b;enlist x]}', key, sync=True) # delete key from `.

    def __dir__(self):
        if hasattr(self, 'ctx'): # Temporary workaround to disable the context interface over IPC.
            return sorted({
                *super().__dir__(),
                *dir(self.ctx),
                *dir(self.ctx.__getattr__('q')),
            })
        else:
            return super().__dir__()

    def _register(self,
                  name: Optional[str] = None,
                  path: Optional[Union[Path, str]] = None,
    ) -> str:
        """Obtain the definitions from a q/k script.

        Parameters:
            name: Name of the context to be loaded. If `path` is not provided, [a file whose name
                matches will be searched for](#script-search-logic), and loaded if found. If `path`
                is provided, `name` will be used as the name of the context that the script at
                `path` is executed in.
            path: Path to the script to load. If `name` is not provided, it will default to the
                filename sans extension.

        Returns:
            The attribute name for the newly loaded module.
        """
        if name is None and path is None:
            raise ValueError('Module name or path must be provided.')
        if path is not None:
            path = Path(path).resolve(strict=True)
        else:
            path = _first_resolved_path([''.join(x) for x in it.product(
                # `str(Path/@)[:-1]` gets the path with a trailing path separator
                (str(x/'@')[:-1] for x in self.paths),
                ('.', ''),
                (name,),
                ('.q', '.k'),
                ('', '_')
            )])
        if name is None: # defaults to filename at end of path sans extension
            name = path.stem
        prev_ctx = self('string system"d"', sync=True)
        try:
            self(
                f'{"" if name[0] == "." else "."}{name}:(enlist`)!enlist(::);'
                f'system "d {"" if name[0] == "." else "."}{name}";'
                f'system "l {path.as_posix()}"',
                sync=True,
            )
            return name[1:] if name[0] == '.' else name
        finally:
            self('{system"d ",x}', prev_ctx, sync=True)

    @property
    def paths(self):
        """List of locations for the context interface to find q scripts in.

        Defaults to the current working directory and `$QHOME`.
        """
        return object.__getattribute__(self, '_paths')

    @paths.setter
    def paths(self, paths: List[Union[str, Path]]):
        resolved = []
        for path in paths:
            if not isinstance(path, Path):
                path = Path(path)
            try:
                resolved.append(path.resolve(strict=True))
            except FileNotFoundError:
                warn(f"Module path '{path!r}' not found", RuntimeWarning)
        object.__setattr__(self, '_paths', resolved)

    @property
    def max_num_threads(self):
        return int(self('system"s 0N"', sync=True))

    @max_num_threads.setter
    def max_num_threads(self, value):
        raise AttributeError("Can't set attribute - the maximum number of threads available to "
                             "q is fixed on startup by the '-s' flag.")

    @property
    def num_threads(self):
        return int(self('system"s"', sync=True))

    @num_threads.setter
    def num_threads(self, value):
        self(f'system"s {int(value)}"', sync=True)


# Import order matters here, so the imports are not ordered conventionally.
from .console import QConsole
from .ctx import default_paths, QContext
from .query import QSQL
from .read import QReader
from .write import QWriter

from . import console
from . import exceptions
from . import wrappers
from .embedded_q import EmbeddedQ, q
from ._version import version as __version__
from .exceptions import *
from .ipc import QConnection
from .util import QARGS, QHOME, QLIC
from .wrappers import *


__all__ = sorted([
    'EmbeddedQ',
    'Q',
    'QARGS',
    'QConnection',
    'QConsole',
    'QContext',
    'QHOME',
    'QReader',
    'QWriter',
    'QLIC',
    'console',
    'ctx',
    'exceptions',
    'install_into_QHOME',
    'licensed',
    'toq',
    'util',
    'q',
    *exceptions.__all__,
    *wrappers.__all__,
])


def __dir__():
    return __all__


def install_into_QHOME():
    """Copies `pykx.q` into `$QHOME`."""
    shutil.copy(Path(__file__).parent/'pykx.q', QHOME)
