import os
from pathlib import Path
from platform import system

from .exceptions import PyKXException, PyKXWarning


def _num_available_cores() -> int:
    if hasattr(os, 'sched_getaffinity'):
        return len(os.sched_getaffinity(0))
    elif system() == 'Windows': # nocov
        import ctypes # nocov
        import ctypes.wintypes # nocov
        kernel32 = ctypes.WinDLL('kernel32') # nocov
        DWORD_PTR = ctypes.wintypes.WPARAM # nocov
        PDWORD_PTR = ctypes.POINTER(DWORD_PTR) # nocov
        GetCurrentProcess = kernel32.GetCurrentProcess # nocov
        GetCurrentProcess.restype = ctypes.wintypes.HANDLE # nocov
        GetProcessAffinityMask = kernel32.GetProcessAffinityMask # nocov
        GetProcessAffinityMask.argtypes = (ctypes.wintypes.HANDLE, PDWORD_PTR, PDWORD_PTR) # nocov
        mask = DWORD_PTR() # nocov
        if not GetProcessAffinityMask(GetCurrentProcess(), ctypes.byref(mask), ctypes.byref(DWORD_PTR())): # nocov
            raise PyKXException("Call to 'GetProcessAffinityMask' failed") # nocov
        return bin(mask.value).count('1') # nocov
    # No way to determine the number of cores available in this case. The CPU count will work in
    # most scenarios, and for those in which it does not there is no known workaround.
    return os.cpu_count() # nocov


cdef I _qinit(I (*qinit)(I, C**, C*, C*, C*), qhome_str: str, qlic_str: str, args: List[str]) except *:
    num_cores_as_bytes = str(_num_available_cores()).encode()
    default_args = (b'q', b'-q', b'-s', num_cores_as_bytes)
    # LIST comprehension on user-provided args checks for errors before we start allocating memory
    normalized_args = (*default_args, *[x.encode() for x in args])
    cdef int argc = len(normalized_args)
    cdef char** argv = <char**>PyMem_Malloc(sizeof(char*) * argc)
    for i, arg in enumerate(normalized_args):
        argv[i] = strncpy(<char*>PyMem_Malloc(len(arg) + 1), arg + b'\0', len(arg) + 1)
    qhome_bytes = bytes(Path(__file__).parent.absolute()/'lib')
    qlic_bytes = qlic_str.encode()
    init_code = qinit(argc, argv, qhome_bytes, qlic_bytes, NULL)
    # Calling qinit sets QHOME, so we set it back to the original value here
    os.environ['QHOME'] = qhome_str
    return init_code


cdef char* _libq_path
cdef void* _q_handle


# The `PYKX_QINIT_CHECK` env var indicates that the current process is a Python subprocess made to
# check if qinit can be initialized without error. The error might be something relatively harmless
# like a license error, but could also cause a segfault. The subprocess allows us to safely attempt
# calling `_qinit`, and return the result for the parent to decide if it should use licensed mode,
# or fallback to unlicensed mode.
# Data is passed in via the `PYKX_QINIT_CHECK` env var to reduce PyKX's import time - why do the
# work to gather that data when the parent process has already done it?
qinit_check_data = os.environ.get('PYKX_QINIT_CHECK')
if qinit_check_data is not None: # nocov
    _core_q_lib_path, _qhome_str, _qlic_str, _qargs = qinit_check_data.split(';') # nocov
    import shlex # nocov
    _qargs = shlex.split(_qargs) # nocov
    _libq_path_py = _core_q_lib_path.encode() # nocov
    _libq_path = _libq_path_py # nocov
    _q_handle = dlopen(_libq_path, RTLD_NOW | RTLD_GLOBAL) # nocov
    qinit = <I (*)(I, C**, C*, C*, C*)>dlsym(_q_handle, 'qinit') # nocov
    os._exit(_qinit(qinit, _qhome_str, _qlic_str, _qargs)) # nocov


from libc.string cimport strncpy
from cpython.mem cimport PyMem_Malloc
from libc.stdint cimport *

from typing import List, Union
from warnings import warn
import subprocess
import sys

from .util import find_core_lib, once, normalize_to_bytes, QARGS, QHOME, QLIC


if '--licensed' in QARGS and '--unlicensed' in QARGS:
    raise PyKXException("$QARGS includes mutually exclusive flags '--licensed' and '--unlicensed'")


under_q = 'PYKX_UNDER_Q' in os.environ


cdef inline K r1k(x):
    return r1(<K><uintptr_t>x._addr)


cdef inline uintptr_t _keval(const char* code, K k1, K k2, K k3, K k4, K k5, K k6, K k7, K k8):
    # TODO: Provide the option to release the GIL for long-running queries. Such a change should be
    # accompanied with the addition of a lock around `k` to avoid corruption of q's memory.
    # Having a long-running `k` call running alongside some unrelated Python/C code is valid, but
    # having multiple `k`/`dot`/`r0`/etc. calls running simultaneously cannot be done safely.
    return <uintptr_t>k(0, <char* const>code, k1, k2, k3, k4, k5, k6, k7, k8, NULL)


def keval(code: bytes, k1=None, k2=None, k3=None, k4=None, k5=None, k6=None, k7=None, k8=None):
    # This code is ugly, but Cython can turn it into a switch statement.
    if k1 is None:
        return _keval(code, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL)
    elif k2 is None:
        return _keval(code, r1k(k1), NULL, NULL, NULL, NULL, NULL, NULL, NULL)
    elif k3 is None:
        return _keval(code, r1k(k1), r1k(k2), NULL, NULL, NULL, NULL, NULL, NULL)
    elif k4 is None:
        return _keval(code, r1k(k1), r1k(k2), r1k(k3), NULL, NULL, NULL, NULL, NULL)
    elif k5 is None:
        return _keval(code, r1k(k1), r1k(k2), r1k(k3), r1k(k4), NULL, NULL, NULL, NULL)
    elif k6 is None:
        return _keval(code, r1k(k1), r1k(k2), r1k(k3), r1k(k4), r1k(k5), NULL, NULL, NULL)
    elif k7 is None:
        return _keval(code, r1k(k1), r1k(k2), r1k(k3), r1k(k4), r1k(k5), r1k(k6), NULL, NULL)
    elif k8 is None:
        return _keval(code, r1k(k1), r1k(k2), r1k(k3), r1k(k4), r1k(k5), r1k(k6), r1k(k7), NULL)
    else:
        return _keval(code, r1k(k1), r1k(k2), r1k(k3), r1k(k4), r1k(k5), r1k(k6), r1k(k7), r1k(k8))


if under_q: # nocov
    if '--unlicensed' in QARGS: # nocov
        warn("The '--unlicensed' flag has no effect when running under a q process", # nocov
             PyKXWarning) # nocov
    _q_handle = dlopen(NULL, RTLD_NOW | RTLD_GLOBAL) # nocov
    licensed = True # nocov
else:
    # To make Cython happy, we indirectly assign Python values to `_libq_path`
    if '--unlicensed' in QARGS:
        _libq_path_py = bytes(find_core_lib('e'))
        _libq_path = _libq_path_py
        _q_handle = dlopen(_libq_path, RTLD_NOW | RTLD_GLOBAL)
        licensed = False
    else:
        _core_q_lib_path = find_core_lib('q')
        _qinit_check_proc = subprocess.run(
            (str(Path(sys.executable).as_posix()), '-c', 'import pykx'),
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            text=True,
            env={
                **os.environ,
                'PYKX_QINIT_CHECK': ';'.join((
                    str(_core_q_lib_path),
                    str(QHOME),
                    str(QLIC),
                    # Use the env var directly because `util.QARGS` has already split the args.
                    os.environ.get('QARGS', ''),
                )),
            }
        )
        _qinit_output = '    ' + '    '.join(_qinit_check_proc.stdout.strip().splitlines(True))
        if _qinit_check_proc.returncode: # Fallback to unlicensed mode
            if _qinit_output != '    ':
                _capout_msg = f' Captured output from initialization attempt:\n{_qinit_output}'
            else:
                _capout_msg = ''
            if '--licensed' in QARGS:
                raise PyKXException(f'Failed to initialize embedded q.{_capout_msg}')
            else:
                warn('Failed to initialize embedded q; falling back to unlicensed mode, which '
                     'has limited functionality. Refer to https://code.kx.com/pykx/modes.html '
                     f'for more information.{_capout_msg}',
                     PyKXWarning)
            _libq_path_py = bytes(find_core_lib('e'))
            _libq_path = _libq_path_py
            _q_handle = dlopen(_libq_path, RTLD_NOW | RTLD_GLOBAL)
            licensed = False
        else: # Start in licensed mode
            if _qinit_output != '    ':
                warn(f'Unexpected output from initialzation of embedded q:\n{_qinit_output}',
                     PyKXWarning)
            _libq_path_py = bytes(_core_q_lib_path)
            _libq_path = _libq_path_py
            _q_handle = dlopen(_libq_path, RTLD_NOW | RTLD_GLOBAL)
            qinit = <I (*)(I, C**, C*, C*, C*)>dlsym(_q_handle, 'qinit')
            qinit_return_code = _qinit(qinit, str(QHOME), str(QLIC), QARGS)
            if qinit_return_code: # nocov
                dlclose(_q_handle) # nocov
                licensed = False # nocov
                raise PyKXException( # nocov
                    f'Non-zero qinit return code {qinit_return_code} despite successful pre-check') # nocov
            licensed = True

kG = <G* (*)(K x)>dlsym(_q_handle, 'kG')
kC = <G* (*)(K x)>dlsym(_q_handle, 'kC')
kU = <U* (*)(K x)>dlsym(_q_handle, 'kU')
kS = <S* (*)(K x)>dlsym(_q_handle, 'kS')
kH = <H* (*)(K x)>dlsym(_q_handle, 'kH')
kI = <I* (*)(K x)>dlsym(_q_handle, 'kI')
kJ = <J* (*)(K x)>dlsym(_q_handle, 'kJ')
kE = <E* (*)(K x)>dlsym(_q_handle, 'kE')
kF = <F* (*)(K x)>dlsym(_q_handle, 'kF')
kK = <K* (*)(K x)>dlsym(_q_handle, 'kK')

b9 = <K (*)(I mode, K x)>dlsym(_q_handle, 'b9')
d9 = <K (*)(K x)>dlsym(_q_handle, 'd9')
dj = <I (*)(I date)>dlsym(_q_handle, 'dj')
dl = <K (*)(V* f, J n)>dlsym(_q_handle, 'dl')
dot = <K (*)(K x, K y)>dlsym(_q_handle, 'dot')
ee = <K (*)(K x)>dlsym(_q_handle, 'ee')
ja = <K (*)(K* x, V*)>dlsym(_q_handle, 'ja')
jk = <K (*)(K* x, K y)>dlsym(_q_handle, 'jk')
js = <K (*)(K* x, S s)>dlsym(_q_handle, 'js')
jv = <K (*)(K* x, K y)>dlsym(_q_handle, 'jv')
k = <K (*)(I handle, const S s, ...)>dlsym(_q_handle, 'k')
ka = <K (*)(I t)>dlsym(_q_handle, 'ka')
kb = <K (*)(I x)>dlsym(_q_handle, 'kb')
kc = <K (*)(I x)>dlsym(_q_handle, 'kc')
kd = <K (*)(I x)>dlsym(_q_handle, 'kd')
ke = <K (*)(F x)>dlsym(_q_handle, 'ke')
kf = <K (*)(F x)>dlsym(_q_handle, 'kf')
kg = <K (*)(I x)>dlsym(_q_handle, 'kg')
kh = <K (*)(I x)>dlsym(_q_handle, 'kh')
ki = <K (*)(I x)>dlsym(_q_handle, 'ki')
kj = <K (*)(J x)>dlsym(_q_handle, 'kj')
knk = <K (*)(I n, ...)>dlsym(_q_handle, 'knk')
knt = <K (*)(J n, K x)>dlsym(_q_handle, 'knt')
kp = <K (*)(S x)>dlsym(_q_handle, 'kp')
kpn = <K (*)(S x, J n)>dlsym(_q_handle, 'kpn')
krr = <K (*)(const S s)>dlsym(_q_handle, 'krr')
ks = <K (*)(S x)>dlsym(_q_handle, 'ks')
kt = <K (*)(I x)>dlsym(_q_handle, 'kt')
ktd = <K (*)(K x)>dlsym(_q_handle, 'ktd')
ktj = <K (*)(H _type, J x)>dlsym(_q_handle, 'ktj')
ktn = <K (*)(I _type, J length)>dlsym(_q_handle, 'ktn')
ku = <K (*)(U x)>dlsym(_q_handle, 'ku')
kz = <K (*)(F x)>dlsym(_q_handle, 'kz')
m9 = <V (*)()>dlsym(_q_handle, 'm9')
okx = <I (*)(K x)>dlsym(_q_handle, 'okx')
orr = <K (*)(const S)>dlsym(_q_handle, 'orr')
r0 = <V (*)(K k)>dlsym(_q_handle, 'r0')
r1 = <K (*)(K k)>dlsym(_q_handle, 'r1')
sd0 = <V (*)(I d)>dlsym(_q_handle, 'sd0')
sd0x = <V (*)(I d, I f)>dlsym(_q_handle, 'sd0x')
sd1 = <K (*)(I d, f)>dlsym(_q_handle, 'sd1')
sd1 = <K (*)(I d, f)>dlsym(_q_handle, 'sd1')
sn = <S (*)(S s, J n)>dlsym(_q_handle, 'sn')
ss = <S (*)(S s)>dlsym(_q_handle, 'ss')
sslInfo = <K (*)(K x)>dlsym(_q_handle, 'sslInfo')
vak = <K (*)(I x, const S s, va_list l)>dlsym(_q_handle, 'vak')
vaknk = <K (*)(I, va_list l)>dlsym(_q_handle, 'vaknk')
ver = <I (*)()>dlsym(_q_handle, 'ver')
xD = <K (*)(K x, K y)>dlsym(_q_handle, 'xD')
xT = <K (*)(K x)>dlsym(_q_handle, 'xT')
ymd = <I (*)(I year, I month, I day)>dlsym(_q_handle, 'ymd')
