"""Interface to q contexts and scripts which define a namespace.

The context interface provides an easy to way access q contexts (also known as namespaces when at
the top level). For more information about contexts/namespaces in q please refer to
[Chapter 12 of Q for Mortals](https://code.kx.com/q4m3/12_Workspace_Organization/).

Warning: The context interface has been temporarily disabled for `pykx.QConnection` instances.
    Due to the discovery of issues with the use of the context interface over IPC, it has been
    temporarily disabled to prevent misleading behaviour and unexpected errors from occuring. It
    will be re-enabled in a future version of PyKX.

Both the local q instance at `pykx.q`, and `pykx.Connection` instances, have attributes for q
namespaces, which are exposed in Python as `pykx.QContext` objects. These context objects have
attributes for their members, which can either be sub-contexts or K objects. For example:

- `pykx.q.Q` is a `KdbContext` instance for the builtin `.Q` context/namespace
- `pykx.q.ctxA.ctxB` is a `KdbContext` instance for the `.ctxA.ctxB` context
- `pykx.q.ctxA.ctxB.kObject` is a [`pykx.K`][pykx.K] instance for the `.ctxA.ctxB.kObject` K object

Just as in q, the `.q` context is accessible at the top-level, so for instance instead of accessing
`pykx.q.q.md5`, you can access it as `pykx.q.md5`. Some q builtins cannot be accessed like this
such as `or` and `not` as these result in Python syntax errors. Such functions can instead be
accessed either with `getattr`, or by evaluating them as q code (e.g. `pykx.q('not')`).

Accessing attributes of `pykx.q` (or a `pykx.ipc.Connection` instance) which do not correspond to
a context that has been loaded in memory in q will result in it
[trying to find a script with a matching name](#script-search-logic). This process is detailed in
the flowchart below:

```mermaid
graph LR
  A([`pykx.q.ctxA`<br>is accessed.]);
  B{Is `.ctxA` <br>defined in<br>memory in q?};
  C{Is `.ctxA` <br>defined in<br>memory in q?};
  D["Search for a<br>matching script<br>(details below)."];
  E{Has a<br>matching script<br>been found?};
  F([`AttributeError`<br>is raised.]);
  G([`QContext`<br>is returned.]);
  H["Switch to the<br>context `.ctxA`<br>(`system &quot;d .ctxA&quot;`),
    <br>execute the script,<br>then switch back."];

  A --> B;
  B --No--> D;
  B --Yes--> G;
  C --No--> F;
  C --Yes--> G;
  E --No--> F;
  E --Yes--> H;
  D --> E;
  H --> C;
```

The fact that the context might not be defined even after the context interface changes the context
and executes the script might be confusing. This can happen because the script can switch into
other contexts, which overrides the context switch done by the context interface. Additionally the
script might use fully qualified names for its definitions, which can bypass the effect of
switching contexts.

Note that context switches persists across `pykx.q` calls (but not `pykx.QConnection(...)` calls).
One should take care when switching contexts, as unexpectedly being in an different context can
result in undesirable behaviour. `QContext` objects are Python context managers, which means they
can be used with the `with` statement like so:

```python
# q code here executes in the global context
with q.myctx:
    # q code here executes in the `.myctx` context
    pass
# q code here executes in the global context
```

If you would like to switch into a q context using a string for the context name, use `getattr`
like so:

```python
# q code here executes in the global context
with getattr(q, 'myctx'):
    # q code here executes in the `.myctx` context
    pass
# q code here executes in the global context
```

## Script Search Logic

When the context interface cannot find a namespace (i.e. a top-level context) that is being
accessed it attempts to find a q/k script that has a matching name. This process is done via a
depth first search of a tree where each node corresponds to part of the path, and each leaf
corresponds to a possible file. Only the first file found that exists is executed. If none of the
files exist then an `AttributeError` is raised.

The layers of the tree are as follows:

- Each of the paths in `pykx.q.paths`/`pykx.ipc.Connection(...).paths` (which defaults to
  `pykx.ctx.default_paths`)
- `.` prefix or not
- The name of the attribute accessed (i.e. `pykx.q.script` -> `script`)
- `.q` or `.k`
- No trailing `_` or a trailing `_` ([n.b. why a q/k script path would end with an
  underscore](https://code.kx.com/q/basics/syscmds/#_-hide-q-code))

So for example if `pykx.q.script` was accessed, the context `.script` was not defined in memory in
q, and `paths` was set to `['.', pykx.util.QHOME]`
(where `pykx.util.QHOME == pathlib.Path('/opt/kdb')`), then the following paths would be checked
in order until one is found to exist, or they have all been checked:

1.  `./.script.q`
2.  `./.script.q_`
3.  `./.script.k`
4.  `./.script.k_`
5.  `./script.q`
6.  `./script.q_`
7.  `./script.k`
8.  `./script.k_`
9.  `/opt/kdb/.script.q`
10. `/opt/kdb/.script.q_`
11. `/opt/kdb/.script.k`
12. `/opt/kdb/.script.k_`
13. `/opt/kdb/script.q`
14. `/opt/kdb/script.q_`
15. `/opt/kdb/script.k`
16. `/opt/kdb/script.k_`

## Best Practices

To take full advantage of the automatic script loading one should ensure that every q/k script
defines at most one public context. Ideally every q/k script should define exactly one context,
and the name of the context should be equivalent to the name of the file without the file
extension. For instance, `script.q` should place its definitions within the `.script` namespace.
This ensures that when the context interface executes a script to load a context, it doesn't load
in more contexts than intended. Furthermore the context name matching the file name ensures that
when that file is executed because its name matches the desired context, that context will actually
be defined.

When these best practices cannot be followed it may be impossible to use the automatic loading of
scripts via the context interface. In that case we can resort to manually loading scripts either
by executing the q code `system "l <path to script>"`, or by calling `pykx.q._register` with the
path to the script.

When switching contexts within a script, one should always save the context they were in prior to
their context switch, and then switch back into it afterwards, rather than explicitly switching
into the global context.
"""

from __future__ import annotations

from functools import lru_cache

from pathlib import Path
from typing import Tuple

from . import Q
from .exceptions import PyKXException, QError
from .util import QHOME
from .wrappers import Identity


__all__ = [
    'CurrentDirectory',
    'QContext',
    'ZContext',
    'default_paths',
]


def __dir__():
    return __all__


# We resolve the module paths, so if the path of the current directory is actually used, and then
# the directory is changed, the current directory will no longer be searched for q modules. This
# class represents the current directory no matter where the program is currently executing, while
# still allowing the current directory to be removed from the search path by altering the module
# paths attribute.
class CurrentDirectory(type(Path())):
    """``pathlib.Path`` instance for the current directory regardless of directory changes."""
    def __init__(self):
        super().__init__()

    def resolve(self, *args, **kwargs):
        return self


default_paths = [
    CurrentDirectory(),
    QHOME.resolve(strict=True),
    (Path(__file__).parent/'lib').resolve(strict=True),
]


def _fully_qualified_name(name: str, parent: QContext) -> str:
    """Constructs the fully qualified name of a context given its name and parent."""
    names = [name]
    while parent is not None:
        names.append(parent._name)
        parent = parent._parent
    return '.'.join(reversed(names))


class QContext:
    def __init__(self, q: Q, name: str, parent: QContext):
        """Interface to a q context.

        Members of the context be accessed as if the `QContext` object was a dictionary, or by
        dotting off of the `QContext` object.

            q: The q instance in which the context exists.
            name: The name of the context.
            parent: The parent context as a `QContext`, or `None` in the case of the global context.
        """
        super().__setattr__('_q', q)
        super().__setattr__('_name', name)
        super().__setattr__('_parent', parent)
        super().__setattr__('_fqn', _fully_qualified_name(name, parent))
        super().__setattr__('_prev_ctx_stack', [])

    @property
    def _context_keys(self) -> Tuple[str]:
        return (
            *(x for x in self._q('key', self._fqn, sync=True).py() if x),
            *(() if self._fqn else ('z',)),
        )

    def __iter__(self):
        return iter(self._context_keys)

    def _invalidate_cache(self):
        """Clears the cached context, forcing it to be reloaded the next time it is accessed."""
        if hasattr(self.__getattr__, 'cache_clear'):
            self.__getattr__.cache_clear()

    @lru_cache(maxsize=None)
    def __getattr__(self, key):
        if key == 'z' and self._fqn == '':
            return ZContext(self)
        if self._fqn in {'', '.q'} and key in self._q.reserved_words:
            # Reserved words aren't actually part of the `.q` context dict
            return self._q(key, sync=True)
        try:
            attr = self._q(
                'k){x:. x;if[99h~@x;if[` in!x;if[(::)~x`;:`$"_pykx_ctx_marker"]]]x}',
                f'{self._fqn}.{key}',
                sync=True
            )
        except QError:
            raise AttributeError(
                f"'{type(self).__module__}.{type(self).__name__}' object has no attribute {key!r}"
            ) from None
        if attr.t == -11 and attr.py() == '_pykx_ctx_marker':
            return QContext(self._q, key, self)
        return attr

    def __setattr__(self, key, value):
        if key in self._q.reserved_words or (self._fqn == '' and key in self._q.q):
            raise PyKXException('Cannot assign to reserved word or overwrite q namespace.')
        self._q('set', f'{self._fqn}.{key}', value, sync=True)
        self._invalidate_cache()

    def __delattr__(self, key):
        if self._fqn == '':
            raise PyKXException('Cannot delete from the global context.')
        self._q('{![x;();0b;enlist y]}', self._fqn, key, sync=True)
        self._invalidate_cache()

    __getitem__ = __getattr__
    __setitem__ = __setattr__
    __delitem__ = __delattr__

    def __enter__(self):
        # HACK: It's difficult to do a type-check here because that would introduce a cyclic import
        # error. Instead we use `hasattr(self._q, 'fileno')` as a proxy for the type check.
        if hasattr(self._q, 'fileno'): # IPC connections have a 'fileno' attribute.
            raise PyKXException('Context cannot be switch over IPC.')
        self._prev_ctx_stack.append(self._q(
            '{r:system"d";system "d ",string x;r}',
            self._fqn if self._fqn else '.',
            sync=True
        ))

    def __exit__(self, ex_type, ex_value, ex_traceback):
        self._q('{system "d ",string x}', self._prev_ctx_stack.pop(), sync=True)

    def __repr__(self):
        return f'<{type(self).__module__}.{type(self).__name__} of {self._fqn} with ' \
               f'[{", ".join(self._context_keys)}]>'

    def __dir__(self):
        return sorted({*dir(super()), *self._context_keys})


class ZContext(QContext):
    """Special interface to handle the .z context.

    The .z context in q is not a normal context; it lacks a dictionary. To access it one must
    access its attributes directly.
    """
    _no_default = ('ac', 'bm', 'exit', 'pc', 'pd', 'pg', 'ph', 'pi', 'pm',
                   'po', 'pp', 'pq', 'ps', 'pw', 'vs', 'wc', 'wo', 'ws', 'zd')

    _has_default = ('a', 'b', 'c', 'D', 'd', 'e', 'f', 'H', 'h', 'i', 'K', 'k',
                    'l', 'N', 'n', 'o', 'P', 'p', 'q', 'T', 't', 'u', 'W', 'w',
                    'X', 'x')

    _unsupported_keys_with_msg = {
        's': 'cannot refer to self outside of a function, so .z.s is not '
             'exposed through the context interface.',
        'ex': '.z.ex is only available during a debugging session, so it is '
              'not exposed through the context interface.',
        'ey': '.z.ey is only available during a debugging session, so it is '
              'not exposed through the context interface.',
        'ts': '.z.ts is not exposed through the context interface because the '
              'main loop is inactive in PyKX.',
        'z': 'The q datetime type is deprecated, and so the datetime '
             'object .z.z is not exposed through the context interface.',
        'Z': 'The q datetime type is deprecated, and so the datetime '
             'object .z.Z is not exposed through the context interface.'
    }

    _context_keys = {*_has_default, *_no_default}

    def __init__(self, global_context: QContext):
        super().__init__(global_context._q, 'z', global_context)

    def __getattr__(self, key):
        if key in self._no_default:
            try:
                return self._q(f'.z.{key}', sync=True)
            except QError:
                pass
            return self._q('::', sync=True)
        elif key in self._has_default:
            return self._q(f'.z.{key}', sync=True)
        elif key in self._unsupported_keys_with_msg:
            raise AttributeError(f'{key}: {self._unsupported_keys_with_msg[key]}')
        raise AttributeError(key)

    def __setattr__(self, key, value):
        if key in self._no_default and (value is None or isinstance(value, Identity)):
            self._q(f'\\x .z.{key}', sync=True)
        else:
            super().__setattr__(key, value)

    def __delattr__(self, key):
        self.__setattr__(key, None)

    def __dir__(self):
        return sorted({
            *object.__dir__(self),
            *self._has_default,
            *self._no_default,
        })
