from abc import ABCMeta
from textwrap import dedent
from typing import Optional

from . import ipc
from . import Q
from . import toq
from . import wrappers
from .core import keval as _keval, licensed
from .exceptions import LicenseException
from .util import find_core_lib, pykx_dir


__all__ = [
    'EmbeddedQ',
    'q',
]


def __dir__():
    return __all__


# The metaclass of a derived class (EmbeddedQ) must be a (non-strict) subclass of the metaclasses of
# all its bases (ABCMeta from Q).
class ABCMetaSingleton(ABCMeta):
    def __init__(self, name, bases, _dict):
        super().__init__(name, bases, _dict)
        self._singleton_instance = None

    def __call__(self):
        if self._singleton_instance is None:
            self._singleton_instance = super().__call__()
        return self._singleton_instance


class EmbeddedQ(Q, metaclass=ABCMetaSingleton):
    """Interface for q within the current process; can be called to execute q code."""
    def __init__(self):
        if licensed:
            pykx_lib_path = (pykx_dir/'pykx').as_posix()
            self(dedent(f'''
                setenv[`UNDER_PYTHON;"true"]; // env var for embedPy to load appropriately
                2:[`$"{pykx_lib_path}";(`k_pykx_init; 1)][`$"{find_core_lib('q').as_posix()}"];
                `.pykx.i.pyfunc set (`$"{pykx_lib_path}") 2: (`k_pyfunc; 2);
                `.pykx.pyeval set .[;10b] (`$"{pykx_lib_path}")2: (`k_pyrun; 3);
                `.pykx.pyexec set .[;11b] (`$"{pykx_lib_path}")2: (`k_pyrun; 3);
                `.pykx.modpow set {{((`$"{pykx_lib_path}") 2: (`k_modpow; 3))["j"$x;"j"$y;$[z~(::);(::);"j"$z]]}};
            ''')) # noqa: E501
        super().__init__()

    def __repr__(self):
        return 'pykx.q'

    def __call__(self, query: str, *args, sync: Optional[bool] = None) -> wrappers.K:
        """Run code in the q instance.

        Parameters:
            query: The code to run in the q instance.
            args: Arguments to the q query. Each argument will be converted into a `pykx.K` object.
                Up to 8 arguments can be provided, as that is the maximum supported by q.
            sync: Keyword argument provided solely for conformity with `pykx.QConnection`. All
                queries against the embedded q instance are synchronous regardless of what this
                parameter is set to. Setting this keyword argument to `False` results in q generic
                null (`::`) being returned, so as to conform with `pykx.QConnection`. This
                conformity enables one to call any `pykx.Q` instance the same way regardless of
                whether it is a `pykx.EmbeddedQ` or `pykx.QConnection` instance. For cases where
                the query executing asynchronously (and returning after it has been issued, rather
                than after is is done executing) is actually important, one can discriminate
                between `pykx.Q` instances using `isinstance` as normal.

        Returns:
            The value obtained by evaluating the `query` within the current process.
        """
        if not licensed:
            raise LicenseException("run q code via 'pykx.q'")
        if len(args) > 8:
            raise TypeError('Too many parameters - q queries cannot have more than 8 parameters')
        result = _keval(query.encode(), *[wrappers.K(x) for x in args])
        if sync is None or sync:
            return wrappers.factory(result, incref=False)
        else:
            return self('::', sync=True)


q = EmbeddedQ()

# HACK: some modules are relient on the EmbeddedQ instance, so we provide it to them here
ipc._init(q)
toq._init(q)
wrappers._init(q)
