"""PyKX q IPC interface.

The IPC communication module provided here works differently than may be expected for users
familiar with the KX IPC interfaces provided for Java and C#. Unlike these interfaces it does not
directly convert the encoded data received over the q IPC protocol to an analogous type in Python,
but rather stores the object within q memory space as a `pykx.K` object for defered conversion.

This has major benefits with regards to the flexibility of the interface. In particular, the
[`pykx.K`][pykx.K] conversion methods (i.e. `py`, `np`, `pd`, and `pa`), use the same logic as they
do when converting `pykx.K` objects that were created by an embedded q instance.

The IPC interface works when running with or without a q license. Refer to
[the modes of operation documentation](../modes.md) for more details.

Currently functions pulled in over IPC execute locally within PyKX.

!!! Warning
    There are plans to change this behaviour so that when `q` is an instance of `QConnection` each
    of these examples would run on the `q` server.

```python
# Executes locally
q.save('t')
q('save')('t')
# Executes on the server
q('save', 't')
```
"""

from libc.stdint cimport *

from pykx cimport core

from ctypes import c_void_p, CDLL
import os
from typing import List, Optional, Union
from weakref import finalize

from . import Q
from .core import licensed
from .exceptions import PyKXException
from .util import find_core_lib, get_default_args, normalize_to_bytes, normalize_to_str, \
    pykx_dir, system
from .wrappers import Function, K, factory


__all__ = [
    'QConnection',
]


def __dir__():
    return __all__


def _init(_q):
    global q
    q = _q


class QConnection(Q):
    _lib = CDLL(str(find_core_lib('e')))

    _ipc_errors = {
        0: 'Authentication error',
        -1: 'Connection error',
        -2: 'Timeout error',
        -3: 'OpenSSL initialization failed',
    }

    def __init__(self,
                 host: Union[str, bytes] = 'localhost',
                 port: int = None,
                 *,
                 username: Union[str, bytes] = '',
                 password: Union[str, bytes] = '',
                 timeout: float = 0.0,
                 large_messages: bool = True,
                 tls: bool = False,
                 unix: bool = False,
                 sync: bool = True,
    ):
        """Interface with a q process using the q IPC protocol.
        
        Instances of this class represent an open connection to a q process, which can be sent
        messages synchronously or asynchronously by calling it as a function.

        Parameters:
            host: The host name to which a connection is to be established.
            port: The port to which a connection is to be established.
            username: Username for q connection authorization.
            password: Password for q connection authorization.
            timeout: Timeout for blocking socket operations in seconds. If set to `0`, the socket
                will be non-blocking.
            large_messages: Whether support for messages >2GB should be enabled.
            tls: Whether TLS should be used.
            unix: Whether a Unix domain socket should be used instead of TCP. If set to `True`, the
                host parameter is ignored.
            sync: Whether a response should be sent (and waited for) after sending a message to q.
                Can be overridden on a per-call basis.

        Examples:

        Connect to a q process on localhost with a required username and password.

        ```python
        pykx.QConnection('localhost', 5001, 'username', 'password')
        ```

        Connect to a q process at IP address 127.0.0.0, on port 5000 with a timeout of 2 seconds
        and TLS enabled.

        ```python
        pykx.QConnection('127.0.0.1', 5001, timeout=2.0, tls=True)
        ```

        Connect to a q process via a Unix domain socket on port 5001

        ```python
        pykx.QConnection(port=5001, unix=True)
        ```
        """
        object.__setattr__(self, '_connection_info', {
            'host': host,
            'port': port,
            'username': username,
            'password': password,
            'timeout': timeout,
            'large_messages': large_messages,
            'tls': tls,
            'unix': unix,
            'sync': sync,
        })
        if system == 'Windows' and unix: # nocov
            raise TypeError('Unix domain sockets cannot be used on Windows')
        if port is None or not isinstance(port, int):
            raise TypeError('IPC port must be provided')
        credentials = f'{normalize_to_str(username, "Username")}:' \
                      f'{normalize_to_str(password, "Password")}'
        if licensed:
            parts = (
                'unix://' if unix else '',
                'tcps://' if tls else '',
                f'{normalize_to_str(host, "Host")}:' if not unix else '',
                str(port),
                ':',
                credentials,
            )
            object.__setattr__(self, '_handle', int(q.hopen(
                (q.hsym(''.join(parts)), int(timeout * 1000))
            )))
        else:
            host = b'0.0.0.0' if unix else normalize_to_bytes(host, 'Host')
            object.__setattr__(self, '_handle', self._lib.khpunc(
                host,
                port,
                credentials.encode(),
                int(timeout * 1000),
                int(large_messages) | (2 * int(tls))
            ))
            if self._handle <= 0:
                raise PyKXException(self._ipc_errors.get(self._handle, 'Unknown IPC error'))
        object.__setattr__(self, 'closed', False)
        object.__setattr__(self, '_finalizer', finalize(self, self._close, self._handle, self._lib.kclose))
        super().__init__()

    def __repr__(self):
        kwargs = get_default_args(type(self))
        for param_name in tuple(kwargs):
            arg = self._connection_info[param_name]
            if arg == kwargs[param_name]: # Remove if equal to default value
                del kwargs[param_name]
            elif param_name == 'password': # Prevent the password from being logged
                kwargs[param_name] = '********'
            else:
                kwargs[param_name] = arg
        return f'pykx.QConnection({", ".join(f"{k}={v!r}" for k, v in kwargs.items())})'

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @staticmethod
    def _close(handle, kclose) -> None:
        try:
            if licensed:
                q.hclose(handle)
            elif isinstance(handle, int):
                kclose(handle)
        except Exception:
            pass

    def close(self) -> None:
        """Close the connection.

        Examples:

        Open and subsequently close a connection to a q process on localhost:
        
        ```python
        q = pykx.QConnection('localhost', 5001)
        q.close()
        ```

        Using this class with a with-statement should be preferred:

        ```python
        with QConnection('localhost', 5001) as q:
            # do stuff with q
            pass
        # q is closed automatically
        ```
        """
        if not self.closed:
            object.__setattr__(self, 'closed', True)
            self._finalizer()

    @staticmethod
    def _licensed_call(handle: int, query: bytes, parameters: List[K], sync: bool) -> K:
        ret = q(f'{{{handle} x}}', [query, *parameters] if parameters else query)
        if not sync:
            q(f'{handle}[]')
        return ret

    # TODO: can we switch over to exclusively using this approach instead of `_licensed_call`?
    # It would involve making `cls._lib` be either libq or libe depending on if we're licensed.
    @classmethod
    def _unlicensed_call(cls, handle: int, query: bytes, parameters: List[K], sync: bool) -> K:
        parameters = [c_void_p(<uintptr_t>core.r1(<core.K><uintptr_t>x._addr)) for x in parameters]
        cls._lib.k.restype = c_void_p
        val = cls._lib.k(handle, query, *parameters, None)
        if sync:
            return factory(val)
        # TODO: if not sync, flush the handle
        if val == 0:
            raise PyKXException('Async query network error')
        return K(None)

    def __call__(self,
                 query: Union[str, bytes] = '',
                 *parameters,
                 sync: Optional[bool] = None,
    ) -> K:
        """Evaluate a query on the connected q process over IPC.

        Typically the `query` parameter is the name of a function defined on the remote process,
        and `parameters` are the parameters (Python or K objects) for that function. In the
        absence of any parameters the `query` is executed independently.

        Parameters:
            query: A q expression to be evaluated.
            parameters: Parameters for the query. These parameters are converted to q objects
                using `pykx.toq`. This should be taken into account when using Python objects for q
                queries with specific type requirements.
            sync: Whether the query should be executed synchronous (and the response waited for)
                or asynchronous (and the response discarded).
                will use the `sync` instance variable which is set when the `QConnection` is
                initialized.

        Raises:
            RuntimeError: a closed IPC connection was used.
            TypeError: Too many parameters were provided - q queries cannot have more than 8
                parameters.
            ValueError: Attempted to send a Python function over IPC.

        Examples:

        ```python
        q = pykx.QConnection(host='localhost', port=5002)
        ```

        Call an anonymous function with 2 parameters

        ```python
        q('{y+til x}', 10, 5)
        ```

        Execute a q query with no parameters

        ```python
        q('til 10')
        ```

        Call an anonymous function asynchronously with 3 parameters

        ```python
        q('{x set y+til z}', 'async_query', 10, 5, sync=False)
        ```

        Call an anonymous function asynchronously by default with 3 parameters

        ```python
        q = pykx.QConnection(host='localhost', port=5002, sync=False)
        # Because `sync=False`, all calls on this q instance are async by default:
        q('{x set y+til z}', 'async_query', 10, 5)
        # But we can issue sync calls by overriding the `sync` option on a per-call basis:
        q('{x set y+til z}', 'async_query', 10, 5, sync=True)
        ```
        """
        if self.closed:
            raise RuntimeError('Attempted to use a closed IPC connection')
        if len(parameters) > 8:
            raise TypeError('Too many parameters - q queries cannot have more than 8 parameters')
        prev_types = [type(x) for x in parameters]
        if sync is None:
            sync = self._connection_info['sync']
        handle = self._handle if sync else -self._handle
        parameters = [K(x) for x in parameters]
        for a, b in zip(prev_types, (type(x) for x in parameters)):
            if issubclass(b, Function) and not issubclass(a, Function):
                raise ValueError('Cannot send Python function over IPC')
        handler = self._licensed_call if licensed else self._unlicensed_call
        return handler(handle, normalize_to_bytes(query, 'Query'), parameters, sync)

    def fileno(self):
        """The handle of the connection."""
        return self._handle


# Patch QConnection to remove context interface until it has been fixed. The current implementation
# has some severe issues around how it handles functions, which is most of what the context
# interface is used for.
QConnection.__getattr__ = object.__getattribute__
QConnection.__setattr__ = object.__setattr__
QConnection.__delattr__ = object.__delattr__
