#define KXVER 3

#ifdef _WIN32
#define EXPORT __declspec(dllexport)
#else
#define EXPORT
#endif

#include <dlfcn.h>
#include <stdio.h>
#include "k.h"
#include "Python.h"


static void (*r0_ptr)(K);
static K (*r1_ptr)(K);
static K (*kj_ptr)(J);
static K (*ks_ptr)(S);
static K (*ktn_ptr)(I, J);

static PyObject* builtins;
static PyObject* toq_module;
static PyObject* toq;
static PyObject* wrappers_module;
static PyObject* exception_tracker;
static PyObject* factory;
static PyObject* UUID;
static PyObject* thread_get_ident;
static PyObject* POSITIONAL_ONLY;
static PyObject* POSITIONAL_OR_KEYWORD;
static PyObject* VAR_POSITIONAL;
static PyObject* KEYWORD_ONLY;
static PyObject* VAR_KEYWORD;
static PyObject* error_preamble;


EXPORT K k_pykx_init(K k_q_lib_path) {
    void* q_lib = dlopen(k_q_lib_path->s, RTLD_NOW | RTLD_GLOBAL);
    r0_ptr = (void (*)(K))dlsym(q_lib, "r0");
    r1_ptr = (K (*)(K))dlsym(q_lib, "r1");
    kj_ptr = (K (*)(J))dlsym(q_lib, "kj");
    ks_ptr = (K (*)(S))dlsym(q_lib, "ks");
    ktn_ptr = (K (*)(I, J))dlsym(q_lib, "ktn");

    PyGILState_STATE gstate;
    gstate = PyGILState_Ensure();

    builtins = PyModule_GetDict(PyImport_ImportModule("builtins"));
    toq_module = PyModule_GetDict(PyImport_AddModule("pykx.toq"));
    toq = PyDict_GetItemString(toq_module, "toq");
    wrappers_module = PyModule_GetDict(PyImport_AddModule("pykx.wrappers"));
    exception_tracker = PyDict_GetItemString(wrappers_module, "_current_exception");
    factory = PyDict_GetItemString(wrappers_module, "factory");
    UUID = PyDict_GetItemString(PyModule_GetDict(PyImport_AddModule("uuid")), "UUID");
    thread_get_ident = PyDict_GetItemString(PyModule_GetDict(PyImport_AddModule("threading")), "get_ident");
    PyObject* Parameter = PyDict_GetItemString(PyModule_GetDict(PyImport_AddModule("inspect")), "Parameter");
    POSITIONAL_ONLY = PyObject_GetAttrString(Parameter, "POSITIONAL_ONLY");
    POSITIONAL_OR_KEYWORD = PyObject_GetAttrString(Parameter, "POSITIONAL_OR_KEYWORD");
    VAR_POSITIONAL = PyObject_GetAttrString(Parameter, "VAR_POSITIONAL");
    KEYWORD_ONLY = PyObject_GetAttrString(Parameter, "KEYWORD_ONLY");
    VAR_KEYWORD = PyObject_GetAttrString(Parameter, "VAR_KEYWORD");
    error_preamble = PyUnicode_FromString("Error in pykx q extension library: ");

    PyGILState_Release(gstate);
    return (K)0;
}


K k_py_error() {
    if (!PyErr_Occurred()) return (K)0;

    PyObject* ex_type;
    PyObject* ex_value;
    PyObject* ex_traceback;
    PyErr_Fetch(&ex_type, &ex_value, &ex_traceback);
    PyErr_NormalizeException(&ex_type, &ex_value, &ex_traceback);
    if (ex_traceback) PyException_SetTraceback(ex_value, ex_traceback);

    // Build a q error object with the repr of the exception value as its message. The full
    // traceback is provided as the cause to the QError that will be raised.
    PyObject* ex_repr = PyObject_CallMethod(ex_value, "__repr__", NULL);
    K k = ks_ptr(PyUnicode_AsUTF8(ex_repr));
    k->t = -128;
    Py_DECREF(ex_repr);

    // Store the Python exception in `k._current_exception` so that when the k module raises the
    // QError it can provide the Python exception as cause via
    // `raise QError from _current_exception[threading.getident()]`.
    // We use a dict to track exceptions on a per-thread basis. Once the exception has been used
    // by the k module, it is set to `None` in the `exception_tracker` dictionary.
    PyObject* thread_id = PyObject_CallFunction(thread_get_ident, NULL);
    if (PyDict_SetItem(exception_tracker, thread_id, ex_value) && PyErr_Occurred()) {
        PyErr_WriteUnraisable(ex_value);
    }
    Py_DECREF(thread_id);

    return k;
}


void construct_args_kwargs(PyObject* params, PyObject** args, PyObject** kwargs, size_t num_args, PyObject* py_k_args[8]) {
    PyObject* param;
    PyObject* kind;
    PyObject* name;
    PyObject* tmp; // temp var to aid with proper reference management
    *args = PyList_New(0);
    *kwargs = PyDict_New();
    for (size_t i = 0; i < num_args - 1; ++i) {
        param = PyList_GetItem(params, i);
        kind = PyObject_GetAttrString(param, "kind");
        if (PyObject_RichCompareBool(kind, POSITIONAL_ONLY, Py_EQ) || PyObject_RichCompareBool(kind, POSITIONAL_OR_KEYWORD, Py_EQ)) {
            PyList_Append(*args, py_k_args[i]);
        }
        else if (PyObject_RichCompareBool(kind, VAR_POSITIONAL, Py_EQ)) {
            for (Py_ssize_t j = 0; j < PySequence_Size(py_k_args[i]); ++j) {
                tmp = PySequence_GetItem(py_k_args[i], j);
                PyList_Append(*args, tmp);
                Py_DECREF(tmp);
            }
        }
        else if (PyObject_RichCompareBool(kind, KEYWORD_ONLY, Py_EQ)) {
            name = PyObject_GetAttrString(param, "name");
            PyDict_SetItem(*kwargs, name, py_k_args[i]);
            Py_DECREF(name);
        }
        else if (PyObject_RichCompareBool(kind, VAR_KEYWORD, Py_EQ)) {
            // turn the keys into strings, because the keys of a kwargs dict must be strings, not pykx.SymbolAtom objects
            PyObject* d = PyObject_CallFunctionObjArgs(PyDict_GetItemString(builtins, "dict"), py_k_args[i], NULL);
            PyObject* tmp = PyObject_CallMethod(d, "keys", NULL);
            PyObject* keys = PyObject_CallFunctionObjArgs(PyDict_GetItemString(builtins, "list"), tmp, NULL);
            Py_DECREF(tmp);
            for (Py_ssize_t j = 0; j < PySequence_Size(keys); ++j) {
                tmp = PySequence_GetItem(keys, j);
                PyObject* key = PyObject_CallMethod(tmp, "py", NULL);
                PyObject* value = PyObject_CallMethod(d, "pop", "(O)", tmp);
                Py_DECREF(tmp);
                PyDict_SetItem(d, key, value);
                Py_DECREF(value);
                Py_DECREF(key);
            }
            Py_DECREF(keys);
            PyDict_Update(*kwargs, d);
            Py_DECREF(d);
        }
        Py_DECREF(kind);
    }
    tmp = *args;
    *args = PyList_AsTuple(*args);
    Py_DECREF(tmp);
}


EXPORT K k_pyfunc(K k_guid_string, K k_args) {
    PyGILState_STATE gstate;
    gstate = PyGILState_Ensure();
    K k = (K)0; // the K object which will be returned

    PyObject* py_k_args[8]; // args to the Python callable as `pykx.K` objects
    // use `pykx.wrappers.factory` to convert each of the K objects to Python objects
    for (int i = 0; i < k_args->n - 1; ++i) {
        py_k_args[i] = PyObject_CallFunction(factory, "(L)", (uintptr_t)kK(k_args)[i+1]);
        if ((k = k_py_error())) {
            PyGILState_Release(gstate);
            return k;
        }
    }

    // get the uuid for the python function
    PyObject* guid_string = PyUnicode_FromStringAndSize((const char*)k_guid_string->G0, k_guid_string->n);
    if ((k = k_py_error())) {
        PyGILState_Release(gstate);
        return k;
    }
    PyObject* guid = PyObject_CallFunctionObjArgs(UUID, guid_string, NULL);
    if ((k = k_py_error())) {
        PyGILState_Release(gstate);
        return k;
    }

    // get the python function
    PyObject* converted_callables = PyDict_GetItemString(toq_module, "converted_callables");
    PyObject* pyfunc_tuple = PyDict_GetItemWithError(converted_callables, guid);
    if (!pyfunc_tuple) PyErr_SetObject(PyExc_KeyError, guid);
    if ((k = k_py_error())) {
        PyGILState_Release(gstate);
        return k;
    }
    Py_DECREF(guid);
    PyObject* pyfunc = PyTuple_GET_ITEM(pyfunc_tuple, 0);
    PyObject* params = PyTuple_GET_ITEM(pyfunc_tuple, 1);
    if ((k = k_py_error())) {
        PyGILState_Release(gstate);
        return k;
    }
    Py_INCREF(pyfunc);
    Py_INCREF(params);

    // construct the args and kwargs for the pyfunc
    PyObject* args; // positional arguments for Python callables
    PyObject* kwargs; // keyword arguments for Python callables
    construct_args_kwargs(params, &args, &kwargs, k_args->n, py_k_args);
    Py_DECREF(params);
    if ((k = k_py_error())) {
        Py_XDECREF(args);
        Py_XDECREF(kwargs);
        Py_DECREF(pyfunc);
        PyGILState_Release(gstate);
        return k;
    }

    // call the python function
    PyObject* py_ret = PyObject_Call(pyfunc, args, kwargs);
    Py_XDECREF(args);
    Py_XDECREF(kwargs);
    Py_DECREF(pyfunc);
    if ((k = k_py_error())) {
        PyGILState_Release(gstate);
        return k;
    }

    // Convert the returned value to q
    PyObject* py_k_ret = PyObject_CallFunctionObjArgs(toq, py_ret, NULL);
    Py_DECREF(py_ret);
    if ((k = k_py_error())) {
        PyGILState_Release(gstate);
        return k;
    }
    PyObject* py_addr = PyObject_GetAttrString(py_k_ret, "_addr");
    k = (K)PyLong_AsLongLong(py_addr);
    Py_DECREF(py_addr);

    PyGILState_Release(gstate);
    return k;
}


// k_eval_or_exec == 0 -> eval the code string
// k_eval_or_exec == 1 -> exec the code string
EXPORT K k_pyrun(K k_ret, K k_eval_or_exec, K k_code_string) {
    PyGILState_STATE gstate;
    gstate = PyGILState_Ensure();
    K k;
    void* code_string = PyMem_Calloc(k_code_string->n + 1, 1);
    strncpy(code_string, (const char *)k_code_string->G0, k_code_string->n);
    PyObject* ctx = PyModule_GetDict(PyImport_AddModule("__main__"));
    PyObject* py_ret = PyRun_String(
        code_string,
        k_eval_or_exec->g ? Py_file_input : Py_eval_input,
        ctx,
        ctx
    );
    PyMem_Free(code_string);

    if (!k_ret->g) {
        if (PyErr_Occurred()) PyErr_Print();
        else Py_DECREF(py_ret);
        PyGILState_Release(gstate);
        return (K)0;
    }

    if ((k = k_py_error())) {
        PyGILState_Release(gstate);
        return k;
    }
    PyObject* py_k_ret = PyObject_CallFunctionObjArgs(toq, py_ret, NULL);
    Py_DECREF(py_ret);
    if ((k = k_py_error())) {
        PyGILState_Release(gstate);
        return k;
    }
    PyObject* py_addr = PyObject_GetAttrString(py_k_ret, "_addr");
    k = (K)PyLong_AsLongLong(py_addr);
    Py_DECREF(py_addr);
    PyGILState_Release(gstate);
    return k;
}


inline J modpow(J base, J exp, J mod) {
    J result = 1;
    while (exp > 0) {
        if (exp & 1) result = (result * base) % mod;
        base = (base * base) % mod;
        exp >>= 1;
    }
    return result;
}


EXPORT K k_modpow(K k_base, K k_exp, K k_mod_arg) {
    K result;
    K k_mod;

    if (k_mod_arg->t == 101 && k_mod_arg->g == 0) {
        k_mod = kj_ptr(9223372036854775807); // default modulo is 2^63-1
    }
    else {
        k_mod = r1_ptr(k_mod_arg);
    }

    if (k_base->t >= 0 && k_exp->t >= 0 && k_mod->t >= 0) {
        if (k_base->n != k_exp->n || k_exp->n != k_mod->n) {
            result = ks_ptr("length");
            result->t = -128;
        } else {
            result = ktn_ptr(7, k_base->n);
            for (J x = 0; x < k_base->n; ++x) kJ(result)[x] = modpow(kJ(k_base)[x], kJ(k_exp)[x], kJ(k_mod)[x]);
        }
    } else if (k_base->t >= 0 && k_exp->t >= 0) {
        if (k_base->n != k_exp->n) {
            result = ks_ptr("length");
            result->t = -128;
        }
        else {
            result = ktn_ptr(7, k_base->n);
            for (J x = 0; x < k_base->n; ++x) kJ(result)[x] = modpow(kJ(k_base)[x], kJ(k_exp)[x], k_mod->j);
        }
    } else if (k_base->t >= 0 && k_mod->t >= 0) {
        if (k_base->n != k_mod->n) {
            result = ks_ptr("length");
            result->t = -128;
        }
        result = ktn_ptr(7, k_base->n);
        for (J x = 0; x < k_base->n; ++x) kJ(result)[x] = modpow(kJ(k_base)[x], k_exp->j, kJ(k_mod)[x]);
    } else if (k_base->t >= 0) {
        result = ktn_ptr(7, k_base->n);
        for (J x = 0; x < k_base->n; ++x) kJ(result)[x] = modpow(kJ(k_base)[x], k_exp->j, k_mod->j);
    } else if (k_exp->t >= 0 && k_mod->t >= 0) {
        if (k_exp->n != k_mod->n) {
            result = ks_ptr("length");
            result->t = -128;
        }
        else {
            result = ktn_ptr(7, k_exp->n);
            for (J x = 0; x < k_exp->n; ++x) kJ(result)[x] = modpow(k_base->j, kJ(k_exp)[x], kJ(k_mod)[x]);
        }
    } else if (k_exp->t >= 0) {
        result = ktn_ptr(7, k_exp->n);
        for (J x = 0; x < k_exp->n; ++x) kJ(result)[x] = modpow(k_base->j, kJ(k_exp)[x], k_mod->j);
    } else if (k_mod->t >= 0) {
        result = ktn_ptr(7, k_mod->n);
        for (J x = 0; x < k_mod->n; ++x) kJ(result)[x] = modpow(k_base->j, k_exp->j, kJ(k_mod)[x]);
    } else {
        result = kj_ptr(modpow(k_base->j, k_exp->j, k_mod->j));
    }
    r0_ptr(k_mod);
    return result;
}
