"""Query interfaces for PyKX."""

from random import choices
from string import ascii_letters
from typing import Dict, List, Optional, Union

from . import Q
from . import wrappers as k


__all__ = [
    'QSQL',
]


def __dir__():
    return __all__


consolidate = '{$[0>type x;x;(0h>v 0)&1~count v:distinct type each x;raze x;x]}'


class QSQL:
    """Generates and submits functional q SQL queries.

    Instances of this class can be accessed as the `qsql` attribute of any [`pykx.Q`][pykx.Q]. For
    instance, `pykx.q.qsql`, or `pykx.QConnection(...).qsql`.

    The `QSQL` class provides Python users with a method of querying q simple, keyed, splayed and
    partitioned tables using a single set of functionality.

    This is achieved by wrapping the logic contained within the q functional select, exec, update,
    and delete functionality. For more information on this functionality please refer to [Chapter 9
    Section 12 of Q for Mortals](https://code.kx.com/q4m3/9_Queries_q-sql/#912-functional-forms).

    While it is also conceivable that the interface could compile a qSQL statement to achieve the
    same end goal there are a number of advantages to using the more complex functional form.

    1. Users that are unfamiliar with q who use the interface are introduced to the more powerful
        version of querying with q, while still operating within a familiar setting.
    2. Using the functional form provides the ability when running functional updates to update the
        q tables with data derived from Python:

        ```python
        qtable = pykx.q('([]1 2 3;4 5 6)')
        pykx.q.qsql.update(qtable, {'x': [10, 20, 30]})
        ```

    3. It makes development and maintenance of the interface easier when dealing across the forms
        of supported table within q within which the functional forms of interacting with tables
        are more natural.
    """

    def __init__(self, q: Q):
        self._q = q

    def select(self,
               table: Union[k.Table, str],
               columns: Optional[Union[Dict[str, str], k.Dictionary]] = None,
               where: Optional[Union[List[str], str, k.SymbolAtom, k.SymbolVector]] = None,
               by: Optional[Union[Dict[str, str], k.Dictionary]] = None
    ) -> k.K:
        """Apply a q style select statement on tables defined within the process.

        This implementation follows the q functional select syntax with limitations on
        structures supported for the various clauses a result of this.

        Parameters:
            table: The q table or name of the table (provided the table is named within the q
                memory space) on which the select statement is to be applied.
            columns: A dictionary mapping the name to be given to a column and the logic to be
                applied in aggregation to that column both as strings.
            where: Conditional filtering used to select subsets of the data on which by-clauses and
                appropriate aggregations are to be applied.
            by: A dictionary mapping the names to be assigned to the produced columns and the
                columns whose results are used to construct the groups of the by clause.

        Examples:

        Define a q table in python, and give it a name in q memory

        ```python
        qtab = pykx.q('([]col1:100?`a`b`c;col2:100?1f;col3:100?0b;col4:100?10f)')
        pykx.q['qtab'] = qtab
        ```

        Select all items in the table

        ```python
        pykx.q.qsql.select(qtab)
        pykx.q.qsql.select('qtab')
        ```

        Filter table based on various where conditions

        ```python
        pykx.q.qsql.select(qtab, where='col2<0.5')
        pykx.q.qsql.select(qtab, where=['col1=`a', 'col2<0.3'])
        ```

        Retrieve statistics by grouping data on symbol columns

        ```python
        pykx.q.qsql.select(qtab, columns={'maxCol2': 'max col2'}, by={'col1': 'col1'})
        pykx.q.qsql.select(qtab, columns={'avgCol2': 'avg col2', 'minCol4': 'min col4'}, by={'col1': 'col1'})
        ```

        Retrieve grouped statistics with restrictive where condition

        ```python
        pykx.q.qsql.select(qtab, columns={'avgCol2': 'avg col2', 'minCol4': 'min col4'}, by={'col1': 'col1'}, where='col3=0b')
        ```
        """ # noqa: E501
        return self._seud(table, 'select', columns, where, by)

    def exec(self,
             table: Union[k.Table, str],
             columns: Optional[Union[Dict[str, str], k.Dictionary]] = None,
             where: Optional[Union[List[str], str, k.SymbolAtom, k.SymbolVector]] = None,
             by: Optional[Union[Dict[str, str], k.Dictionary]] = None
    ) -> k.K:
        """
        Apply a q style exec statement on tables defined within the process.

        This implementation follows the q functional exec syntax with limitations on structures
        supported for the various clauses a result of this.

        Parameters:
            table: The q table or name of the table (provided the table is named within the q
                memory space) on which the exec statement is to be applied.
            columns: A dictionary mapping the name to be given to a column and the logic to be
                applied in aggregation to that column both as strings. A string defining a single
                column to be retrieved from the table as a list.
            where: Conditional filtering used to select subsets of the data on which by clauses and
                appropriate aggregations are to be applied.
            by: A dictionary mapping the names to be assigned to the produced columns and the
                the columns whose results are used to construct the groups of the by clause.

        Examples:

        Define a q table in python and named in q memory

        ```python
        qtab = pykx.q('([]col1:100?`a`b`c;col2:100?1f;col3:100?0b;col4:100?10f)')
        pykx.q['qtab'] = qtab
        ```

        Select last item of the table

        ```python
        pykx.q.qsql.exec(qtab)
        pykx.q.qsql.exec('qtab')
        ```

        Retrieve a column from the table as a list

        ```python
        pykx.q.qsql.exec(qtab, 'col3')
        ```

        Retrieve a set of columns from a table as a dictionary

        ```python
        pykx.q.qsql.exec(qtab, {'symcol': 'col1'})
        pykx.q.qsql.exec(qtab, {'symcol': 'col1', 'boolcol': 'col3'})
        ```

        Filter columns from a table based on various where conditions

        ```python
        pykx.q.qsql.exec(qtab, 'col3', where='col1=`a')
        pykx.q.qsql.exec(qtab, {'symcol': 'col1', 'maxcol4': 'max col4'}, where=['col1=`a', 'col2<0.3'])
        ```

        Retrieve data grouping by data on symbol columns

        ```python
        pykx.q.qsql.exec(qtab, 'col2', by={'col1': 'col1'})
        pykx.q.qsql.exec(qtab, columns={'maxCol2': 'max col2'}, by={'col1': 'col1'})
        pykx.q.qsql.exec(qtab, columns={'avgCol2': 'avg col2', 'minCol4': 'min col4'}, by={'col1': 'col1'})
        ```

        Retrieve grouped statistics with restrictive where condition

        ```python
        pykx.q.qsql.exec(qtab, columns={'avgCol2': 'avg col2', 'minCol4': 'min col4'}, by={'col1': 'col1'}, where='col3=0b')
        ```
        """ # noqa: E501
        return self._seud(table, 'exec', columns, where, by)

    def update(self,
               table: Union[k.Table, str],
               columns: Optional[Union[Dict[str, str], k.Dictionary]] = None,
               where: Optional[Union[List[str], str, k.SymbolAtom, k.SymbolVector]] = None,
               by: Optional[Union[Dict[str, str], k.Dictionary]] = None,
               modify: bool = False
    ) -> k.K:
        """
        Apply a q style update statement on tables defined within the process.

        This implementation follows the q functional update syntax with limitations on
        structures supported for the various clauses a result of this.

        Parameters:
            table: The q table or name of the table (provided the table is named within the q
                memory space) on which the update statement is to be applied.
            columns: A dictionary mapping the name of a column present in the table or one to be
                added to the contents which are to be added to the column, this content can be a
                string denoting q data or the equivalent Python data.
            where: Conditional filtering used to select subsets of the data on which by-clauses and
                appropriate aggregations are to be applied.
            by: A dictionary mapping the names to be assigned to the produced columns and the
                columns whose results are used to construct the groups of the by clause.
            modify: Whether the result of a delete is to be saved. This holds when `table` is the
                name of a table in q memory, as outlined at:
                https://code.kx.com/q/basics/qsql/#result-and-side-effects.

        Examples:

        Define a q table in python and named in q memory

        ```python
        qtab = pykx.q('([]name:`tom`dick`harry; age:28 29 35; hair:`fair`dark`fair; eye:`green`brown`gray)')
        pykx.q['qtab'] = qtab
        ```

        Update all the contents of a column

        ```python
        pykx.q.qsql.update(qtab, {'eye': '`blue`brown`green'})
        pykx.q.qsql.update(qtab, {'age': [25, 30, 31]})
        ```

        Update the content of a column restricting scope using a where clause

        ```python
        pykx.q.qsql.update(qtab, {'eye': ['blue']}, where='hair=`fair')
        ```

        Define a q table suitable for by clause example

        ```python
        byqtab = pykx.q('([]p:100?`a`b`c;name:100?`nut`bolt`screw;color:100?`red`green`blue;weight:0.5*100?20;city:100?`london`paris`rome)')
        pykx.q['byqtab'] = byqtab
        ```

        Apply an update grouping based on a by phrase

        ```python
        pykx.q.qsql.update(byqtab, {'weight': 'avg weight'}, by={'city': 'city'})
        ```

        Apply an update grouping based on a by phrase and persist the result using the modify keyword

        ```python
        pykx.q.qsql.update('byqtab', columns={'weight': 'avg weight'}, by={'city': 'city'}, modify=True)
        pykx.q['byqtab']
            ```
        """ # noqa: E501
        return self._seud(table, 'update', columns, where, by, modify)

    def delete(self,
               table: Union[k.Table, str],
               columns: Optional[Union[List[str], k.SymbolVector]] = None,
               where: Optional[Union[List[str], str, k.SymbolAtom, k.SymbolVector]] = None,
               modify: bool = False
    ) -> k.K:
        """
        Apply a q style delete statement on tables defined within the process.

        This implementation follows the q functional delete syntax with limitations on
        structures supported for the various clauses a result of this.

        Parameters:
            table: The q table or name of the table (provided the table is named within the q
                memory space) on which the delete statement is to be applied.
            columns: Denotes the columns to be deleted from a table.
            where: Conditional filtering used to select subsets of the data which are to be
                deleted from the table.
            modify: Whether the result of a delete is to be saved. This holds when `table` is the
                name of a table in q memory, as outlined at:
                https://code.kx.com/q/basics/qsql/#result-and-side-effects.

        Examples:

        Define a q table in python and named in q memory

        ```python
        qtab = pykx.q('([]name:`tom`dick`harry;age:28 29 35;hair:`fair`dark`fair;eye:`green`brown`gray)')
        pykx.q['qtab'] = qtab
        ```

        Delete all the contents of the table

        ```python
        pykx.q.qsql.delete(qtab)
        pykx.q.qsql.delete('qtab')
        ```

        Delete single and multiple columns from the table

        ```python
        pykx.q.qsql.delete(qtab, 'age')
        pykx.q.qsql.delete('qtab', ['age', 'eye'])
        ```

        Delete rows of the dataset based on where condition

        ```python
        pykx.q.qsql.delete(qtab, where='hair=`fair')
        pykx.q.qsql.delete('qtab', where=['hair=`fair', 'age=28'])
        ```

        Delete a column from the dataset named in q memory and persist the result using the
        modify keyword

        ```python
        pykx.q.qsql.delete('qtab', 'age', modify=True)
        pykx.q['qtab']
            ```
        """ # noqa: E501
        if columns is not None and where is not None:
            raise TypeError("'where' and 'columns' clauses cannot be used simultaneously in a "
                            "delete statement")
        return self._seud(table, 'delete', columns, where, None, modify)

    def _seud(self, table, query_type, columns=None, where=None, by=None, modify=False) -> k.K:
        select_clause = self._generate_clause(columns, 'columns', query_type)
        by_clause = self._generate_clause(by, 'by', query_type)
        where_clause = self._generate_clause(where, 'where', query_type)
        original_table = table
        if isinstance(table, k.K):
            if modify:
                raise TypeError("'modify' argument can only be used when 'table' is saved as a "
                                "named object in q memory")
            randstring = ''.join(choices(ascii_letters, k=32))
            table_name = f'.pykx.i._{randstring}'
            self._q[table_name] = table
            table = table_name
        elif not isinstance(table, str):
            raise TypeError("'table' must be a K object or string denoting an item in q memory")
        query_char = '!' if query_type in ('delete', 'update') else '?'
        table_code = f'`$"{table}"' if modify else f'get`$"{table}"'
        try:
            return self._q(
                f'{query_char}[{table_code};;;]',
                where_clause,
                by_clause,
                select_clause,
                sync=True,
            )
        finally:
            if isinstance(original_table, k.K):
                self._q(f'![`.pykx.i;();0b;enlist[`$"_{randstring}"]]', sync=True)

    def _generate_clause(self, clause_value, clause_name, query_type):
        if clause_value is None:
            if clause_name in ('columns', 'where'):
                b = query_type == 'delete' and clause_name == 'columns'
                return self._q('`symbol$()', sync=True) if b else []
            elif clause_name == 'by':
                return [] if query_type == 'exec' else False
        else:
            if clause_name in ('columns', 'by'):
                return self._generate_clause_columns_by(clause_value, clause_name, query_type)
            elif clause_name == 'where':
                return self._generate_clause_where(clause_value)
        raise TypeError(f"Provided clause_name '{clause_name}' not supported")

    def _generate_clause_columns_by(self, clause_value, clause_name, query_type):
        if isinstance(clause_value, dict):
            return self._generate_clause_columns_by_dict(clause_value)
        elif clause_name == 'columns' and query_type == 'delete':
            if isinstance(clause_value, str):
                clause_value = [clause_value]
            return self._q(
                'raze',
                [self._q('parse', k.CharVector(x), sync=True) for x in clause_value],
                sync=True,
            )
        elif (query_type in ['select', 'exec']) and (clause_name in ['columns', 'by']):
            if isinstance(clause_value, list):
                return self._q('{x!x}', self._q(consolidate, clause_value, sync=True), sync=True)
            elif isinstance(clause_value, str) and query_type == 'select':
                return self._q('{x!x}enlist@', clause_value, sync=True)
            return k.K(clause_value)
        elif isinstance(clause_value, k.K):
            return clause_value
        raise TypeError(f"Unsupported type for '{clause_name}' clause")

    def _generate_clause_columns_by_dict(self, clause_value):
        clause_dict = {}
        for key, val in clause_value.items():
            clause_dict[key] = self._q('parse', k.CharVector(val), sync=True) \
                if isinstance(val, str) \
                else self._q(consolidate, val, sync=True)
        return k.K(clause_dict)

    def _generate_clause_where(self, clause_value) -> k.List:
        if isinstance(clause_value, k.List):
            return clause_value # clause value is a parse tree
        if isinstance(clause_value, k.BooleanVector):
            return self._q('enlist', clause_value, sync=True)
        if isinstance(clause_value, str):
            clause_value = [clause_value]
        try:
            return k.K([self._q('parse', k.CharVector(x), sync=True) for x in clause_value])
        except Exception as ex:
            raise TypeError("Unsupported type for 'where' clause") from ex
