"""Converts a Python object to a `pykx.K` object.

`pykx.toq` can be called to automatically select the appropriate `pykx.toq.from_*` function
based on the type of the provided Python object. Refer to the documentation for each of the
`pykx.toq.from_*` functions for more details about how each conversion works.

Note: `pykx.toq.from_*` methods should not be used directly
    `pykx.toq` or one of the `pykx.K` subclassses should instead be used to convert a value into a
    `pykx.K` object. The `pykx.toq.from_*` functions are automatically called by these higher-level
    interfaces. They are documented here to explain what conversions from Python to q are supported,
    and how they are performed.

Instead of calling e.g. `pykx.toq('qwerty', ktype=pykx.CharVector)`, one can instantiate the
desired type directly like so: `pykx.CharVector('qwerty')`.

Examples:

```python
>>> import pykx as kx
>>> kx.toq('grok')
pykx.SymbolAtom(pykx.q('`grok'))
>>> kx.toq('grok', ktype=kx.CharVector)
pykx.CharVector(pykx.q('"grok"'))
```

Parameters:
    x (Any): A Python object which is to be converted into a `pykx.K` object.
    ktype (Union[pykx.K, int], defaults to `None`): Desired `pykx.K` subclass (or type number) for
        the returned value. If `None`, the type is inferred from `x`.

Returns:
    (pykx.K): The provided Python object as an analogous `pykx.K` object.
"""

from libc.stddef cimport *
from libc.stdint cimport *
from libc.string cimport memcpy

cimport numpy as np

from pykx cimport wrappers as k
from pykx cimport core

import datetime
from ctypes import CDLL
from inspect import signature
from pathlib import Path
import platform
import sys
import traceback
from types import ModuleType
from typing import Any, Callable, Optional, Union
from uuid import UUID, uuid4 as random_uuid

import numpy as np
import pandas as pd

from . import wrappers as k
from ._pyarrow import pyarrow as pa
from .core import licensed
from .exceptions import LicenseException, PyArrowUnavailable, PyKXException, QError
from .util import df_from_arrays, find_core_lib, slice_to_range, system


__all__ = [
    'from_arrow',
    'from_bytes',
    'from_callable',
    'from_datetime_timedelta',
    'from_datetime_datetime',
    'from_dict',
    'from_ellipsis',
    'from_fileno',
    'from_float',
    'from_int',
    'from_k',
    'from_list',
    'from_numpy_datetime64',
    'from_numpy_ndarray',
    'from_numpy_timedelta64',
    'from_none',
    'from_pandas_dataframe',
    'from_pandas_index',
    'from_pandas_series',
    'from_pathlib_path',
    'from_range',
    'from_slice',
    'from_str',
    'from_uuid_UUID',
    'from_tuple',
]


def __dir__():
    return __all__


def _init(_q):
    global q
    q = _q


if system == 'Windows': # nocov
    import msvcrt # nocov


cdef inline core.K _k(x):
    return <core.K><uintptr_t>x._addr


# nanoseconds between 1970-01-01 and 2000-01-01
DEF TIMESTAMP_OFFSET = 946684800000000000
#      months between 1970-01-01 and 2000-01-01
DEF MONTH_OFFSET = 360
#        days between 1970-01-01 and 2000-01-01
DEF DATE_OFFSET = 10957


if not licensed:
    # Initialize the q memory system
    CDLL(str(find_core_lib('e'))).khp('', -1)


np_int_types = (
    np.bool_, np.byte, np.ubyte, np.short, np.ushort, np.intc, np.uintc,
    np.int_, np.uint, np.longlong, np.ulonglong, np.int8, np.int16, np.int32,
    np.int64, np.uint8, np.uint16, np.uint32, np.uint64, np.intp, np.uintp
)

np_float_types = (
    np.half, np.float16, np.single, np.double, np.longdouble, np.float32,
    np.float64, np.float_
)

type_number_to_pykx_k_type = {**k.type_number_to_pykx_k_type, **{
    98: k.Table,
    99: k.Dictionary,
    101: k.UnaryPrimative,
}}


_overflow_error_by_size = {
    'long': OverflowError('q long must be in the range [-9223372036854775808, 9223372036854775807]'),
    'int': OverflowError('q int must be in the range [-2147483648, 2147483647]'),
    'short': OverflowError('q short must be in the range [-32768, 32767]'),
    'byte': OverflowError('q byte must be in the range [0, 255]'),
    'boolean': OverflowError('q boolean must be in the range [0, 1]'),
}


def _first_isinstance(x, t):
    try:
        return isinstance(x[0], t)
    except IndexError:
        return False


def _conversion_TypeError(x, input_type, output_type):
    if output_type is None:
        output_type = 'K object'
    x_repr = repr(x) if isinstance(x, str) else f"'{x!r}'"
    return TypeError(f"Cannot convert {input_type} {x_repr} to {output_type}")


KType = Union[k.K, int]


def _resolve_k_type(ktype: KType) -> Optional[k.K]:
    """Resolve the pykx.K type represented by the ktype parameter."""
    if ktype is None:
        return None
    elif isinstance(ktype, int):
        try:
            return type_number_to_pykx_k_type[ktype]
        except KeyError:
            raise TypeError(f'Numeric ktype {ktype!r} does not exist')
    try:
        if issubclass(ktype, k.K):
            return ktype
    except TypeError:
        pass
    raise TypeError(f'ktype {ktype!r} unrecognized')


def _default_converter(x, ktype: Optional[KType] = None):
    raise _conversion_TypeError(x, f'Python object of type {type(x)}', ktype)


def from_none(x: None, ktype: Optional[KType] = None) -> k.Identity:
    """Converts `None` into a `pykx.Identity` object.

    Parameters:
        x: The `None` singleton object.
        ktype: Desired `pykx.K` subclass (or type number) for the returned value. If `None`, the
            type is inferred from `x`. The following values are supported:

            `ktype`         | Returned value
            --------------- | ----------------------------------
            `None`          | Same as for `ktype=pykx.Identity`.
            `pykx.Identity` | Generic null (`::`).

    Returns:
        A `pykx.Identity` instance, i.e. generic null.
    """
    cdef core.K kx
    if licensed:
        return q('::') # `::` always has the same address in memory, which is occasionally relevant
    kx = core.ka(101)
    kx.g = 0
    kx.j = 0
    return k._factory(<uintptr_t>kx, False)


def from_k(x: k.K, ktype: Optional[KType] = None) -> k.K:
    """Converts a `pykx.K` object into a `pykx.K` object.

    Parameters:
        x: A `pykx.K` instance.
        ktype: Desired `pykx.K` subclass (or type number) for the returned value. If `None`, the
            type is inferred from `x`. The following values are supported:

            `ktype`         | Returned value
            --------------- | ---------------------------------------
            `None`          | The provided `pykx.K` object unchanged.
            The type of `x` | Same as for `ktype=None`.

    Raises:
        TypeError: `ktype` is not `None` and does not match the provided `pykx.K` object.

    Returns:
        The provided `pykx.K` instance.
    """
    if ktype is not None and not isinstance(x, ktype):
        raise TypeError(f'{x!r} must be an instance of {ktype!r}')
    return x

def from_int(x: int, ktype: Optional[KType] = None) -> k.IntegralNumericAtom:
    """Converts an `int` into an instance of a subclass of `pykx.IntegralNumericAtom`.

    Parameters:
        x: The `int` that will be converted into a `pykx.IntegralNumericAtom` object.
        ktype: Desired `pykx.K` subclass (or type number) for the returned value. If `None`, the
            type is inferred from `x`. The following values are supported:

            +--------------------+----------------------------------------------------------------+
            |`ktype`             | Returned value                                                 |
            +====================+================================================================+
            | `None`             | Same as for `ktype=pykx.LongAtom`, except when `x` is a        |
            |                    | `bool`, in which case it is the same as for                    |
            |                    | `ktype=pykx.BooleanAtom`.                                      |
            +--------------------+----------------------------------------------------------------+
            | `pykx.LongAtom`    | The `int` as a signed 64 bit integer in q.                     |
            +--------------------+----------------------------------------------------------------+
            | `pykx.IntAtom`     | The `int` as a signed 32 bit integer in q.                     |
            +--------------------+----------------------------------------------------------------+
            | `pykx.ShortAtom`   | The `int` as a signed 16 bit integer in q.                     |
            +--------------------+----------------------------------------------------------------+
            | `pykx.ByteAtom`    | The `int` as a unsigned 8 bit integer in q.                    |
            +--------------------+----------------------------------------------------------------+
            | `pykx.BooleanAtom` | The `int` as a boolean in q.                                   |
            +--------------------+----------------------------------------------------------------+

    Raises:
        TypeError: Unsupported `ktype` for a `float`.
        OverflowError: The provided `int` is out of bounds for the selected
            `pykx.IntegralNumericAtom` subclass.

    Warning: The edges of the bounds have special meanings in q.
        For `pykx.LongAtom`, `pykx.IntAtom`, and `pykx.ShortAtom`, the lower bound of the
        range is the null value for that type, one larger than that is the negative infinity for
        that type, and the upper bound of the range is positive infinity for that type. An
        exception will not be raised if `int` objects with these values are provided. Refer to
        the nulls and infinities PyKX documentation for more information about q nulls and
        infinities, and how to handle them with PyKX.

    Returns:
        An instance of a `pykx.IntegralNumericAtom` subclass.
    """
    cdef core.K kx
    if ktype is None or issubclass(ktype, k.LongAtom):
        if not -9223372036854775808 <= x <= 9223372036854775807:
            raise _overflow_error_by_size['long']
        kx = core.kj(x)
    elif issubclass(ktype, k.IntAtom):
        if not -2147483648 <= x <= 2147483647:
            raise _overflow_error_by_size['int']
        kx = core.ki(x)
    elif issubclass(ktype, k.ShortAtom):
        if not -32768 <= x <= 32767:
            raise _overflow_error_by_size['short']
        kx = core.kh(x)
    elif issubclass(ktype, k.ByteAtom):
        if not 0 <= x <= 255:
            raise _overflow_error_by_size['byte']
        kx = core.kg(x)
    elif issubclass(ktype, k.BooleanAtom):
        if not 0 <= x <= 1:
            raise _overflow_error_by_size['boolean']
        kx = core.kb(x)
    else:
        raise _conversion_TypeError(x, 'Python int', ktype)
    return k._factory(<uintptr_t>kx, False)


def from_float(x: float, ktype: Optional[KType] = None) -> k.NonIntegralNumericAtom:
    """Converts a `float` into an instance of a subclass of `pykx.NonIntegralNumericAtom`.

    Parameters:
        x: The `float` that will be converted into a `pykx.NonIntegralNumericAtom` object.
        ktype: Desired `pykx.K` subclass (or type number) for the returned value. If `None`, the
            type is inferred from `x`. The following values are supported:

            `ktype`          | Returned value
            ---------------- | ------------------------------------------
            `None`           | Same as for `ktype=pykx.FloatAtom`.
            `pykx.FloatAtom` | The `float` as a signed 64 bit float in q.
            `pykx.RealAtom`  | The `float` as a signed 32 bit float in q.

    Raises:
        TypeError: Unsupported `ktype` for a `float`.

    Returns:
        An instance of a `pykx.NonIntegralNumericAtom` subclass.
    """
    cdef core.K kx
    if ktype is None or issubclass(ktype, k.FloatAtom):
        kx = core.kf(x)
    elif issubclass(ktype, k.RealAtom):
        kx = core.ke(x)
    else:
        raise _conversion_TypeError(x, 'Python float', ktype)
    return k._factory(<uintptr_t>kx, False)


def from_str(x: str,
             ktype: Optional[KType] = None,
) -> Union[k.SymbolAtom, k.SymbolVector, k.CharAtom]:
    """Converts a `str` into an instance of a string-like subclass of `pykx.K`.

    Parameters:
        x: The `str` that will be converted into a `pykx.K` object.
        ktype: Desired `pykx.K` subclass (or type number) for the returned value. If `None`, the
            type is inferred from `x`. The following values are supported:

            `ktype`           | Returned value
            ----------------- | ------------------------------------
            `None`            | Same as for `ktype=pykx.SymbolAtom`.
            `pykx.SymbolAtom` | The `str` as a q symbol.
            `pykx.CharVector` | The `str` as a q character vector.
            `pykx.CharAtom`   | The `str` as a q character atom.

    Raises:
        TypeError: Unsupported `ktype` for a `str`.
        ValueError: `ktype=pykx.CharAtom`, but `x` is not 1 character long.

    Returns:
        An instance of a `pykx.SymbolAtom`, `pykx.CharVector`, or `pykx.CharAtom`.
    """
    cdef core.K kx
    cdef bytes as_bytes = x.encode('utf-8')
    if ktype is None or issubclass(ktype, k.SymbolAtom):
        kx = core.ks(as_bytes)
    elif issubclass(ktype, k.CharAtom):
        if len(as_bytes) != 1:
            raise ValueError(
                f"'pykx.CharAtom' can only be created from a single character, not {x!r}"
            )
        kx = core.kc(ord(as_bytes) if x else 32)
    elif issubclass(ktype, k.CharVector):
        kx = core.kpn(as_bytes, len(as_bytes))
    else:
        raise _conversion_TypeError(x, 'Python str', ktype)
    return k._factory(<uintptr_t>kx, False)


def from_bytes(x: bytes,
               ktype: Optional[KType] = None,
) -> Union[k.SymbolAtom, k.SymbolVector, k.CharAtom]:
    """Converts a `bytes` object into an instance of a string-like subclass of `pykx.K`.

    Parameters:
        x: The `bytes` object that will be converted into a `pykx.K` object.
        ktype: Desired `pykx.K` subclass (or type number) for the returned value. If `None`, the
            type is inferred from `x`. The following values are supported:


            +-------------------+---------------------------------------------------------+
            | `ktype`           | Returned value                                          |
            +===================+=========================================================+
            | `None`            | Same as for `ktype=pykx.CharVector` when `len(x) != 1`. |
            |                   | Same as for `ktype=pykx.CharAtom` when `len(x) == 1`.   |
            +-------------------+---------------------------------------------------------+
            | `pykx.CharVector` | The `str` as a q character vector.                      |
            +-------------------+---------------------------------------------------------+
            | `pykx.SymbolAtom` | The `str` as a q symbol.                                |
            +-------------------+---------------------------------------------------------+
            | `pykx.CharAtom`   | The `str` as a q character atom.                        |
            +-------------------+---------------------------------------------------------+

    Raises:
        TypeError: Unsupported `ktype` for `bytes`.
        ValueError: `ktype=pykx.CharAtom`, but `x` is not 1 character long.

    Returns:
        An instance of a `pykx.CharVector`, `pykx.SymbolAtom`, or `pykx.CharAtom`.
    """
    cdef core.K kx
    if (ktype is None and len(x) == 1) or (ktype is not None and issubclass(ktype, k.CharAtom)):
        if len(x) != 1:
            raise ValueError(
                f"'pykx.CharAtom' can only be created from a single character, not {x!r}"
            )
        kx = core.kc(ord(x) if x else 32)
    elif ktype is None or issubclass(ktype, k.CharVector):
        kx = core.kpn(x, len(x))
    elif issubclass(ktype, k.SymbolAtom):
        kx = core.ks(x)
    else:
        raise _conversion_TypeError(x, 'Python bytes', ktype)
    return k._factory(<uintptr_t>kx, False)


def from_uuid_UUID(x: UUID, ktype: Optional[KType] = None) -> k.GUIDAtom:
    """Converts a `uuid.UUID` into a `pykx.GUIDAtom`.

    Parameters:
        x: The `uuid.UUID` that will be converted into a `pykx.GUIDAtom`.
        ktype: Desired `pykx.K` subclass (or type number) for the returned value. If `None`, the
            type is inferred from `x`. The following values are supported:

            `ktype`         | Returned value
            --------------- | ----------------------------------
            `None`          | Same as for `ktype=pykx.GUIDAtom`.
            `pykx.GUIDAtom` | The `uuid.UUID` as a q GUID atom.

    Raises:
        TypeError: Unsupported `ktype` for `uuid.UUID`.

    Returns:
        An instance of `pykx.GUIDAtom`.
    """
    cdef core.K kx
    if ktype is not None and not issubclass(ktype, k.GUIDAtom):
        raise _conversion_TypeError(x, repr('uuid.UUID'), ktype)

    u = x.int
    u = (u & 0x0000000000000000FFFFFFFFFFFFFFFF) << 64 | (u & 0xFFFFFFFFFFFFFFFF0000000000000000) >> 64 # noqa
    u = (u & 0x00000000FFFFFFFF00000000FFFFFFFF) << 32 | (u & 0xFFFFFFFF00000000FFFFFFFF00000000) >> 32 # noqa
    u = (u & 0x0000FFFF0000FFFF0000FFFF0000FFFF) << 16 | (u & 0xFFFF0000FFFF0000FFFF0000FFFF0000) >> 16 # noqa
    u = (u & 0x00FF00FF00FF00FF00FF00FF00FF00FF) << 8  | (u & 0xFF00FF00FF00FF00FF00FF00FF00FF00) >> 8   # noqa
    cdef uint64_t upper_bits = (u & (-1 ^ 0xFFFFFFFFFFFFFFFF)) >> 64
    cdef uint64_t lower_bits = u & 0xFFFFFFFFFFFFFFFF
    cdef uint64_t data[2]
    data[0] = lower_bits
    data[1] = upper_bits
    cdef core.U guid
    guid.g = <char*>data
    return k._factory(<uintptr_t>core.ku(guid), False)


# This mapping of k types to Numpy dtypes is for from_list to convert lists of elements of a
# single type to a vector instead of a generic list.
_raze_with_np = {
    k.BooleanVector: bool,
    k.ByteVector: np.uint8,
    k.ShortVector: np.int16,
    k.IntVector: np.int32,
    k.LongVector: np.int64,
    k.RealVector: np.float32,
    k.FloatVector: np.float64,
}


def from_list(x: list, ktype: Optional[KType] = None) -> k.Vector:
    """Converts a `list` into an instance of a subclass of `pykx.Vector`.

    Note: q lists are vectors
        In q, a list (also called a "general list") is a vector of k objects. Non-list vectors
        (also called "typed vectors" or simply "vectors") are vectors of one particular type.

    Parameters:
        x: The `list` that will be converted into a `pykx.Vector`.
        ktype: Desired `pykx.K` subclass (or type number) for the returned value. If `None`, the
            type is inferred from `x`. The following values are supported:

            +----------------------+--------------------------------------------------------------+
            | `ktype`              | Returned value                                               |
            +======================+==============================================================+
            | `None`               | Same as for `ktype=pykx.List`.                               |
            +----------------------+--------------------------------------------------------------+
            | `pykx.List`          | The `list` as a q list. Each element of the Python list will |
            |                      | be converted independently into a `pykx.K` object, and added |
            |                      | to the q list.                                               |
            +----------------------+--------------------------------------------------------------+
            | `pykx.BooleanVector` | The `list` as a `pykx.BooleanVector`. This is equivalent to  |
            |                      | converting a Numpy array `array(x, dtype=bool)` to q.        |
            +----------------------+--------------------------------------------------------------+
            | `pykx.ByteVector`    | The `list` as a `pykx.ByteVector`. This is equivalent to     |
            |                      | converting a Numpy array `array(x, dtype=np.uint8)` to q.    |
            +----------------------+--------------------------------------------------------------+
            | `pykx.ShortVector`   | The `list` as a `pykx.ShortVector`. This is equivalent to    |
            |                      | converting a Numpy array `array(x, dtype=np.int16)` to q.    |
            +----------------------+--------------------------------------------------------------+
            | `pykx.IntVector`     | The `list` as a `pykx.IntVector`. This is equivalent to      |
            |                      | converting a Numpy array `array(x, dtype=np.int32)` to q.    |
            +----------------------+--------------------------------------------------------------+
            | `pykx.LongVector`    | The `list` as a `pykx.LongVector`. This is equivalent to     |
            |                      | converting a Numpy array `array(x, dtype=np.int64)` to q.    |
            +----------------------+--------------------------------------------------------------+
            | `pykx.RealVector`    | The `list` as a `pykx.RealVector`. This is equivalent to     |
            |                      | converting a Numpy array `array(x, dtype=np.float32)` to q.  |
            +----------------------+--------------------------------------------------------------+
            | `pykx.FloatVector`   | The `list` as a `pykx.FloatVector`. This is equivalent to    |
            |                      | converting a Numpy array `array(x, dtype=np.float64)` to q.  |
            +----------------------+--------------------------------------------------------------+

    Raises:
        TypeError: Unsupported `ktype` for `list`.
        ValueError: Could not convert some elements of `x` in accordance with the `ktype`.

    Returns:
        An instance of a subclass of `pykx.Vector`.
    """
    if ktype is not None and not issubclass(ktype, k.List):
        try:
            return from_numpy_ndarray(np.array(x, dtype=_raze_with_np.get(ktype)), ktype)
        except TypeError as ex:
            raise _conversion_TypeError(x, 'Python list', ktype) from ex
    cdef core.K kx = core.ktn(0, len(x))
    for i, item in enumerate(x):
        # No good way to specify the ktype for nested types
        (<core.K*>kx.G0)[i] = core.r1(_k(toq(item)))
    return k._factory(<uintptr_t>kx, False)


def from_tuple(x: tuple, ktype: Optional[KType] = None) -> k.Vector:
    """Converts a `tuple` into an instance of a subclass of `pykx.Vector`.

    See Also:
        [`from_list`][pykx.toq.from_list]

    Parameters:
        x: The `tuple` that will be converted into a `pykx.Vector`.
        ktype: Desired `pykx.K` subclass (or type number) for the returned value. If `None`, the
            type is inferred from `x`. The following values are supported:

            +----------------------+--------------------------------------------------------------+
            | `ktype`              | Returned value                                               |
            +======================+==============================================================+
            | `None`               | Same as for `ktype=pykx.List`.                               |
            +----------------------+--------------------------------------------------------------+
            | `pykx.List`          | The `tuple` as a q list. Each element of the Python list     |
            |                      | will be converted independently into a `pykx.K` object, and  |
            |                      | added to the q list.                                         |
            +----------------------+--------------------------------------------------------------+
            | `pykx.BooleanVector` | The `tuple` as a `pykx.BooleanVector`. This is equivalent to |
            |                      | converting a Numpy array `array(x, dtype=bool)` to q.        |
            +----------------------+--------------------------------------------------------------+
            | `pykx.ByteVector`    | The `tuple` as a `pykx.ByteVector`. This is equivalent to    |
            |                      | converting a Numpy array `array(x, dtype=np.uint8)` to q.    |
            +----------------------+--------------------------------------------------------------+
            | `pykx.ShortVector`   | The `tuple` as a `pykx.ShortVector`. This is equivalent to   |
            |                      | converting a Numpy array `array(x, dtype=np.int16)` to q.    |
            +----------------------+--------------------------------------------------------------+
            | `pykx.IntVector`     | The `tuple` as a `pykx.IntVector`. This is equivalent to     |
            |                      | converting a Numpy array `array(x, dtype=np.int32)` to q.    |
            +----------------------+--------------------------------------------------------------+
            | `pykx.LongVector`    | The `tuple` as a `pykx.LongVector`. This is equivalent to    |
            |                      | converting a Numpy array `array(x, dtype=np.int64)` to q.    |
            +----------------------+--------------------------------------------------------------+
            | `pykx.RealVector`    | The `tuple` as a `pykx.RealVector`. This is equivalent to    |
            |                      | converting a Numpy array `array(x, dtype=np.float32)` to q.  |
            +----------------------+--------------------------------------------------------------+
            | `pykx.FloatVector`   | The `tuple` as a `pykx.FloatVector`. This is equivalent to   |
            |                      | converting a Numpy array `array(x, dtype=np.float64)` to q.  |
            +----------------------+--------------------------------------------------------------+

    Raises:
        TypeError: Unsupported `ktype` for `tuple`.
        ValueError: Could not convert some elements of `x` in accordance with the `ktype`.

    Returns:
        An instance of a subclass of `pykx.Vector`.
    """
    if ktype is not None and not issubclass(ktype, k.Vector):
        raise _conversion_TypeError(x, 'Python tuple', ktype)
    return from_list(list(x), ktype=ktype)


def from_dict(x: dict, ktype: Optional[KType] = None) -> k.Dictionary:
    """Converts a `dict` into a `pykx.Dictionary`.

    Parameters:
        x: The `dict` that will be converted into a `pykx.Dictionary`.
        ktype: Desired `pykx.K` subclass (or type number) for the returned value. If `None`, the
            type is inferred from `x`. The following values are supported:

            +-------------------+-----------------------------------------------------------------+
            | `ktype`           | Returned value                                                  |
            +===================+=================================================================+
            | `None`            | Same as for `ktype=pykx.Dictionary`.                            |
            +-------------------+-----------------------------------------------------------------+
            | `pykx.Dictionary` | The `dict` as a q dictionary. Each key and value of the Python  |
            |                   |  dictionary will be converted independently into a `pykx.K`     |
            |                   |  object, and added to the q dictionary.                         |
            +-------------------+-----------------------------------------------------------------+

    Raises:
        TypeError: Unsupported `ktype` for `dict`.

    Returns:
        An instance of `pykx.Dictionary`.
    """
    if ktype is not None and ktype != k.Dictionary:
        raise _conversion_TypeError(x, 'Python dict', ktype)
    cdef core.K kx
    if all(isinstance(key, (str, k.SymbolAtom)) for key in x.keys()):
        k_keys = toq(np.array([str(key) for key in x.keys()]))
    else:
        k_keys = toq(list(x.keys()))
    k_values = toq(list(x.values()))
    kx = core.xD(core.r1(_k(k_keys)), core.r1(_k(k_values)))
    return k._factory(<uintptr_t>kx, True)


# keys are supported ndarray k types, values are # of bytes per element
supported_np_temporal_types = {
    k.TimestampVector: 8,
    k.MonthVector: 4,
    k.DateVector: 4,
    k.TimespanVector: 8,
    k.MinuteVector: 4,
    k.SecondVector: 4,
    k.TimeVector: 4,
}


supported_np_nontemporal_types = {
    k.List: 8,
    k.BooleanVector: 1,
    k.GUIDVector: 16,
    k.ByteVector: 1,
    k.ShortVector: 2,
    k.IntVector: 4,
    k.LongVector: 8,
    k.RealVector: 4,
    k.FloatVector: 8,
    k.CharVector: 1,
    k.SymbolVector: 8,
}


supported_ndarray_k_types = {**supported_np_temporal_types, **supported_np_nontemporal_types}


def _listify(x):
    """Convert all arrays except the lowest level into lists."""
    if len(x.shape) > 1:
        return [_listify(y) for y in list(x)]
    return x


def _resolve_ndarray_k_type(x, ktype):
    if ktype is not None:
        return ktype
    if x.dtype.char == 'U' or (x.dtype == object and _first_isinstance(x, str)):
        return k.SymbolVector
    elif x.dtype.char == 'S':
        return k.CharVector
    elif x.dtype == np.float64:
        return k.FloatVector
    elif x.dtype == np.float32:
        return k.RealVector
    elif x.dtype == np.int64:
        return k.LongVector
    elif x.dtype == np.int32:
        return k.IntVector
    elif x.dtype == np.int16:
        return k.ShortVector
    elif x.dtype == 'timedelta64[ms]':
        return k.TimeVector
    elif x.dtype == 'timedelta64[ns]':
        return k.TimespanVector
    elif x.dtype == 'datetime64[ns]':
        return k.TimestampVector
    elif x.dtype == 'timedelta64[s]':
        return k.SecondVector
    elif x.dtype == 'datetime64[M]':
        return k.MonthVector
    elif x.dtype == 'datetime64[D]':
        return k.DateVector
    elif x.dtype == 'timedelta64[m]':
        return k.MinuteVector
    elif x.dtype == np.uint8:
        return k.ByteVector
    elif (x.dtype == object and _first_isinstance(x, UUID)) or x.dtype == complex:
        return k.GUIDVector
    elif x.dtype == bool:
        return k.BooleanVector
    elif x.dtype == object or len(x) == 0:
        return k.List
    else:
        raise TypeError(f'ktype cannot be inferred from Numpy dtype {x.dtype}')


def from_numpy_ndarray(x, ktype: Optional[KType] = None) -> k.Vector:
    """Converts a `numpy.ndarray` into a `pykx.Vector`.

    Warning: Data is always copied when converting from Numpy to q.
        While many conversions from q to Python types can be performed without incurring a copy,
        the reverse is not true. As mentioned in
        [the q C API documentation](https://code.kx.com/q/interfaces/capiref/#k-object), vectors in
        q store their metadata in a fixed memory location relative to their data, so the data for
        an array in Python's memory cannot be shared by q, since q would have to store the metadata
        in a memory location it has no ownership of.

    Because q only supports 1-dimensional arrays (vectors), Numpy arrays with more than 1 dimension
    will have all every level converted into a list of lists except for the lowest, for which each
    vector is converted to a q vector independently.

    For example, in the following case, even though Numpy stores all of the data contiguously in
    memory with a single `numpy.ndarray` object wrapping it, q will store it as 3 vectors, with a
    list containing them:

    ```python
    >>> np.arange(12).reshape(3, 4)
    array([[ 0,  1,  2,  3],
            [ 4,  5,  6,  7],
            [ 8,  9, 10, 11]])
    >>> kx.K(np.arange(12).reshape(3, 4))
    pykx.List(pykx.q('
    0 1 2  3
    4 5 6  7
    8 9 10 11
    '))
    ```

    Numpy masked arrays can be used to indicate null data for integer types.

    Parameters:
        x: The `numpy.ndarray` that will be converted into a `pykx.Vector`.
        ktype: Desired `pykx.K` subclass (or type number) for the returned value. If `None`, the
            type is inferred from `x`. The following values are supported:

            +------------------------+------------------------------------------------------------+
            | `ktype`                | Returned value                                             |
            +========================+============================================================+
            | `None`                 | The returned vector is as if the `ktype` had been set      |
            |                        | based off of the Numpy `dtype`, roughly following this     |
            |                        | mapping:                                                   |
            |                        |                                                            |
            |                        | +----------------------+------------------------+          |
            |                        | | `x.dtype`            | Inferred `ktype`       |          |
            |                        | +======================+========================+          |
            |                        | | `object`             | `pykx.List`            |          |
            |                        | +----------------------+------------------------+          |
            |                        | | `object` (uuid.UUID) | `pykx.GUIDVector`      |          |
            |                        | +----------------------+------------------------+          |
            |                        | | `object` (str)       | `pykx.SymbolVector`    |          |
            |                        | +----------------------+------------------------+          |
            |                        | | `'U'`                | `pykx.SymbolVector`    |          |
            |                        | +----------------------+------------------------+          |
            |                        | | `'S'`                | `pykx.CharVector`      |          |
            |                        | +----------------------+------------------------+          |
            |                        | | `np.float64`         | `pykx.FloatVector`     |          |
            |                        | +----------------------+------------------------+          |
            |                        | | `np.float32`         | `pykx.RealVector`      |          |
            |                        | +----------------------+------------------------+          |
            |                        | | `np.int64`           | `pykx.LongVector`      |          |
            |                        | +----------------------+------------------------+          |
            |                        | | `np.int32`           | `pykx.IntVector`       |          |
            |                        | +----------------------+------------------------+          |
            |                        | | `np.int16`           | `pykx.ShortVector`     |          |
            |                        | +----------------------+------------------------+          |
            |                        | | `'timedelta64[ms]'`  | `pykx.TimeVector`      |          |
            |                        | +----------------------+------------------------+          |
            |                        | | `'timedelta64[ns]'`  | `pykx.TimespanVector`  |          |
            |                        | +----------------------+------------------------+          |
            |                        | | `'datetime64[ns]'`   | `pykx.TimestampVector` |          |
            |                        | +----------------------+------------------------+          |
            |                        | | `'timedelta64[s]'`   | `pykx.SecondVector`    |          |
            |                        | +----------------------+------------------------+          |
            |                        | | `'datetime64[M]'`    | `pykx.MonthVector`     |          |
            |                        | +----------------------+------------------------+          |
            |                        | | `'datetime64[D]'`    | `pykx.DateVector`      |          |
            |                        | +----------------------+------------------------+          |
            |                        | | `'timedelta64[m]'`   | `pykx.MinuteVector`    |          |
            |                        | +----------------------+------------------------+          |
            |                        | | `np.uint8`           | `pykx.ByteVector`      |          |
            |                        | +----------------------+------------------------+          |
            |                        | | `bool`               | `pykx.BooleanVector`   |          |
            |                        | +----------------------+------------------------+          |
            +------------------------+------------------------------------------------------------+
            | `pykx.List`            | The Numpy array as a q general list. Equivalent to         |
            |                        | `pykx.List(x.tolist())`.                                   |
            +------------------------+------------------------------------------------------------+
            | `pykx.BooleanVector`   | The Numpy array as a vector of q booleans, which each take |
            |                        | up 8 bits in memory.                                       |
            +------------------------+------------------------------------------------------------+
            | `pykx.ByteVector`      | The Numpy array as a vector of q unsigned 8 bit intergers. |
            +------------------------+------------------------------------------------------------+
            | `pykx.GUIDVector`      | The Numpy array as a vector of q GUIDs, which each take up |
            |                        | 128 bits in memory.                                        |
            +------------------------+------------------------------------------------------------+
            | `pykx.ShortVector`     | The Numpy array as a vector of q signed 16 bit integers.   |
            +------------------------+------------------------------------------------------------+
            | `pykx.IntVector`       | The Numpy array as a vector of q signed 32 bit integers.   |
            +------------------------+------------------------------------------------------------+
            | `pykx.LongVector`      | The Numpy array as a vector of q signed 64 bit integers.   |
            +------------------------+------------------------------------------------------------+
            | `pykx.RealVector`      | The Numpy array as a vector of q 32 bit floats.            |
            +------------------------+------------------------------------------------------------+
            | `pykx.FloatVector`     | The Numpy array as a vector of q 64 bit floats.            |
            +------------------------+------------------------------------------------------------+
            | `pykx.CharVector`      | The Numpy array as a vector of q characters. Each          |
            |                        | characters in q is a single byte, which has implications   |
            |                        | for multi-byte characters in e.g. Unicode strings.         |
            +------------------------+------------------------------------------------------------+
            | `pykx.SymbolVector`    | The Numpy array as a vector of q symbols. Each symbol is   |
            |                        | permanently interned into q's memory.                      |
            +------------------------+------------------------------------------------------------+
            | `pykx.TimestampVector` | The Numpy array as a vector of q timestamp, i.e. signed 64 |
            |                        | bit integers which represent the number of nanoseconds     |
            |                        | since the q epoch: `2000-01-01T00:00:00.000000000`. The    |
            |                        | data from the Numpy array will be incremented to adjust    |
            |                        | its epoch from the standard epoch                          |
            |                        | (`1970-01-01T00:00:00.000000000`) to the q epoch.          |
            +------------------------+------------------------------------------------------------+
            | `pykx.MonthVector`     | The Numpy array as a vector of q months, i.e. signed 32    |
            |                        | bit intergers which represent the number of months since   |
            |                        | the q epoch: `2000-01`. The data from the Numpy array will |
            |                        | be incremented to adjust its epoch from the standard epoch |
            |                        | (`1970-01`) to the q epoch.                                |
            +------------------------+------------------------------------------------------------+
            | `pykx.DateVector`      | The Numpy array as a vector of q dates, i.e. signed 32 bit |
            |                        | intergers which represent the number of days since the q   |
            |                        | epoch: `2000-01-01`. The data from the Numpy array will be |
            |                        | incremented to adjust its epoch from the standard epoch    |
            |                        | (`1970-01-01`) to the q epoch.                             |
            +------------------------+------------------------------------------------------------+
            | `pykx.TimespanVector`  | The Numpy array as a vector of q timespans, i.e. signed 64 |
            |                        | bit integers which represent a number of nanoseconds.      |
            +------------------------+------------------------------------------------------------+
            | `pykx.MinuteVector`    | The Numpy array as a vector of q minutes, i.e. signed 32   |
            |                        | bit integers which represent a number of minutes.          |
            +------------------------+------------------------------------------------------------+
            | `pykx.SecondVector`    | The Numpy array as a vector of q seconds, i.e. signed 32   |
            |                        | bit integers which represent a number of seconds.          |
            +------------------------+------------------------------------------------------------+
            | `pykx.TimeVector`      | The Numpy array as a vector of q times, i.e. signed 32 bit |
            |                        | integers which represent a number of milliseconds.         |
            +------------------------+------------------------------------------------------------+

    Raises:
        TypeError: Unsupported `ktype` for `numpy.ndarray`.
        ValueError: Numpy array is not C-contiguous, which is required for `ktype`.

    Returns:
        An instance of a subclass of `pykx.Vector`.
    """
    ktype = _resolve_ndarray_k_type(x, ktype)

    if ktype not in supported_ndarray_k_types:
        raise _conversion_TypeError(x, repr('numpy.ndarray'), ktype)

    if isinstance(x, np.ma.MaskedArray):
        if x.dtype.kind != 'i':
            raise TypeError(
                'Non-integral masked array conversions to q are not yet implemented'
            )
        x = np.ma.MaskedArray(x, copy=False, fill_value=-2 ** (x.itemsize * 8 - 1)).filled()

    if issubclass(ktype, k.List):
        return from_list(x.tolist(), ktype=k.List)

    # q doesn't support n-dimensional vectors, so we treat them as lists to preserve the shape
    if len(x.shape) > 1:
        return from_list(_listify(x), ktype=k.List)

    if issubclass(ktype, k.CharVector):
        if str(x.dtype).endswith('U1'):
            return from_bytes(''.join(x).encode())
        elif str(x.dtype).endswith('S1'):
            return from_bytes(b''.join(x))
        raise _conversion_TypeError(x, repr('numpy.ndarray'), ktype)
    elif ktype not in supported_np_temporal_types:
        if hasattr(x.data, 'c_contiguous') and not x.data.c_contiguous:
            raise ValueError('Numpy array must be C-contiguous to convert to k vector')

    cdef core.J n = x.size
    cdef core.K kx = core.ktn(ktype.t, n)
    cdef bytes as_bytes
    cdef uintptr_t data

    if issubclass(ktype, k.GUIDVector) and x.dtype == object:
        for i in range(n):
            (<np.complex128_t*>kx.G0)[i] = k.UUID_to_complex(x[i])
    elif issubclass(ktype, k.SymbolVector):
        x = x.astype('U')
        for i in range(n):
            as_bytes = bytes(x[i], 'utf-8')
            (<char**>kx.G0)[i] = core.sn(as_bytes, len(as_bytes))
    elif ktype in supported_np_temporal_types:
        if issubclass(ktype, k.TimestampVector) or issubclass(ktype, k.TimespanVector):
            offset = TIMESTAMP_OFFSET if issubclass(ktype, k.TimestampVector) else 0
            for i in range(n):
                (<core.J*>kx.G0)[i] = x[i] - offset
        else:
            if issubclass(ktype, k.MonthVector):
                offset = MONTH_OFFSET
            elif issubclass(ktype, k.DateVector):
                offset = DATE_OFFSET
            else:
                offset = 0
            for i in range(n):
                (<int*>kx.G0)[i] = x.astype(int)[i] - offset
    elif ktype in supported_np_nontemporal_types:
        itemsize = supported_ndarray_k_types[ktype]
        if itemsize != x.itemsize:
            core.r0(kx)
            raise TypeError('Item size mismatch when converting Numpy ndarray to q: q item size '
                            f'({itemsize}) != Numpy item size ({x.itemsize})')
        data = x.__array_interface__['data'][0]
        memcpy(<void*>kx.G0, <void*>data, n * itemsize)
    return k._factory(<uintptr_t>kx, False)


def from_pandas_dataframe(x: pd.DataFrame,
                          ktype: Optional[KType] = None
) -> Union[k.Table, k.KeyedTable]:
    """Converts a `pandas.DataFrame` into a `pykx.Table` or `pykx.KeyedTable` as appropriate.

    The index and columns of the dataframe are each converted independently, and then the results
    of those conversions are used to assemble the resulting q table or keyed table.

    See Also:
        - [`from_pandas_index`][pykx.toq.from_pandas_index]
        - [`from_pandas_series`][pykx.toq.from_pandas_series]

    Parameters:
        x: The `pandas.DataFrame` that will be converted into a `pykx.Table` or `pykx.KeyedTable`.
        ktype: Desired `pykx.K` subclass (or type number) for the returned value. If `None`, the
            type is inferred from `x`. The following values are supported:

            +-------------------+-----------------------------------------------------------------+
            | `ktype`           | Returned value                                                  |
            +===================+=================================================================+
            | `None`            | The returned value is as if the `ktype` had been set based off  |
            |                   | of the index of the dataframe. If the index has an integer      |
            |                   | type, starts at 0, and increases by 1 for every row, then a     |
            |                   | `pykx.Table` is returned. Otherwise, a `pykx.KeyedTable` is     |
            |                   | returned. Note that `pykx.KeyedTable` is a subclass of          |
            |                   | `pykx.Dictionary`, not `pykx.Table`. Both are subclasses of     |
            |                   | `pykx.Mapping`.                                                 |
            +-------------------+-----------------------------------------------------------------+
            | `pykx.Table`      | The `pandas.DataFrame` as a q table.                            |
            +-------------------+-----------------------------------------------------------------+
            | `pykx.KeyedTable` | The `pandas.DataFrame` as a q keyed table.                      |
            +-------------------+-----------------------------------------------------------------+

    Raises:
        TypeError: Unsupported `ktype` for `pandas.DataFrame`.

    Returns:
        An instance of `pykx.Table` or `pykx.KeyedTable`.
    """
    cdef core.K kx

    if ktype is None:
        if pd.Index(np.arange(0, len(x))).equals(x.index):
            ktype = k.Table
        else:
            ktype = k.KeyedTable

    if issubclass(ktype, k.Table):
        kx = core.xT(core.r1(_k(from_dict({k: v.to_numpy() for k, v in x.iteritems()}))))
        if kx == NULL:
            raise PyKXException('Failed to create table from k dictionary')
    elif issubclass(ktype, k.KeyedTable):
        k_keys = from_pandas_index(x.index)
        values = [x[c].values for c in x.columns]
        k_values = from_pandas_dataframe(
            df_from_arrays(pd.CategoricalIndex(x.columns), values, pd.RangeIndex(len(x)))
        )
        kx = core.xD(core.r1(_k(k_keys)), core.r1(_k(k_values)))
        if kx == NULL:
            raise PyKXException('Failed to create k dictionary (keyed table)')
    else:
        raise _conversion_TypeError(x, repr('pandas.DataFrame'), ktype)
    return k._factory(<uintptr_t>kx, False)


def from_pandas_series(x: pd.Series, ktype: Optional[KType] = None) -> k.Vector:
    """Converts a `pandas.Series` into an instance of a subclass of `pykx.Vector`.

    The given series is converted to a Numpy array, and then converted to a q vector as Numpy
    arrays generally are.

    See Also:
        [`from_numpy_ndarray`][pykx.toq.from_numpy_ndarray]

    Parameters:
        x: The `pandas.Series` that will be converted into a `pykx.Vector`.
        ktype: Desired `pykx.K` subclass (or type number) for the returned value. If `None`,
            the type is inferred from `x` using the same logic as `from_numpy_ndarray` applied
            to `x.to_numpy()`.

    Raises:
        TypeError: Unsupported `ktype` for `pandas.Series`.

    Returns:
        An instance of a subclass of `pykx.Vector`.
    """
    # Convert a series to q by converting to Numpy, the from Numpy to q; this is zero copy where
    # possible: https://pandas.pydata.org/docs/reference/api/pandas.Series.to_numpy.html
    # By using Numpy, we save ourselves from having to handle the Pandas dtypes separately.
    if x.dtype.kind != 'f' and x.isnull().values.any():
        if x.dtype.kind != 'i':
            raise TypeError(
                'Non-integral masked array conversions to q are not yet implemented'
            )
        arr = x.to_numpy(copy=False, na_value=-2 ** (x.dtype.itemsize * 8 - 1))
    else:
        arr = x.to_numpy(copy=False)
    return toq(arr[0] if (1,) == arr.shape else arr, ktype=ktype)


def from_pandas_index(x: pd.Index, ktype: Optional[KType] = None) -> Union[k.Vector, k.Table]:
    """Converts a `pandas.Index` into a `pykx.Vector` or `pykx.Table` as appropriate.

    See Also:
        [`from_numpy_ndarray`][pykx.toq.from_numpy_ndarray]

    Parameters:
        x: The `pandas.Index` that will be converted into a `pykx.Table` or `pykx.KeyedTable`.
        ktype: Desired `pykx.K` subclass (or type number) for the returned value. If `None`, the
            type is inferred from `x`. The following values are supported:

            +---------+---------------------------------------------------------------------------+
            | `ktype` | Returned value                                                            |
            +=========+===========================================================================+
            | `None`  | The resulting value is determined based on the kind of Pandas index       |
            |         | provided. If `x` is an instance of                                        |
            |         | `pandas.core.indexes.numeric.NumericIndex`, then it will be converted     |
            |         | into a Numpy array, then that will be converted into a `pykx.Vector` via  |
            |         | `from_numpy_ndarray`. If `x` is an instance of `pandas.MultiIndex`, then  |
            |         | each of its levels will be converted into a Numpy array, and assembled    |
            |         | into a `pykx.Table` with column names equal to the names of the levels of |
            |         | the index if they are named. For any level which is not named, its index  |
            |         | within `x.levels` is used as its name.                                    |
            +---------+---------------------------------------------------------------------------+

    Raises:
        TypeError: Unsupported `ktype` for `pandas.Index`.

    Returns:
        An instance of `pykx.Vector` or `pykx.Table`.
    """
    if isinstance(x, pd.core.indexes.numeric.NumericIndex):
        # Handles DatetimeIndex/TimedeltaIndex in addition to all of the regular int/float indexes
        return from_numpy_ndarray(x.to_numpy())
    elif isinstance(x, pd.MultiIndex):
        d = {}
        for i, level in enumerate(x.levels):
            d[level.name if level.name is not None else str(i)] = level[x.codes[i]].to_numpy()
        index_dict = from_dict(d)
        return k._factory(<uintptr_t>core.xT(_k(index_dict)), False)
    raise _conversion_TypeError(x, 'Pandas index', ktype)


def from_arrow(x: Union['pa.Array', 'pa.Table'],
               ktype: Optional[KType] = None,
) -> Union[k.Vector, k.Table]:
    """Converts PyArrow arrays/tables into PyKX vectors/tables, respectively.

    Conversions from PyArrow to q are performed by converting the PyArrow array/table to pandas
    first, which avoids copying data when possible, then converting the resulting Pandas
    datastructure to q using `from_pandas_series` or `from_pandas_dataframe` as appropriate.

    See Also:
        - [`from_pandas_dataframe`][pykx.toq.from_pandas_dataframe]
        - [`from_pandas_series`][pykx.toq.from_pandas_series]

    Warning: Memory usage may spike during this function.
        Normally converting from Python to q results in one copy of the data being made, but
        because this function first converts to Pandas, it's possible for there to be a temporary
        third copy made. `See the PyArrow docs about converting to Pandas for more details
        <https://arrow.apache.org/docs/python/pandas.html#memory-usage-and-zero-copy>`_.

    Parameters:
        x: The `pyarrow.Array`  that will be converted into a `pykx.Vector`, or the
            `pyarrow.Table` that will be converted into a `pykx.Table`.
        ktype: Desired `pykx.K` subclass (or type number) for the returned value. If `None`,
            the type is inferred from `x`. This argument is propagated to the Pandas conversion
            functions (which in turn propagate it to the Numpy conversion functions).

    Raises:
        TypeError: Cannot convert PyArrow extension array to `ktype`.

    Returns:
        An instance of `pykx.Vector` or `pykx.Table`.
    """
    # To convert Arrow datastructures to q, we first convert it to Pandas. This conversion avoids
    # copies where possible, but often results in some amount of data being copied.
    # https://arrow.apache.org/docs/python/pandas.html#memory-usage-and-zero-copy
    if pa is None:
        raise PyArrowUnavailable
    if isinstance(x, pa.ExtensionArray):
        raise _conversion_TypeError(x, 'Arrow extension array', ktype)
    return toq(x.to_pandas(), ktype=ktype)


def from_datetime_datetime(x: datetime.datetime,
                           ktype: Optional[KType] = None,
) -> k.TemporalFixedAtom:
    """Converts a `datetime.datetime` into an instance of a subclass of `pykx.TemporalFixedAtom`.

    Parameters:
        x: The `datetime.datetime` that will be converted into a `pykx.TemporalFixedAtom`.
        ktype: Desired `pykx.K` subclass (or type number) for the returned value. If `None`, the
            type is inferred from `x`. The following values are supported:

            +----------------------+--------------------------------------------------------------+
            | `ktype`              | Returned value                                               |
            +======================+==============================================================+
            | `None`               | Same as for `ktype=pykx.TimestampAtom`.                      |
            +----------------------+--------------------------------------------------------------+
            | `pykx.TimestampAtom` | The `datetime.datetime` as a q timestamp atom, which is a    |
            |                      | signed 64 bit value representing the number of nanoseconds   |
            |                      | since the q epoch: `2000-01-01T00:00:00.000000000`.          |
            +----------------------+--------------------------------------------------------------+
            | `pykx.MonthAtom`     | The `datetime.datetime` as a q month atom, which is a signed |
            |                      | 32 bit value representing the number of months since the q   |
            |                      | epoch: `2000-01`.                                            |
            +----------------------+--------------------------------------------------------------+
            | `pykx.DateAtom`      | The `datetime.datetime` as a q date atom, which is a signed  |
            |                      | 32 bit value representing the number of days since the q     |
            |                      | epoch: `2000-01-01`.                                         |
            +----------------------+--------------------------------------------------------------+

    Raises:
        TypeError: Unsupported `ktype` for `datetime.datetime`.
        NotImplementedError: The q datetime type is deprecated, so support for converting to it
            from Python has not been implemented.

    Returns:
        An instance of a subclass of `pykx.TemporalFixedAtom`.
    """
    cdef core.K kx
    epoch = datetime.datetime(2000, 1, 1)
    if ktype is None or issubclass(ktype, k.TimestampAtom):
        d = x - epoch
        t = 1000 * (d.microseconds + 1000000 * (d.seconds + d.days * 86400))
        kx = core.ktj(-12, t)
    elif issubclass(ktype, k.MonthAtom):
        kx = core.ki(12 * (x.year - epoch.year) + x.month - epoch.month)
        kx.t = -13
    elif issubclass(ktype, k.DateAtom):
        kx = core.kd((x - epoch).days)
    elif issubclass(ktype, k.DatetimeAtom):
        raise NotImplementedError
    else:
        raise _conversion_TypeError(x, repr('datetime.datetime'), ktype)
    return k._factory(<uintptr_t>kx, False)


def from_datetime_timedelta(x: datetime.timedelta,
                            ktype: Optional[KType] = None,
) -> k.TemporalSpanAtom:
    """Converts a `datetime.timedelta` into an instance of a subclass of `pykx.TemporalSpanAtom`.

    Parameters:
        x: The `datetime.timedelta` that will be converted into a `pykx.TemporalSpanAtom`.
        ktype: Desired `pykx.K` subclass (or type number) for the returned value. If `None`, the
            type is inferred from `x`. The following values are supported:

            +---------------------+---------------------------------------------------------------+
            | `ktype`             | Returned value                                                |
            +=====================+===============================================================+
            | `None`              | Same as for `ktype=pykx.TimespanAtom`.                        |
            +---------------------+---------------------------------------------------------------+
            | `pykx.TimespanAtom` | The `datetime.timedelta` as a q timespan atom, which is a     |
            |                     | signed 64 bit value representing a number of nanoseconds.     |
            +---------------------+---------------------------------------------------------------+
            | `pykx.MinuteAtom`   | The `datetime.timedelta` as a q minute atom, which is a       |
            |                     | signed 32 bit value representing a number of minutes.         |
            +---------------------+---------------------------------------------------------------+
            | `pykx.SecondAtom`   | The `datetime.timedelta` as a q second atom, which is a       |
            |                     | signed 32 bit value representing a number of seconds.         |
            +---------------------+---------------------------------------------------------------+
            | `pykx.TimeAtom`     | The `datetime.timedelta` as a q date atom, which is a signed  |
            |                     | 32 bit value representing a number of milliseconds.           |
            +---------------------+---------------------------------------------------------------+

    Raises:
        TypeError: Unsupported `ktype` for `datetime.timedelta`.

    Returns:
        An instance of a subclass of `pykx.TemporalSpanAtom`.
    """
    cdef core.K kx
    if ktype is None or issubclass(ktype, k.TimespanAtom):
        t = 1000 * (x.microseconds + 1000000 * (x.seconds + x.days * 86400))
        kx = core.ktj(-16, t)
    elif issubclass(ktype, k.MinuteAtom):
        kx = core.ki(x.total_seconds() // 60)
        kx.t = -17
    elif issubclass(ktype, k.SecondAtom):
        kx = core.ki(x.total_seconds())
        kx.t = -18
    elif issubclass(ktype, k.TimeAtom):
        kx = core.kt(x.days * 86400000 + x.seconds * 1000 + x.microseconds // 1000)
    else:
        raise _conversion_TypeError(x, repr('datetime.timedelta'), ktype)
    return k._factory(<uintptr_t>kx, False)


def from_numpy_datetime64(x: np.datetime64,
                          ktype: Optional[KType] = None,
) -> k.TemporalFixedAtom:
    """Converts a `numpy.datetime64` into an instance of a subclass of `pykx.TemporalFixedAtom`.

    Parameters:
        x: The `numpy.datetime64` that will be converted into a `pykx.TemporalFixedAtom`.
        ktype: Desired `pykx.K` subclass (or type number) for the returned value. If `None`, the
            type is inferred from `x`. The following values are supported:

            +----------------------+--------------------------------------------------------------+
            | `ktype`              | Returned value                                               |
            +======================+==============================================================+
            | `None`               | Same as for `ktype=pykx.TimestampAtom`.                      |
            +----------------------+--------------------------------------------------------------+
            | `pykx.TimestampAtom` | The `numpy.datetime64` as a q timestamp atom, which is a     |
            |                      | signed 64 bit value representing the number of nanoseconds   |
            |                      | since the q epoch: `2000-01-01T00:00:00.000000000`.          |
            +----------------------+--------------------------------------------------------------+
            | `pykx.MonthAtom`     | The `numpy.datetime64` as a q month atom, which is a signed  |
            |                      | 32 bit value representing the number of months since the q   |
            |                      | epoch: `2000-01`.                                            |
            +----------------------+--------------------------------------------------------------+
            | `pykx.DateAtom`      | The `numpy.datetime64` as a q date atom, which is a signed   |
            |                      | 32 bit value representing the number of days since the q     |
            |                      | epoch: `2000-01-01`.                                         |
            +----------------------+--------------------------------------------------------------+

    Raises:
        TypeError: Unsupported `ktype` for `numpy.datetime64`.
        NotImplementedError: The q datetime type is deprecated, so support for converting to it
            from Python has not been implemented.

    Returns:
        An instance of a subclass of `pykx.TemporalFixedAtom`.
    """
    cdef core.K kx
    if ktype is None or issubclass(ktype, k.TimestampAtom):
        kx = core.ktj(-12, x.astype(np.dtype('datetime64[ns]')).astype(np.int64) - TIMESTAMP_OFFSET) # noqa
    elif issubclass(ktype, k.MonthAtom):
        kx = core.ki(x.astype(np.dtype('datetime64[M]')).astype(int) - MONTH_OFFSET)
        kx.t = -13
    elif issubclass(ktype, k.DateAtom):
        kx = core.kd(x.astype(np.dtype('datetime64[D]')).astype(int) - DATE_OFFSET)
    elif issubclass(ktype, k.DatetimeAtom):
        raise NotImplementedError
    else:
        raise _conversion_TypeError(x, repr('numpy.datetime64'), ktype)
    return k._factory(<uintptr_t>kx, False)


def from_numpy_timedelta64(x: np.timedelta64,
                           ktype: Optional[KType] = None,
) -> k.TemporalSpanAtom:
    """Converts a `numpy.timedelta64` into an instance of a subclass of `pykx.TemporalSpanAtom`.

    Parameters:
        x: The `numpy.timedelta64` that will be converted into a `pykx.TemporalSpanAtom`.
        ktype: Desired `pykx.K` subclass (or type number) for the returned value. If `None`, the
            type is inferred from `x`. The following values are supported:

            +---------------------+---------------------------------------------------------------+
            | `ktype`             | Returned value                                                |
            +=====================+===============================================================+
            | `None`              | Same as for `ktype=pykx.TimespanAtom`.                        |
            +---------------------+---------------------------------------------------------------+
            | `pykx.TimespanAtom` | The `numpy.timedelta64` as a q timespan atom, which is a      |
            |                     | signed 64 bit value representing a number of nanoseconds.     |
            +---------------------+---------------------------------------------------------------+
            | `pykx.MinuteAtom`   | The `numpy.timedelta64` as a q minute atom, which is a signed |
            |                     | 32 bit value representing a number of minutes.                |
            +---------------------+---------------------------------------------------------------+
            | `pykx.SecondAtom`   | The `numpy.timedelta64` as a q second atom, which is a signed |
            |                     | 32 bit value representing a number of seconds.                |
            +---------------------+---------------------------------------------------------------+
            | `pykx.TimeAtom`     | The `numpy.timedelta64` as a q time atom, which is a signed   |
            |                     | 32 bit value representing a number of milliseconds.           |
            +---------------------+---------------------------------------------------------------+

    Raises:
        TypeError: Unsupported `ktype` for `numpy.timedelta64`.

    Returns:
        An instance of a subclass of `pykx.TemporalSpanAtom`.
    """
    cdef core.K kx
    if ktype is None or issubclass(ktype, k.TimespanAtom):
        kx = core.ktj(-16, x.astype(np.dtype('timedelta64[ns]')).astype(np.int64))
    elif issubclass(ktype, k.MinuteAtom):
        kx = core.ki(x.astype(np.dtype('timedelta64[m]')).astype(int))
        kx.t = -17
    elif issubclass(ktype, k.SecondAtom):
        kx = core.ki(x.astype(np.dtype('timedelta64[s]')).astype(int))
        kx.t = -18
    elif issubclass(ktype, k.TimeAtom):
        kx = core.kt(x.astype(np.dtype('timedelta64[ms]')).astype(int))
    else:
        raise _conversion_TypeError(x, repr('numpy.timedelta64'), ktype)
    return k._factory(<uintptr_t>kx, False)


def from_slice(x: slice, ktype: Optional[KType] = None) -> k.IntegralNumericVector:
    """Converts a `slice` into an instance of a subclass of `pykx.IntegralNumericVector`.

    Slice objects are used by Python for indexing. In q, indexing is done by applying a numeric
    atom or vector to the object being indexed. As such, `from_slice` works by converting the
    given `slice` into a `pykx.IntegralNumericVector`, which can then be applied to an
    indexable q object.

    Parameters:
        x: The `slice` that will be converted into a `pykx.IntegralNumericVector`.
        ktype: Desired `pykx.K` subclass (or type number) for the returned value. If `None`, the
            type is inferred from `x`. The following values are supported:

            `ktype`              | Returned value
            -------------------- | -----------------------------------------------------
            `None`               | Same as for `ktype=pykx.LongVector`.
            `pykx.LongVector`    | The `slice` as a q vector of signed 64 bit integers.
            `pykx.IntVector`     | The `slice` as a q vector of signed 32 bit integers.
            `pykx.IntVector`     | The `slice` as a q vector of signed 16 bit integers.
            `pykx.ByteVector`    | The `slice` as a q vector of unsigned 8 bit integers.
            `pykx.BooleanVector` | The `slice` as a q vector of unsigned 1 bit integers.

    Raises:
        TypeError: Unsupported `ktype` for `slice`.
        OverflowError: The minimum or maximum value in the slice is out of bounds for the selected
            `ktype`.

    Returns:
        An instance of a subclass of `pykx.IntegralNumericVector`.
    """
    if x.stop is None:
        raise ValueError(f'Cannot convert endless slice {x!r} to vector')
    return from_range(slice_to_range(x, x.stop), ktype=ktype)


_range_k_type_map = {
    k.LongVector: ('long', np.int64, (-9223372036854775808, 9223372036854775807)),
    k.IntVector: ('int', np.int32, (-2147483648, 2147483647)),
    k.ShortVector: ('short', np.int16, (-32768, 32767)),
    k.ByteVector: ('byte', np.int8, (0, 255)),
    k.BooleanVector: ('boolean', bool, (0, 1)),
}


def from_range(x: range, ktype: Optional[KType] = None) -> k.IntegralNumericVector:
    """Converts a `range` into an instance of a subclass of `pykx.IntegralNumericVector`.

    Parameters:
        x: The `range` that will be converted into a `pykx.IntegralNumericVector`.
        ktype: Desired `pykx.K` subclass (or type number) for the returned value. If `None`, the
            type is inferred from `x`. The following values are supported:

            `ktype`              | Returned value
            -------------------- | -----------------------------------------------------
            `None`               | Same as for `ktype=pykx.LongVector`.
            `pykx.LongVector`    | The `range` as a q vector of signed 64 bit integers.
            `pykx.IntVector`     | The `range` as a q vector of signed 32 bit integers.
            `pykx.IntVector`     | The `range` as a q vector of signed 16 bit integers.
            `pykx.ByteVector`    | The `range` as a q vector of unsigned 8 bit integers.
            `pykx.BooleanVector` | The `range` as a q vector of unsigned 1 bit integers.

    Raises:
        TypeError: Unsupported `ktype` for `range`.
        OverflowError: The minimum or maximum value in the range is out of bounds for the selected
            `ktype`.

    Returns:
        An instance of a subclass of `pykx.IntegralNumericVector`.
    """
    if ktype is None:
        ktype = k.LongVector
    if not issubclass(ktype, k.IntegralNumericVector):
        raise _conversion_TypeError(x, 'Python range', ktype)
    name, dtype, (lower_bound, upper_bound) = _range_k_type_map[ktype]
    if lower_bound - 1 in x or upper_bound + 1 in x:
        raise _overflow_error_by_size[name]
    return from_numpy_ndarray(np.arange(x.start, x.stop, x.step, dtype=dtype), ktype=ktype)


def from_pathlib_path(x: Path, ktype: Optional[KType] = None) -> k.SymbolAtom:
    """Converts a `pathlib.Path` into a q handle symbol.

    In q a symbol atom that begin with a `:` is known as an "hsym", or handle symbol. These
    symbols represent local or remote paths or resources.
    
    Windows paths are reformatted as POSIX paths.

    This function does not make the path absolute.

    Parameters:
        x: The `pathlib.Path` that will be converted into an hsym.
        ktype: Desired `pykx.K` subclass (or type number) for the returned value. If `None`, the
            type is inferred from `x`. The following values are supported:

            +-------------------+-----------------------------------------------------------------+
            | `ktype`           | Returned value                                                  |
            +===================+=================================================================+
            | `None`            | Same as for `ktype=pykx.SymbolAtom`.                            |
            +-------------------+-----------------------------------------------------------------+
            | `pykx.SymbolAtom` | The `pathlib.Path` as a q handle symbol, which will be the path |
            |                   | as text, prefixed by `:`, as a POSIX path.                      |
            +-------------------+-----------------------------------------------------------------+

    Raises:
        TypeError: Unsupported `ktype` for `pathlib.Path`.

    Returns:
        An instance of a subclass of `pykx.IntegralNumericVector`.
    """
    return from_str(f'{"" if str(x)[:1] == ":" else ":"}{x.as_posix()}')


def from_ellipsis(x: Ellipsis, ktype: Optional[KType] = None) -> pykx.ProjectionNull:
    """Converts an `Ellipsis` (`...`) into a q projection null.

    Warning: Projection nulls are unwieldy.
        If you aren't sure you know what you are doing, then you probably don't need to use a
        projection null.

    PyKX uses Python's `...` singleton to represent the q projection null, which is similar to
    generic null (see [`from_none`][pykx.toq.from_none]), but indicates to q not just that there
    is missing data, but that it should be filled in with the next q object that is applied onto
    it.

    This is how q creates function projections with later parameters specified. For instance,
    the q function projection `{x+y}[;7]` has an argument list of 2 q objects: a projection null,
    followed by a long atom with a value of 7.

    Because PyKX treats `...` as projection null, q projections can be created in Python like so:

    ```python
    >>> f = pykx.K(lambda x, y: x + y)
    >>> projection = f(..., 7)
    >>> projection(2)
    pykx.LongAtom(pykx.q('9'))
    ```

    Parameters:
        x: The `Ellipsis` singleton object.
        ktype: Desired `pykx.K` subclass (or type number) for the returned value. If `None`, the
            type is inferred from `x`. The following values are supported:

            `ktype`               | Returned value
            --------------------- | ----------------------------------------
            `None`                | Same as for `ktype=pykx.ProjectionNull`.
            `pykx.ProjectionNull` | An instance of `pykx.ProjectionNull`.

    Raises:
        TypeError: Unsupported `ktype` for `Ellipsis`.

    Returns:
        An instance of a subclass of `pykx.ProjectionNull`.
    """
    if ktype is not None and not issubclass(ktype, k.ProjectionNull):
        raise _conversion_TypeError(x, 'Ellipsis', ktype)
    return q('value[(;)]1')


# TODO: After Python 3.7 support is dropped, use a `typing.Protocol` for `x`
def from_fileno(x, ktype: Optional[KType] = None) -> k.IntAtom:
    """Converts an object with a `fileno` attribute to a `pykx.IntAtom`.

    In q, int atoms that match open file descriptors can be called as if they were functions. Refer
    to https://code.kx.com/q/basics/handles/ for more information.

    The `fileno` attribute of `x` can either be a Python `int`, or a function that returns
    an `int` when called. The `int` should represent an open file descriptor.

    Parameters:
        x: The `Ellipsis` singleton object.
        ktype: Desired `pykx.K` subclass (or type number) for the returned value. If `None`, the
            type is inferred from `x`. The following values are supported:

            `ktype`        | Returned value
            -------------- | ---------------------------------------------
            `None`         | Same as for `ktype=pykx.IntAtom`.
            `pykx.IntAtom` | The open file descriptor as a `pykx.IntAtom`.

    Raises:
        TypeError: Unsupported `ktype` for an object with the `fileno` attribute.

    Returns:
        A `pykx.IntAtom` that represents an open file descriptor.
    """
    if ktype is not None and not issubclass(ktype, k.IntAtom):
        raise _conversion_TypeError(x, "object with 'fileno' attribute", ktype)
    fd = x.fileno() if callable(x.fileno) else x.fileno
    if system == 'Windows':
        try:
            fd = msvcrt.get_osfhandle(fd)
        except OSError:
            pass
    return from_int(fd, ktype=k.IntAtom)


converted_callables = {}


def from_callable(x: Callable, ktype: Optional[KType] = None) -> k.Lambda:
    """Converts a callable object into a q lambda function.

    The resulting `pykx.Lambda` object works by keeping a reference to the provided callable
    object. When the lambda is called, it calls the Python callable with the arguments it received
    as `pykx.K` objects. Therefore, Python functions that are converted to q functions must be
    able to handle receiving `pykx.K` objects as arguments.

    The return value of the provided Python callable will automatically be converted to a
    `pykx.K` object by having `pykx.K` called on it. You can control how this conversion
    happens by manually ensuring the return value of the function is a `pykx.K` instance.

    Because q does not have keyword arguments, all of the parameters of the Python callable are
    treated as position parameters. `*args` and `**kwargs` are likewise treated as individual
    parameters, which can be provided a `pykx.Vector` and a `pykx.Mapping` respectively.

    Functions in q can have at most 8 parameters, so Python callables being converted to q must
    abide by this limit. Parameters which have defaults set still count towards this limit. An
    `*args` parameters and a `**kwargs` each count as a single parameter towards this limit.

    Note: Inspection of the callable must be possible.
        In order to determine how many parameters a provided callable has, PyKX calls
        `inspect.signature` on it. Some callables may not be introspectable in certain
        implementations of Python. For example, in CPython, some built-in functions defined in C
        provide no metadata about their arguments.

    Parameters:
        x: A callable Python object that can take `pykx.K` as arguments, and returns an object
            that can be converted into a q object via `pykx.K`.
        ktype: Desired `pykx.K` subclass (or type number) for the returned value. If `None`, the
            type is inferred from `x`. The following values are supported:

            `ktype`       | Returned value
            ------------- | ---------------------------------------------------
            `None`        | Same as for `ktype=pykx.Lambda`.
            `pykx.Lambda` | A q lambda that wraps the provided callable object.

    Raises:
        TypeError: Unsupported `ktype` for a callable object.
        ValueError: `x` accepts too many parameters - no more than 8 can be accepted.
        ValueError: `inspect.signature(x)` failed - could not introspect `x`.

    Returns:
        A `pykx.Lambda` that calls the provided Python callable object when it is calls.
    """
    if not licensed:
        raise LicenseException('convert a Python callable to q')
    uuid = random_uuid()
    params = list(signature(x).parameters.values())
    if len(params) > 8:
        raise ValueError('Too many parameters - q functions cannot have more than 8 parameters')
    converted_callables[uuid] = (x, params)
    # Need to prepend all of the params with a common prefix to avoid overwriting a reserved word
    # in q. This prefix is used in the code, but is stripped out by the `pykx.Function` wrappers.
    param_code = ';'.join('PyKXParam' + p.name for p in params)
    # Build a string that when executed by q will create a lambda with the proper prameter names.
    # When the lambda is run it will call .pykx.i.pyfunc with the uuid for this callable, and a
    # list of arguments. The arguments list starts with a ::, and if :: is the only element it is
    # enlisted. This ensures the argument list always has type 0, as otherwise the arguments might
    # not be individual K objects.
    func_code = f'{{[{param_code}] .pykx.i.pyfunc["{uuid}";' \
        f'{"enlist" if len(param_code) == 0 else ""}' \
        f'(::;{param_code})]}}'
    return q(func_code)


class ToqModule(ModuleType):
    def __call__(self, x: Any, ktype: Optional[KType] = None) -> k.K:
        ktype = _resolve_k_type(ktype)
        if x is None:
            converter = from_none
        elif isinstance(x, k.K):
            converter = from_k
        elif isinstance(x, (int, *np_int_types)):
            if isinstance(x, (bool, np.bool_)) and ktype is None:
                ktype = k.BooleanAtom
            converter = from_int
        elif isinstance(x, (float, *np_float_types)):
            converter = from_float
        elif isinstance(x, str):
            converter = from_str
        elif isinstance(x, bytes):
            converter = from_bytes
        elif isinstance(x, UUID):
            converter = from_uuid_UUID
        elif isinstance(x, list):
            converter = from_list
        elif isinstance(x, tuple):
            converter = from_tuple
        elif isinstance(x, dict):
            converter = from_dict
        elif isinstance(x, np.ndarray):
            converter = from_numpy_ndarray
        elif isinstance(x, pd.DataFrame):
            converter = from_pandas_dataframe
        elif isinstance(x, pd.Series):
            converter = from_pandas_series
        elif isinstance(x, pd.Index):
            converter = from_pandas_index
        elif pa is not None and (isinstance(x, pa.Table) or isinstance(x, pa.Array)):
            converter = from_arrow
        elif isinstance(x, datetime.datetime):
            converter = from_datetime_datetime
        elif isinstance(x, datetime.timedelta):
            converter = from_datetime_timedelta
        elif isinstance(x, np.datetime64):
            converter = from_numpy_datetime64
        elif isinstance(x, np.timedelta64):
            converter = from_numpy_timedelta64
        elif isinstance(x, slice):
            converter = from_slice
        elif isinstance(x, range):
            converter = from_range
        elif isinstance(x, Path):
            converter = from_pathlib_path
        elif x is Ellipsis:
            converter = from_ellipsis
        elif hasattr(x, 'fileno'):
            converter = from_fileno
        elif callable(x): # this check should come after all isinstance checks
            converter = from_callable
        else:
            converter = _default_converter
        return converter(x, ktype)


# Set the module type for this module to `ToqModule` so that it can be called via `__call__`.
toq = sys.modules[__name__]
toq.__class__ = ToqModule

# HACK: inject in the definition to avoid a circular import error
k.toq = toq
