"""Wrapper for q datastructures, with conversion functions to Python/Numpy/Pandas/Arrow.

```mermaid
graph LR
  K --> Atom
  K --> Collection
  Atom --> GUIDAtom
  Atom --> NumericAtom
  NumericAtom --> IntegralNumericAtom
  IntegralNumericAtom --> BooleanAtom
  IntegralNumericAtom --> ByteAtom
  IntegralNumericAtom --> ShortAtom
  IntegralNumericAtom --> IntAtom
  IntegralNumericAtom --> LongAtom
  NumericAtom --> NonIntegralNumericAtom
  NonIntegralNumericAtom --> RealAtom
  NonIntegralNumericAtom --> FloatAtom
  Atom --> CharAtom
  Atom --> SymbolAtom
  Atom --> TemporalAtom
  TemporalAtom --> TemporalFixedAtom
  TemporalFixedAtom --> TimestampAtom
  TemporalFixedAtom --> MonthAtom
  TemporalFixedAtom --> DateAtom
  TemporalFixedAtom --> DatetimeAtom
  TemporalAtom --> TemporalSpanAtom
  TemporalSpanAtom --> TimespanAtom
  TemporalSpanAtom --> MinuteAtom
  TemporalSpanAtom --> SecondAtom
  TemporalSpanAtom --> TimeAtom
  Atom --> EnumAtom
  Atom --> Function
  Function --> Lambda
  Function --> UnaryPrimitive
  UnaryPrimitive --> Identity
  Identity --> ProjectionNull
  Function --> Operator
  Function --> Iterator
  Function --> Projection
  Function --> Composition
  Function --> AppliedIterator
  AppliedIterator --> Each
  AppliedIterator --> Over
  AppliedIterator --> Scan
  AppliedIterator --> EachPrior
  AppliedIterator --> EachRight
  AppliedIterator --> EachLeft
  Function --> Foreign
  Collection --> Vector
  Vector --> List
  Vector --> GUIDVector
  Vector --> NumericVector
  NumericVector --> IntegralNumericVector
  IntegralNumericVector --> BooleanVector
  IntegralNumericVector --> ByteVector
  IntegralNumericVector --> ShortVector
  IntegralNumericVector --> IntVector
  IntegralNumericVector --> LongVector
  NumericVector --> NonIntegralNumericVector
  NonIntegralNumericVector --> RealVector
  NonIntegralNumericVector --> FloatVector
  Vector --> CharVector
  Vector --> SymbolVector
  Vector --> TemporalVector
  TemporalVector --> TemporalFixedVector
  TemporalFixedVector --> TimestampVector
  TemporalFixedVector --> MonthVector
  TemporalFixedVector --> DateVector
  TemporalFixedVector --> DatetimeVector
  TemporalVector --> TemporalSpanVector
  TemporalSpanVector --> TimespanVector
  TemporalSpanVector --> MinuteVector
  TemporalSpanVector --> SecondVector
  TemporalSpanVector --> TimeVector
  Vector --> EnumVector
  Collection --> Mapping
  Mapping --> Dictionary
  Dictionary --> KeyedTable
  Mapping --> Table
  Table --> SplayedTable
  SplayedTable --> PartitionedTable
```
"""

from libc.stdint cimport *

from cpython.ref cimport Py_INCREF
cimport cython
cimport numpy as np # Imported ambiguously as "np" by convention for using Numpy with Cython.
cimport numpy as cnp # Imported unambiguously as "cnp" to avoid a Cython issue.

from pykx cimport core

from collections import abc, defaultdict
from datetime import datetime
from datetime import timedelta
from functools import reduce
from inspect import signature
import math
from numbers import Integral, Real
import operator
import os
from platform import python_implementation
import sys
import threading
from typing import Dict, List as _List, Optional, Union
from uuid import UUID
import warnings

from cython import boundscheck
import numpy as np
import pandas as pd

from ._pyarrow import pyarrow as pa
from .core import licensed
from .exceptions import QError, LicenseException, PyArrowUnavailable, PyKXException
from .util import cached_property, df_from_arrays, slice_to_range, subclasses


def _init(_q):
    global q
    q = _q


# nanoseconds between 1970-01-01 and 2000-01-01
DEF TIMESTAMP_OFFSET = 946684800000000000
#      months between 1970-01-01 and 2000-01-01
DEF MONTH_OFFSET = 360
#        days between 1970-01-01 and 2000-01-01
DEF DATE_OFFSET = 10957


# This dictionary holds a Python exception object raised from Python code run under q, if one
# exists. It maps thread IDs to exception objects. Its purpose is to be used in a
# `raise ex from cause` statement. See `pykx.c::k_py_error` for more about how we handle Python
# exceptions raised from under q.
_current_exception = {}


cdef inline core.J _n(x):
    return (<core.K><uintptr_t>x._addr).n


cdef inline core.K _k(x):
    return <core.K><uintptr_t>x._addr


cdef inline np.ndarray k_vec_to_array(object k_vec, int np_type):
    """
    Creates a Numpy array that keeps a reference to the K object until the array's data is
    deallocated.
    """
    cdef np.npy_intp n = _n(k_vec)
    cdef np.ndarray arr = np.PyArray_SimpleNewFromData(1, &n, np_type, <void*>_k(k_vec).G0)
    # Increment the Python ref count of k_vec because PyArray_SetBaseObject will steal a reference
    Py_INCREF(k_vec)
    PyArray_SetBaseObject(arr, k_vec)
    return arr


cdef inline uint64_t byteswap64(uint64_t x):
    x = (x & 0x00000000FFFFFFFF) << 32 | (x & 0xFFFFFFFF00000000) >> 32
    x = (x & 0x0000FFFF0000FFFF) << 16 | (x & 0xFFFF0000FFFF0000) >> 16
    x = (x & 0x00FF00FF00FF00FF) << 8  | (x & 0xFF00FF00FF00FF00) >> 8 # noqa
    return x


cdef object complex_to_UUID(np.complex128_t c):
    return UUID(int=(int((<uint64_t*>&c.real)[0]) << 64) | (<uint64_t*>&c.imag)[0])


cdef complex UUID_to_complex(u: UUID) except *:
    cdef uint64_t upper_bits = (u.int & (-1 ^ 0xFFFFFFFFFFFFFFFF)) >> 64
    cdef uint64_t lower_bits = u.int & 0xFFFFFFFFFFFFFFFF
    upper_bits = byteswap64(upper_bits)
    lower_bits = byteswap64(lower_bits)
    return complex((<double*>&upper_bits)[0], (<double*>&lower_bits)[0])


def _idx_to_k(key, n):
    if isinstance(key, K):
        return key
    if isinstance(key, Integral):
        # replace negative index with equivalent positive index
        if key < 0:
            key = n + key
        if key >= n or key < 0:
            raise IndexError('index out of range')
    elif isinstance(key, slice) and key.stop is None:
        # ensure slices have a stop set
        # TODO: Ensure all indices are positive
        key = slice(key.start,
                    n if (key.step or 1) > 0 else -1,
                    key.step)
    return K(key)


def _rich_convert(x: K, stdlib: bool = True):
    if stdlib:
        return x.py(stdlib=stdlib)

    for converter_name in x._rich_conversion_precedence:
        try:
            return getattr(x, converter_name)()
        except AttributeError:
            pass
    raise PyKXException(f'No converter found for {x!r} - tried {x._rich_conversion_precedence!r}')


# HACK: This gets overwritten by the toq module to avoid a circular import error.
def toq(*args, **kwargs):
    raise NotImplementedError


# A cdef class is used to store the reference in order to guarantee r0 is called
cdef class _K:
    cdef core.K k

    def __cinit__(self, uintptr_t k, bint incref):
        if incref:
            core.r1(<core.K>k)
        self.k = <core.K>k

    def __dealloc__(self):
        core.r0(self.k)


class K:
    _rich_conversion_precedence = ('np', 'pd', 'py')

    def __new__(cls, x):
        return toq(x, ktype=None if cls is K else cls) # TODO: 'strict' and 'cast' flags

    def __init__(self, x): # Signature must match `__new__`
        pass

    @classmethod
    def _from_addr(cls, uintptr_t addr, bint incref=True):
        instance = object.__new__(cls)
        instance._addr = addr
        instance._k = _K(addr, incref)
        instance.__init__(None) # placeholder argument
        return instance

    @property
    def _r(self):
        """The q reference count of the underlying K object - for debugging only!"""
        return _k(self).r # nocov

    def __repr__(self):
        preamble = f'pykx.{type(self).__name__}'
        if not licensed:
            return f'{preamble}._from_addr({hex(self._addr)})'
        data_str = K.__str__(self)
        if '\n' in data_str:
            data_str = f'\n{data_str}\n'
        return f"{preamble}(pykx.q('{data_str}'))"

    def __str__(self):
        if not licensed:
            return repr(self)
        cdef core.K x = core.k(0, <char* const>'.Q.s', core.r1(_k(self)), NULL)
        if x.n == 0:
            s = ''
        else:
            s = np.asarray(<char[:x.n]><char*>x.G0).tobytes().decode()
        core.r0(x)
        if len(s) and s.endswith(os.linesep):
            # Use `len(s) - X` instead of `-X` because wraparound is disabled here
            s = s[:len(s) - len(os.linesep)] # strip trailing newline appened by .Q.s
        return s

    @staticmethod
    def _from_pickle(state: np.ndarray):
        # TODO: Use d9 instead of -9! to make this work without a license.
        if not licensed:
            raise LicenseException('unpickle K objects')
        return q('-9!', state)

    def __reduce__(self):
        # TODO: Use b9 to instead of -8! to make this work without a license.
        if not licensed:
            raise LicenseException('pickle K objects')
        return (type(self)._from_pickle, (q('-8!', self).np(),))

    def _compare(self, other, op_str):
        try:
            r = q(op_str, self, other)
        except Exception as ex:
            ex_str = str(ex)
            if ex_str.startswith('length') or ex_str.startswith('type'):
                return q('0b')
            raise
        else:
            if hasattr(r, '__len__') and len(r) == 0:
                # Handle comparisons of empty objects
                if op_str == '=':
                    return q('~', self, other)
                elif op_str == '{not x=y}':
                    return q('{not x~y}', self, other)
            return r

    def __lt__(self, other):
        return self._compare(other, '<')

    def __le__(self, other):
        return self._compare(other, '<=')

    def __eq__(self, other):
        try:
            return self._compare(other, '=')
        except TypeError:
            return q('0b')

    def __ne__(self, other):
        try:
            return self._compare(other, '{not x=y}')
        except TypeError:
            return q('1b')

    def __gt__(self, other):
        return self._compare(other, '>')

    def __ge__(self, other):
        return self._compare(other, '>=')

    def __add__(self, other):
        return q('+', self, other)

    __radd__ = __add__

    def __sub__(self, other):
        return q('-', self, other)

    def __rsub__(self, other):
        return q('-', other, self)

    def __mul__(self, other):
        return q('*', self, other)

    __rmul__ = __mul__

    def __truediv__(self, other):
        return q('%', self, other)

    # __rtruediv__ only has priority over __truediv__ if the right-hand operand
    # is an instance of a subclass of the left-hand operand's class
    def __rtruediv__(self, other):
        return q('%', other, self)

    def __floordiv__(self, other):
        return q('div', self, other)

    def __rfloordiv__(self, other):
        return q('div', other, self)

    def __mod__(self, other):
        return q('mod', self, other)

    def __rmod__(self, other):
        return q('mod', other, self)

    def __divmod__(self, other):
        return tuple(q('{(x div y;x mod y)}', self, other))

    def __rdivmod__(self, other):
        return tuple(q('{(x div y;x mod y)}', other, self))

    # TODO: Do we want to keep .pykx.modpow? Q lacks an integer pow, and
    # while one would be easy to implement in pure q it wouldn't be efficient,
    # nor would it be able to handle a modulo easily. Our options are:
    # A) Keep modpow, despite the complexity, and potential for errors
    # B) Raise an error when __pow__ is used on integers
    # C) Use xexp, and always get floats back regardless of the input type, and
    #    be unable to take advantage of the modulo
    # D) Convert the object to a numpy array, and call pow on that; note that
    #    this approach breaks conformity with how all the other similar methods
    #    are implemented
    # E) Use xexp, but convert back to whatever type the original object was;
    #    this approach also cannot take advantage of modulo, and would require
    #    special handling for each type (i.e. more code than ideal, but less
    #    than option A)
    def __pow__(self, other, modulo=None):
        return q('.pykx.modpow', self, other, modulo)

    def __rpow__(self, other, modulo=None):
        return q('.pykx.modpow', other, self, modulo)

    def __neg__(self):
        return q('neg', self)

    def __pos__(self):
        return self

    def __abs__(self):
        return q('abs', self)

    def __bool__(self):
        if self.is_null:
            return False
        if self.is_inf:
            return True
        if isinstance(self, TemporalAtom):
            # Temporal atoms are false if their underlying value is 0, or if they are null.
            return bool(self.py(raw=True))
        x = self.py()
        if type(self) is type(x):
            raise TypeError(f"Unable to determine truth value of '{self!r}'")
        return bool(x)

    any = __bool__
    all = __bool__

    @property
    def is_atom(self):
        return True

    @property
    def t(self):
        return _k(self).t

    def py(self, *, raw: bool = False, has_nulls: Optional[bool] = None, stdlib: bool = True):
        return self

    def np(self, *, raw: bool = False, has_nulls: Optional[bool] = None):
        return self.py(raw=raw)

    def pd(self, *, raw: bool = False, has_nulls: Optional[bool] = None):
        return self.np(raw=raw)

    def pa(self, *, raw: bool = False, has_nulls: Optional[bool] = None):
        if pa is None:
            raise PyArrowUnavailable # nocov
        return self.pd(raw=raw)


class Atom(K):
    @property
    def is_null(self) -> bool:
        return q('null', self).py()

    @property
    def is_inf(self) -> bool:
        if self.t in {-1, -2, -4, -10, -11}:
            return False
        try:
            type_char = ' bg xhijefcspmdznuvts'[abs(self.t)]
        except IndexError:
            return False
        return q(f'{{any -0W 0W{type_char}~\:x}}')(self).py()

    def __hash__(self):
        return hash(self.py())

    def __invert__(self):
        return ~self.py()

    def __lshift__(self, other):
        return self.py() << other

    def __rlshift__(self, other):
        return other << self.py()

    def __rshift__(self, other):
        return self.py() >> other

    def __rrshift__(self, other):
        return other >> self.py()

    def __and__(self, other):
        return self.py() & other

    def __rand__(self, other):
        return other & self.py()

    def __or__(self, other):
        return self.py() | other

    def __ror__(self, other):
        return other | self.py()

    def __xor__(self, other):
        return self.py() ^ other

    def __rxor__(self, other):
        return other ^ self.py()


class EnumAtom(Atom):
    t = -20

    def py(self, *, raw: bool = False, has_nulls: Optional[bool] = None, stdlib: bool = True):
        if raw:
            return _k(self).j
        return q('value', self).py()


class TemporalAtom(Atom):
    pass


class TemporalSpanAtom(TemporalAtom):
    def py(self, *, raw: bool = False, has_nulls: Optional[bool] = None, stdlib: bool = True):
        if raw:
            return self.np(raw=True)
        return pd.to_timedelta(self.np(), 'us').to_pytimedelta()

    def np(self,
           raw: bool = False,
           has_nulls: Optional[bool] = None,
    ) -> Union[np.timedelta64, int]:
        if raw:
            if self.t == -16:
                return _k(self).j
            return _k(self).i
        if self.is_null:
            return np.timedelta64('NaT')
        if self.t == -16:
            return np.timedelta64(_k(self).j, self._np_type)
        return np.timedelta64(_k(self).i, self._np_type)

    def pd(self, *, raw: bool = False, has_nulls: Optional[bool] = None) -> Union[pd.Timedelta, int]:
        if raw:
            return self.np(raw=True)
        return pd.Timedelta(self.np())


class TemporalFixedAtom(TemporalAtom):
    def py(self, *, raw: bool = False, has_nulls: Optional[bool] = None, stdlib: bool = True):
        if raw:
            return self.np(raw=True)
        return self.np().astype(datetime)

    def np(self, *, raw: bool = False, has_nulls: Optional[bool] = None) -> Union[np.datetime64, int]:
        if raw:
            if self.t == -12:
                return _k(self).j
            return _k(self).i
        if self.is_null:
            return np.datetime64('NaT')
        if self.t == -12:
            epoch_offset = 0 if self.is_inf else self._epoch_offset
            return np.datetime64(_k(self).j + epoch_offset, self._np_type)
        return np.datetime64(_k(self).i + self._epoch_offset, self._np_type)

    def pd(self, *, raw: bool = False, has_nulls: Optional[bool] = None):
        if raw:
            return self.np(raw=True)
        return pd.Timestamp(self.np())


class TimeAtom(TemporalSpanAtom):
    t = -19
    _np_type = 'ms'

    @property
    def is_null(self) -> bool:
        return _k(self).i == -2147483648

    @property
    def is_inf(self) -> bool:
        return abs(_k(self).i) == 2147483647


class SecondAtom(TemporalSpanAtom):
    t = -18
    _np_type = 's'

    @property
    def is_null(self) -> bool:
        return _k(self).i == -2147483648

    @property
    def is_inf(self) -> bool:
        return abs(_k(self).i) == 2147483647


class MinuteAtom(TemporalSpanAtom):
    t = -17
    _np_type = 'm'

    @property
    def is_null(self) -> bool:
        return _k(self).i == -2147483648

    @property
    def is_inf(self) -> bool:
        return abs(_k(self).i) == 2147483647


class TimespanAtom(TemporalSpanAtom):
    t = -16
    _np_type = 'ns'

    @property
    def is_null(self) -> bool:
        return _k(self).j == -9223372036854775808

    @property
    def is_inf(self) -> bool:
        return abs(_k(self).j) == 9223372036854775807


class DatetimeAtom(TemporalFixedAtom):
    t = -15

    def __init__(self, *args, **kwargs):
        warnings.warn('The q datetime type is deprecated', DeprecationWarning)
        super().__init__(*args, **kwargs)

    def np(self, *, raw: bool = False, has_nulls: Optional[bool] = None):
        if raw:
            return _k(self).f
        raise TypeError('The q datetime type is deprecated, and can only be accessed with '
                        'the keyword argument `raw=True`')

    def py(self, *, raw: bool = False, has_nulls: Optional[bool] = None, stdlib: bool = True):
        if raw:
            return _k(self).f
        raise TypeError('The q datetime type is deprecated, and can only be accessed with '
                        'the keyword argument `raw=True`')


class DateAtom(TemporalFixedAtom):
    t = -14
    _np_type = 'D'
    _epoch_offset = DATE_OFFSET

    @property
    def is_null(self) -> bool:
        return _k(self).i == -2147483648

    @property
    def is_inf(self) -> bool:
        return abs(_k(self).i) == 2147483647


class MonthAtom(TemporalFixedAtom):
    t = -13
    _np_type = 'M'
    _epoch_offset = MONTH_OFFSET

    @property
    def is_null(self) -> bool:
        return _k(self).i == -2147483648

    @property
    def is_inf(self) -> bool:
        return abs(_k(self).i) == 2147483647


class TimestampAtom(TemporalFixedAtom):
    t = -12
    _np_type = 'ns'
    _epoch_offset = TIMESTAMP_OFFSET

    @property
    def is_null(self) -> bool:
        return _k(self).j == -9223372036854775808

    @property
    def is_inf(self) -> bool:
        return abs(_k(self).j) == 9223372036854775807

    def py(self, *, raw: bool = False, has_nulls: Optional[bool] = None, stdlib: bool = True):
        # XXX: Since Python datetime objects don't support nanosecond
        #      precision (https://bugs.python.org/issue15443), we have to
        #      convert to datetime64[us] before converting to datetime objects
        if raw:
            return _k(self).j
        return self.np().astype('datetime64[us]').astype(datetime)


class SymbolAtom(Atom):
    t = -11

    @property
    def is_null(self) -> bool:
        return str(self) == ''

    @property
    def is_inf(self) -> bool:
        return False

    @boundscheck(True)
    def __bytes__(self):
        return _k(self).s

    @boundscheck(True)
    def __str__(self):
        return str(_k(self).s, 'utf-8')

    def __add__(self, other):
        return type(self)(str(self) + other)

    def __radd__(self, other):
        return type(self)(other + str(self))

    def __int__(self):
        return int(str(self))

    def __float__(self):
        return float(str(self))

    def __complex__(self):
        return complex(str(self))

    def py(self, *, raw: bool = False, has_nulls: Optional[bool] = None, stdlib: bool = True):
        if raw:
            return bytes(self)
        return str(self)


class CharAtom(Atom):
    t = -10

    @property
    def is_null(self) -> bool:
        return 32 == _k(self).g

    @property
    def is_inf(self) -> bool:
        return False

    def __bytes__(self):
        return self.py()

    def py(self, *, raw: bool = False, has_nulls: Optional[bool] = None, stdlib: bool = True):
        if raw:
            return _k(self).g
        return bytes(chr(_k(self).g), 'utf-8')


class NumericAtom(Atom):
    def __int__(self):
        return int(self.py())

    def __float__(self):
        if self.is_null:
            return float('NaN')
        if self.is_inf:
            return float('inf' if self > 0 else '-inf')
        return float(self.py())

    def __complex__(self):
        return complex(float(self))


class NonIntegralNumericAtom(NumericAtom, Real):
    @property
    def is_null(self) -> bool:
        return math.isnan(self.py())

    @property
    def is_inf(self) -> bool:
        return math.isinf(self.py())

    def __round__(self, ndigits=None):
        return round(self.py(), ndigits)

    def __trunc__(self):
        return math.trunc(self.py())

    def __floor__(self):
        return math.floor(self.py())

    def __ceil__(self):
        return math.ceil(self.py())

    def __pow__(self, other, mod=None):
        if mod is None:
            return q('xexp', self, other)
        raise TypeError(f"pow() 3rd argument not allowed when using '{type(self)}' as the base")

    def __rpow__(self, other, mod=None):
        if mod is None:
            return q('xexp', other, self)
        raise TypeError(f"pow() 3rd argument not allowed when using '{type(self)}' as the exponent")


class FloatAtom(NonIntegralNumericAtom):
    t = -9

    def py(self, *, raw: bool = False, has_nulls: Optional[bool] = None, stdlib: bool = True):
        return _k(self).f

    def np(self, *, raw: bool = False, has_nulls: Optional[bool] = None):
        return np.float64(_k(self).f)


class RealAtom(NonIntegralNumericAtom):
    t = -8

    def py(self, *, raw: bool = False, has_nulls: Optional[bool] = None, stdlib: bool = True):
        return _k(self).e

    def np(self, *, raw: bool = False, has_nulls: Optional[bool] = None):
        return np.float32(_k(self).e)


class IntegralNumericAtom(NumericAtom, Integral):
    def __int__(self):
        if self.is_null:
            raise PyKXException('Cannot convert null integral atom to Python int')
        if self.is_inf:
            raise PyKXException('Cannot convert infinite integral atom to Python int')
        return super().__int__()

    def __index__(self):
        return operator.index(self.py())

    def __round__(self, ndigits=None):
        return round(int(self.py()), ndigits)

    def __trunc__(self):
        return self.py()

    def __floor__(self):
        return self.py()

    def __ceil__(self):
        return self.py()

    def _py_null_or_inf(self, default, raw: bool):
        if not raw and (self.is_null or self.is_inf):
            # By returning the wrapper around the q null/inf when a Python object is requested, we
            # propagate q's behaviour around them - for better and for worse. Notably this ensures
            # symmetric conversions.
            return self
        return default

    def _np_null_or_inf(self, default, raw: bool):
        if not raw:
            if self.is_null:
                raise PyKXException('Numpy does not support null atomic integral values')
            if self.is_inf:
                raise PyKXException('Numpy does not support infinite atomic integral values')
        return default


class LongAtom(IntegralNumericAtom):
    t = -7

    @property
    def is_null(self) -> bool:
        return _k(self).j == -9223372036854775808

    @property
    def is_inf(self) -> bool:
        return abs(_k(self).j) == 9223372036854775807

    def py(self, *, raw: bool = False, has_nulls: Optional[bool] = None, stdlib: bool = True):
        return self._py_null_or_inf(_k(self).j, raw)

    def np(self, *, raw: bool = False, has_nulls: Optional[bool] = None):
        return self._np_null_or_inf(np.int64(_k(self).j), raw)


class IntAtom(IntegralNumericAtom):
    t = -6

    @property
    def is_null(self) -> bool:
        return _k(self).i == -2147483648

    @property
    def is_inf(self) -> bool:
        return abs(_k(self).i) == 2147483647

    def py(self, *, raw: bool = False, has_nulls: Optional[bool] = None, stdlib: bool = True):
        return self._py_null_or_inf(_k(self).i, raw)

    def np(self, *, raw: bool = False, has_nulls: Optional[bool] = None):
        return self._np_null_or_inf(np.int32(_k(self).i), raw)


class ShortAtom(IntegralNumericAtom):
    t = -5

    @property
    def is_null(self) -> bool:
        return _k(self).h == -32768

    @property
    def is_inf(self) -> bool:
        return abs(_k(self).h) == 32767

    def py(self, *, raw: bool = False, has_nulls: Optional[bool] = None, stdlib: bool = True):
        return self._py_null_or_inf(_k(self).h, raw)

    def np(self, *, raw: bool = False, has_nulls: Optional[bool] = None):
        return self._np_null_or_inf(np.int16(_k(self).h), raw)


class ByteAtom(IntegralNumericAtom):
    t = -4

    @property
    def is_null(self) -> bool:
        return False

    @property
    def is_inf(self) -> bool:
        return False

    def py(self, *, raw: bool = False, has_nulls: Optional[bool] = None, stdlib: bool = True):
        return _k(self).g

    def np(self, *, raw: bool = False, has_nulls: Optional[bool] = None):
        return np.uint8(_k(self).g)


class GUIDAtom(Atom):
    t = -2

    @property
    def is_null(self) -> bool:
        return self.py(raw=True) == 0j

    @property
    def is_inf(self) -> bool:
        return False

    def py(self, *, raw: bool = False, has_nulls: Optional[bool] = None, stdlib: bool = True):
        if raw:
            return np.asarray(<np.complex128_t[:1]><np.complex128_t*>_k(self).G0)[0]
        return complex_to_UUID(np.asarray(
            <np.complex128_t[:1]><np.complex128_t*>_k(self).G0).byteswap()[0])


class BooleanAtom(IntegralNumericAtom):
    t = -1

    @property
    def is_null(self) -> bool:
        return False

    @property
    def is_inf(self) -> bool:
        return False

    def py(self, *, raw: bool = False, has_nulls: Optional[bool] = None, stdlib: bool = True):
        if raw:
            return _k(self).g
        return bool(_k(self).g)


class Collection(K):
    @property
    def is_atom(self):
        return False

    def any(self) -> bool:
        return any(x.any() for x in self)

    def all(self) -> bool:
        return all(x.all() for x in self)

    def __bool__(self):
        raise TypeError("The truth value of a 'pykx.Collection' is ambiguous - "
                        "use the '.any()' or '.all()' method, or check the length")

    def __contains__(self, other):
        try:
            first = next(iter(self))
        except StopIteration:
            first = None
        # This special case prevents collections from being equal to atoms in any case. Without
        # this something like `(1,) in q('til 3')` would be `True` when it should be `False`.
        # Other being a str is a special special case, as it is converted to a symbol atom.
        if not isinstance(other, str) and first is not None and \
           not isinstance(first, abc.Collection) and isinstance(other, abc.Collection):
            return False
        return any((x == other) or (x is other) for x in self)

    def __getitem__(self, key):
        if not licensed:
            raise LicenseException('index into K object')
        return q('@', self, _idx_to_k(key, _n(self)))


class Vector(Collection):
    @property
    def has_nulls(self) -> bool:
        return q('{any null x}', self).py()

    @property
    def has_infs(self) -> bool:
        if self.t in {1, 2, 4, 10, 11}:
            return False
        try:
            type_char = ' bg xhijefcspmdznuvts'[self.t]
        except IndexError:
            return False
        return q(f'{{any any -0W 0W{type_char}=\:x}}')(self).py()

    def __len__(self):
        return _k(self).n

    def __iter__(self):
        for i in range(_n(self)):
            yield self._unlicensed_getitem(i)

    def __reversed__(self):
        for i in reversed(range(_n(self))):
            yield self._unlicensed_getitem(i)

    def __array__(self):
        return self.np()

    def __arrow_array__(self, type=None):
        if pa is None:
            raise PyArrowUnavailable # nocov
        return self.pa()

    def py(self, *, raw: bool = False, has_nulls: Optional[bool] = None, stdlib: bool = True):
        return self.np(raw=raw, has_nulls=has_nulls).tolist()

    def pd(self, *, raw: bool = False, has_nulls: Optional[bool] = None):
        return pd.Series(self.np(raw=raw, has_nulls=has_nulls), copy=False)

    def pa(self, *, raw: bool = False, has_nulls: Optional[bool] = None):
        if pa is None:
            raise PyArrowUnavailable # nocov
        return pa.array(self.np(raw=raw, has_nulls=has_nulls))

    def index(self, x, start=None, end=None):
        for i in slice_to_range(slice(start, end), _n(self)):
            if self[i] == x:
                return i
        raise ValueError(f'{x!r} is not in {self!r}')

    def count(self, x):
        return sum((x == v) or (x is v) for v in self)


class List(Vector):
    t = 0

    def _unlicensed_getitem(self, index):
        cdef core.J n = _n(self)
        cdef uintptr_t[:] addrs = <uintptr_t[:n]><uintptr_t*>_k(self).G0
        return _factory(addrs[index], True)

    @property
    def has_nulls(self) -> bool:
        return any(x.is_null if x.is_atom else False for x in self)

    @property
    def has_infs(self) -> bool:
        return any(x.is_inf if x.is_atom else False for x in self)

    def py(self, *, raw: bool = False, has_nulls: Optional[bool] = None, stdlib: bool = True):
        return [_rich_convert(x, stdlib) for x in self]

    def np(self, *, raw: bool = False, has_nulls: Optional[bool] = None):
        """Provides a Numpy representation of the list."""
        cdef uintptr_t[:] addrs, razed_addrs
        cdef Py_ssize_t i
        cdef core.J n = _n(self)
        cdef core.J n0 = 1
        if raw:
            # Should use `np.NPY_UINTP`, but that isn't defined for whatever reason. It's just a
            # typedef for `np.NPY_UINT`, so we'll use that instead.
            return k_vec_to_array(self, np.NPY_UINT).astype(np.uintp, copy=False)

        if n == 0:
            return np.empty((0,), dtype=object)

        addrs = <uintptr_t[:n]><uintptr_t*>_k(self).G0
        n0 = (<core.K><uintptr_t>addrs[0]).n
        ty = (<core.K><uintptr_t>addrs[0]).t

        # Optimization for nested vectors:
        if licensed and n0 >= 1 and ty >= 0 and ty < 21 and ty != 10:
            # q query to check if should raze then return razed vector or null
            # (should raze if all element are of same size and same vector type except CharVector)
            query = \
                b'{if[1=count distinct count each x;'          \
                b'    if[1=count types: distinct type each x;' \
                b'        if[types[0] in "h"$(1+til 20)_9;'    \
                b'            : raze x]]];'                    \
                b'    0N}'
            razed = _factory(<uintptr_t>core.k(0, query, core.r1(_k(self)), NULL), incref=False)
            if isinstance(razed, Vector):
                razed = razed.np()
                razed.shape = (n, -1)
                arr = np.empty(n, dtype=object)
                for i in range(n):
                    arr[i] = razed[i].view()
                return arr
        # Fallback:
        arr = np.empty(n, dtype=object)
        for i in range(n):
            if 10 == (<core.K>(addrs[i])).t:
                arr[i] = _rich_convert(CharVector._from_addr(addrs[i]))
            else:
                arr[i] = _rich_convert(_factory(addrs[i], True), stdlib=False)
        return arr


class NumericVector(Vector):
    def _unlicensed_getitem(self, index):
        # XXX: This is inefficient. It converts the whole vector to Numpy to get a single item.
        return vector_to_atom[type(self)](self.np(has_nulls=False)[index])


class IntegralNumericVector(NumericVector):
    @property
    def has_nulls(self) -> bool:
        return (self._base_null_value == self.np(raw=True)).any()

    @property
    def has_infs(self) -> bool:
        a = self.np(raw=True)
        return (a == self._base_inf_value).any() or (a == -self._base_inf_value).any()

    def py(self, *, raw: bool = False, has_nulls: Optional[bool] = None, stdlib: bool = True):
        if raw:
            return self.np(raw=True, has_nulls=has_nulls).tolist()
        return [x if x.is_null else x.py() for x in self]

    def np(self, *, raw: bool = False, has_nulls: Optional[bool] = None):
        if raw:
            has_nulls = False
        if has_nulls is None or has_nulls:
            nulls = self.np(raw=True) == self._base_null_value
            if has_nulls is None:
                has_nulls = nulls.any()
        arr = k_vec_to_array(self, self._np_type)
        if has_nulls:
            return np.ma.MaskedArray(
                arr,
                mask=nulls,
                copy=False,
                # Set the fill value to the value of the underlying null in q
                fill_value=-2 ** (arr.itemsize * 8 - 1)
            )
        return arr

    def pd(self, *, raw: bool = False, has_nulls: Optional[bool] = None):
        arr = self.np(raw=raw, has_nulls=has_nulls)
        if isinstance(arr, np.ma.MaskedArray):
            arr = pd.arrays.IntegerArray(arr, mask=arr.mask, copy=False)
        return pd.Series(arr, copy=False)


class BooleanVector(IntegralNumericVector):
    t = 1
    _np_type = cnp.NPY_BOOL

    @property
    def has_nulls(self) -> bool:
        return False

    @property
    def has_infs(self) -> bool:
        return False

    # Custom np method for this IntegralNumericVector subclass because it never has nulls
    def np(self, *, raw: bool = False, has_nulls: Optional[bool] = None):
        return super().np(raw=raw, has_nulls=False)


def _raw_guids_to_arrow(guids):
    if pa is None:
        raise PyArrowUnavailable # nocov
    return pa.ExtensionArray.from_storage(arrow_uuid_type,
                                          pa.array(guids.view('S16'), pa.binary(16)))


@pd.api.extensions.register_extension_dtype
class PandasUUIDType(pd.api.extensions.ExtensionDtype):
    na_value = 0+0j
    name = 'pykx.uuid'
    kind = 'c'
    type = np.complex128

    def __from_arrow__(self, array):
        if pa is None:
            raise PyArrowUnavailable # nocov
        if isinstance(array, pa.lib.ChunkedArray):
            if len(array.chunks) > 1:
                raise ValueError('Cannot convert multiple chunks from '
                                 'pyarrow.lib.ChunkedArray to Pandas')
            array = array.chunks[0]
        buf = array.buffers()[1]
        np_array = np.asarray(
            <np.complex128_t[:buf.size // 16]><np.complex128_t*>(
                <uintptr_t>buf.address)) # XXX: this probably isn't safe...
        return PandasUUIDArray(np_array)


class PandasUUIDArray(pd.api.extensions.ExtensionArray):
    dtype = PandasUUIDType()

    def __init__(self, array):
        self.array = array

    def __arrow_array__(self, type=None):
        if pa is None:
            raise PyArrowUnavailable # nocov
        return _raw_guids_to_arrow(np.array(self))

    @classmethod
    def _from_sequence(cls, seq):
        return cls(np.asarray(seq))

    @classmethod
    def _from_factorized(cls, fac):
        raise NotImplementedError

    def __getitem__(self, key):
        return np.array(self)[key]

    def __len__(self):
        return len(np.array(self))

    def __eq__(self, other):
        return np.array(self) == other

    def __array__(self):
        return self.array

    def reshape(self, *args, **kwargs):
        return self

    @property
    def nbytes(self):
        return 16 * len(self)

    def isna(self):
        return np.array(self) == 0

    def take(self, indices, allow_fill=False, fill_value=None):
        return np.array(self).take(indices)

    def copy(self):
        return PandasUUIDArray(np.array(self).copy())

    def _concat_same_type(self):
        raise NotImplementedError


if pa is not None:
    class ArrowUUIDType(pa.ExtensionType):
        def __init__(self):
            if pa is None:
                raise PyArrowUnavailable # nocov
            pa.ExtensionType.__init__(self, pa.binary(16), 'pykx.uuid')

        def __arrow_ext_serialize__(self):
            return b''

        @classmethod
        def __arrow_ext_deserialize__(cls, storage_type, serialized):
            return ArrowUUIDType()

        def to_pandas_dtype(self):
            return PandasUUIDType()

    arrow_uuid_type = ArrowUUIDType()
    pa.register_extension_type(arrow_uuid_type)


class GUIDVector(Vector):
    t = 2

    def _unlicensed_getitem(self, index):
        cdef core.J n = _n(self)
        cdef char[:] vector_as_chars = <char[:n*16]><char*>_k(self).G0
        item_as_bytes = bytes(vector_as_chars[index*16:(index+1)*16])
        return GUIDAtom(UUID(bytes=item_as_bytes))

    @property
    def has_nulls(self) -> bool:
        return (0j == self.np(raw=True)).any()

    @property
    def has_infs(self) -> bool:
        return False

    def py(self, *, raw: bool = False, has_nulls: Optional[bool] = None, stdlib: bool = True):
        return [x.py(raw=raw) for x in self]

    def np(self, *, raw: bool = False, has_nulls: Optional[bool] = None):
        if raw:
            # XXX: NPY_COMPLEX128 is the only 128 bit wide Numpy type, and GUIDs are 128 bits wide
            return k_vec_to_array(self, np.NPY_COMPLEX128)

        cdef np.npy_intp n = _n(self)
        arr = np.PyArray_New(np.ndarray, 1, &n, np.NPY_VOID, NULL, _k(self).G0, 16, 0, None)
        return np.array([UUID(bytes=u.tobytes()) for u in arr], object)

    def pa(self, *, raw: bool = False, has_nulls: Optional[bool] = None):
        if pa is None:
            raise PyArrowUnavailable # nocov
        return _raw_guids_to_arrow(self.np(raw=True))

    def pd(self, *, raw: bool = False, has_nulls: Optional[bool] = None):
        if raw:
            return PandasUUIDArray(self.np(raw=raw))
        else:
            return super().pd()


class ByteVector(IntegralNumericVector):
    t = 4
    _np_type = cnp.NPY_UINT8

    @property
    def has_nulls(self) -> bool:
        return False

    @property
    def has_infs(self) -> bool:
        return False

    # Custom np method for this IntegralNumericVector subclass because it never has nulls
    def np(self, *, raw: bool = False, has_nulls: Optional[bool] = None):
        return super().np(raw=raw, has_nulls=False)


class ShortVector(IntegralNumericVector):
    t = 5
    _np_type = cnp.NPY_INT16
    _base_null_value = -32768
    _base_inf_value = 32767



class IntVector(IntegralNumericVector):
    t = 6
    _np_type = cnp.NPY_INT32
    _base_null_value = -2147483648
    _base_inf_value = 2147483647


class LongVector(IntegralNumericVector):
    """A [`pykx.Vector`][] of longs, i.e. signed 64 bit integers.

    Parameters:
        x: A Python object that will be converted into a `LongVector` via [`pykx.toq`][].

    Attributes:
        t (int): Q type number for long vectors.
    """
    t = 7
    _np_type = cnp.NPY_INT64
    _base_null_value = -9223372036854775808
    _base_inf_value = 9223372036854775807


class NonIntegralNumericVector(NumericVector):
    @property
    def has_nulls(self) -> bool:
        a = self.np()
        return np.isnan(np.dot(a, a)) # `np.dot` can be used as a high-performance NaN check

    @property
    def has_infs(self) -> bool:
        return (self.np() == np.inf).any()

    def __pow__(self, other, mod=None):
        if mod is None:
            return q('xexp', self, other)
        raise TypeError(f"pow() 3rd argument not allowed when using '{type(self)}' as the base")

    def __rpow__(self, other, mod=None):
        if mod is None:
            return q('xexp', other, self)
        raise TypeError(f"pow() 3rd argument not allowed when using '{type(self)}' as the exponent")

    def np(self, *, raw: bool = False, has_nulls: Optional[bool] = None):
        return k_vec_to_array(self, self._np_type)


class RealVector(NonIntegralNumericVector):
    t = 8
    _np_type = cnp.NPY_FLOAT32


class FloatVector(NonIntegralNumericVector):
    t = 9
    _np_type = cnp.NPY_FLOAT64


class CharVector(Vector):
    t = 10

    def _unlicensed_getitem(self, index):
        cdef core.J n = _n(self)
        cdef uint8_t[:] x = <uint8_t[:n]><uint8_t*>_k(self).G0
        cdef core.K kx = core.kc(ord(x[index]))
        return _factory(<uintptr_t>kx, False)

    def __bytes__(self):
        if _n(self):
            return self.np().tobytes()
        return b''

    def __str__(self):
        if _n(self):
            return self.np().tobytes().decode()
        return ''

    def py(self, *, raw: bool = False, has_nulls: Optional[bool] = None, stdlib: bool = True):
        return self.np().tobytes()

    def np(self, *, raw: bool = False, has_nulls: Optional[bool] = None):
        return k_vec_to_array(self, np.NPY_UINT8).view('|S1')


class SymbolVector(Vector):
    t = 11

    def _unlicensed_getitem(self, index):
        cdef core.J n = _n(self)
        cdef uintptr_t[:] addrs = <uintptr_t[:n]><uintptr_t*>_k(self).G0
        return SymbolAtom(str(<char*>addrs[index], 'utf-8'))

    def np(self, *, raw: bool = False, has_nulls: Optional[bool] = None):
        cdef core.J n = _n(self)
        dtype=bytes if raw else str
        if n == 0:
            return np.empty((0,), dtype=dtype)
        cdef uintptr_t[:] addrs = <uintptr_t[:n]><uintptr_t*>_k(self).G0
        if raw:
            arr = [bytes(<char*>addrs[i]) for i in range(n)]
        else:
            arr = [str(<char*>addrs[i], 'utf-8') for i in range(n)]
        return np.asarray(arr, dtype=dtype)


class TemporalVector(Vector):
    @property
    def has_nulls(self) -> bool:
        return (self._base_null_value == self.np(raw=True)).any()

    @property
    def has_infs(self) -> bool:
        a = self.np(raw=True)
        return (a == self._base_inf_value).any() or (a == -self._base_inf_value).any()

    def py(self, *, raw: bool = False, has_nulls: Optional[bool] = None, stdlib: bool = True):
        if raw:
            return self.np(raw=True, has_nulls=has_nulls).tolist()
        return [x if x.is_null else x.py() for x in self]

    def np(self, *, raw: bool = False, has_nulls: Optional[bool] = None):
        base_array = k_vec_to_array(self, self._np_base_type)
        if raw:
            return base_array

        array = base_array.astype(self._np_type, copy=False)

        if raw:
            has_nulls = False
        if has_nulls is None or has_nulls:
            nulls = base_array == self._base_null_value
            if has_nulls is None:
                has_nulls = nulls.any()
        if has_nulls:
            is_fixed = isinstance(self, TemporalFixedVector)
            array[nulls] = np.datetime64('NaT') if is_fixed else np.timedelta64('NaT')

        if isinstance(self, TemporalFixedVector):
            # BUG: This will cause 0Wp to overflow and become 1707-09-22. We could check for 0Wp,
            # and not add the offset to it, but is the cost of doing that on every
            # `TimestampVector.np` call worth avoiding this edge case?
            array += self._epoch_offset
        return array


class TemporalFixedVector(TemporalVector):
    def _unlicensed_getitem(self, index):
        # XXX: This is inefficient. It converts the whole vector to Numpy to get a single item.
        return vector_to_atom[type(self)](
            np.array(
                [self.np(raw=True, has_nulls=False)[index] + self._epoch_offset],
                self._np_type,
            )[0]
        )


class TemporalSpanVector(TemporalVector):
    def _unlicensed_getitem(self, index):
        # XXX: This is inefficient. It converts the whole vector to Numpy to get a single item.
        return vector_to_atom[type(self)](self.np(has_nulls=False)[index])


class TimestampVector(TemporalFixedVector):
    t = 12
    _np_base_type = np.NPY_INT64
    _np_type = cnp.dtype('datetime64[ns]')
    _epoch_offset = TIMESTAMP_OFFSET
    _base_null_value = -9223372036854775808
    _base_inf_value = 9223372036854775807

    def py(self, *, raw: bool = False, has_nulls: Optional[bool] = None, stdlib: bool = True):
        # XXX: Since Python datetime objects don't support nanosecond
        #      precision (https://bugs.python.org/issue15443), we have to
        #      convert to datetime64[us] before converting to datetime objects
        if raw:
            return self.np(raw=True, has_nulls=has_nulls).tolist()
        return self.np().astype('datetime64[us]').astype(datetime).tolist()


class MonthVector(TemporalFixedVector):
    t = 13
    _np_base_type = np.NPY_INT32
    _np_type = cnp.dtype('datetime64[M]')
    _epoch_offset = MONTH_OFFSET
    _base_null_value = -2147483648
    _base_inf_value = 2147483647


class DateVector(TemporalFixedVector):
    t = 14
    _np_base_type = np.NPY_INT32
    _np_type = cnp.dtype('datetime64[D]')
    _epoch_offset = DATE_OFFSET
    _base_null_value = -2147483648
    _base_inf_value = 2147483647


class DatetimeVector(TemporalFixedVector):
    t = 15

    def __init__(self, *args, **kwargs):
        warnings.warn('The q datetime type is deprecated', DeprecationWarning)
        super().__init__(*args, **kwargs)

    def np(self, *, raw: bool = False, has_nulls: Optional[bool] = None):
        if raw:
            return k_vec_to_array(self, np.NPY_FLOAT64)
        raise TypeError('The q datetime type is deprecated, and can only be accessed with the '
                        'keyword argument `raw=True`')

    def py(self, *, raw: bool = False, has_nulls: Optional[bool] = None, stdlib: bool = True):
        if raw:
            return self.np(raw=True, has_nulls=has_nulls).tolist()
        raise TypeError('The q datetime type is deprecated, and can only be accessed with the '
                        'keyword argument `raw=True`')


class TimespanVector(TemporalSpanVector):
    t = 16
    _np_base_type = np.NPY_INT64
    _np_type = cnp.dtype('timedelta64[ns]')
    _base_null_value = -9223372036854775808
    _base_inf_value = 9223372036854775807


class MinuteVector(TemporalSpanVector):
    t = 17
    _np_base_type = np.NPY_INT32
    _np_type = cnp.dtype('timedelta64[m]')
    _base_null_value = -2147483648
    _base_inf_value = 2147483647


class SecondVector(TemporalSpanVector):
    t = 18
    _np_base_type = np.NPY_INT32
    _np_type = cnp.dtype('timedelta64[s]')
    _base_null_value = -2147483648
    _base_inf_value = 2147483647


class TimeVector(TemporalSpanVector):
    t = 19
    _np_base_type = np.NPY_INT32
    _np_type = cnp.dtype('timedelta64[ms]')
    _base_null_value = -2147483648
    _base_inf_value = 2147483647


class EnumVector(Vector):
    t = 20

    def np(self, *, raw: bool = False, has_nulls: Optional[bool] = None):
        if raw:
            return k_vec_to_array(self, np.NPY_INT64)
        return q('value', self).np()


# TODO:
# class Anymap(List):
#     pass


# TODO:
# class NestedSymbolEnum(EnumVector):
#     pass


def _check_k_mapping_key(key: Any, k_key: K, valid_keys):
    if isinstance(k_key, abc.Iterable):
        if not all(any((k == x).all() for x in valid_keys) for k in k_key):
            raise KeyError(key)
    elif not any((k_key == x).all() for x in valid_keys):
        raise KeyError(key)


class Mapping(Collection):
    _rich_conversion_precedence = ('pd', 'np', 'py')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._keys = None
        self._values = None

    def __len__(self):
        return len(self._keys)

    def __iter__(self):
        return self._keys.__iter__()

    @property
    def has_nulls(self) -> bool:
        return any(x.is_null if x.is_atom else False for x in self._values)

    @property
    def has_infs(self) -> bool:
        return any(x.is_inf if x.is_atom else False for x in self._values)

    def any(self) -> bool:
        return any(x.any() for x in self._values)

    def all(self) -> bool:
        return all(x.all() for x in self._values)

    def keys(self):
        return self._keys

    def values(self):
        return self._values

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def items(self):
        yield from ((k, self[k]) for k in self)

    def py(self, *, raw: bool = False, has_nulls: Optional[bool] = None, stdlib: bool = True):
        return dict(zip(
            self._keys.py(raw=raw, has_nulls=has_nulls, stdlib=stdlib),
            self._values.py(raw=raw, has_nulls=has_nulls, stdlib=stdlib),
        ))


class Table(Mapping):
    t = 98

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        cdef core.K self_k = _k(self)
        self._keys = _factory(<uintptr_t>(<core.K*>self_k.k.G0)[0], True)
        self._values = _factory(<uintptr_t>(<core.K*>self_k.k.G0)[1], True)

    def __len__(self):
        if licensed:
            return int(q('#:', self))
        return int(len(self._values._unlicensed_getitem(0)))

    def __getitem__(self, key):
        original_key = key
        if isinstance(key, K):
            return super().__getitem__(key)
        try:
            first = next(iter(key))
        except TypeError:
            first = key
        if isinstance(first, (Integral, slice)):
            n = len(self)
            key = _idx_to_k(key, n) # _idx_to_k raises IndexError as appropriate
        else:
            key = K(key)
            _check_k_mapping_key(original_key, key, self._keys)
        return super().__getitem__(key)

    @property
    def flip(self):
        return Dictionary._from_addr(<uintptr_t>_k(self).k)

    def pd(self, *, raw: bool = False, has_nulls: Optional[bool] = None, raw_guids=False):
        if raw_guids and not raw:
            v = [x.np(raw=isinstance(x, GUIDVector), has_nulls=has_nulls) for x in self._values]
            v = [PandasUUIDArray(x) if x.dtype == complex else x for x in v]
        else:
            v = [x.np(raw=raw, has_nulls=has_nulls) for x in self._values]
        return df_from_arrays(pd.Index(self._keys), v, pd.RangeIndex(len(self)))

    def pa(self, *, raw: bool = False, has_nulls: Optional[bool] = None):
        if pa is None:
            raise PyArrowUnavailable # nocov
        return pa.Table.from_pandas(self.pd(raw=raw, has_nulls=has_nulls, raw_guids=True))

    def np(self, *, raw: bool = False, has_nulls: Optional[bool] = None):
        return self.pd(raw=raw, has_nulls=has_nulls).to_records(index=False)


class SplayedTable(Table):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._splay_dir = SymbolAtom._from_addr(self._values._addr)
        self._values = None
        warnings.warn('Splayed tables are not yet implemented', Warning)

    def __getitem__(self, key):
        raise NotImplementedError

    def any(self):
        raise NotImplementedError

    def all(self):
        raise NotImplementedError

    def items(self):
        raise NotImplementedError

    def values(self):
        raise NotImplementedError

    @property
    def flip(self):
        raise NotImplementedError

    def pd(self, *, raw: bool = False, has_nulls: Optional[bool] = None):
        raise NotImplementedError

    def py(self, *, raw: bool = False, has_nulls: Optional[bool] = None, stdlib: bool = True):
        raise NotImplementedError


class PartitionedTable(SplayedTable):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        warnings.warn('Partitioned tables are not yet implemented', Warning)

    def __getitem__(self, key):
        raise NotImplementedError

    def items(self):
        raise NotImplementedError

    def values(self):
        raise NotImplementedError

    @property
    def flip(self):
        raise NotImplementedError

    def pd(self, *, raw: bool = False, has_nulls: Optional[bool] = None):
        raise NotImplementedError

    def py(self, *, raw: bool = False, has_nulls: Optional[bool] = None, stdlib: bool = True):
        raise NotImplementedError


class Dictionary(Mapping):
    t = 99

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        cdef core.K self_k = _k(self)
        self._keys = _factory(<uintptr_t>(<core.K*>self_k.G0)[0], True)
        self._values = _factory(<uintptr_t>(<core.K*>self_k.G0)[1], True)

    def __getitem__(self, key):
        if not isinstance(key, K):
            original_key = key
            key = K(key)
            _check_k_mapping_key(original_key, key, self._keys)
        return super().__getitem__(key)


class KeyedTable(Dictionary):
    def any(self) -> bool:
        return any(x.any() for x in self._values._values)

    def all(self) -> bool:
        return all(x.all() for x in self._values._values)

    def __getitem__(self, key):
        if not isinstance(key, K):
            original_key = key
            key = K(key)
            # Keyed tables with a single key column are handled differently by q
            valid_keys = (
                *(self._keys._values._unlicensed_getitem(0) if len(self._keys._keys) == 1 else ()),
                *zip(*self._keys._values),
            )
            if not any((key == x).all() for x in valid_keys):
                raise KeyError(original_key)
        # XXX: KeyedTable is a subclass of `Dictionary` and `Mapping`, but it has different
        # indexing semantics, so we skip straight to `Collection.__getitem__`.
        return Collection.__getitem__(self, key)

    def __iter__(self):
        yield from zip(*self._keys._values)

    def keys(self):
        return list(self)

    def np(self, *, raw: bool = False, has_nulls: Optional[bool] = None):
        if licensed:
            return q('0!', self).np(raw=raw, has_nulls=has_nulls)
        raise LicenseException('convert a keyed table to a Numpy representation')

    def pd(self, *, raw: bool = False, has_nulls: Optional[bool] = None):
        return df_from_arrays(
            pd.Index(self._values._keys),
            [self._values._values._unlicensed_getitem(i).np(has_nulls=has_nulls) for i in range(len(self._values._keys))],
            pd.MultiIndex.from_arrays(
                [self._keys._values._unlicensed_getitem(i).np(has_nulls=has_nulls) for i in range(len(self._keys._values))],
                names=[self._keys._keys._unlicensed_getitem(i) for i in range(len(self._keys._values))]
            )
        ) # noqa: E501

    def pa(self):
        if pa is None:
            raise PyArrowUnavailable # nocov
        raise NotImplementedError

    def py(self, *, raw: bool = False, has_nulls: Optional[bool] = None, stdlib: bool = True):
        value_keys = self._values._keys.py()
        return dict(zip(
            zip(*self._keys._values.py(raw=raw, has_nulls=has_nulls, stdlib=stdlib)),
            (dict(zip(value_keys, x)) for x in zip(*self._values._values.py(raw=raw, has_nulls=has_nulls, stdlib=stdlib))),
        )) # noqa: E501


class Function(Atom):
    def __bool__(self):
        return True

    @property
    def is_null(self) -> bool:
        return False

    @cached_property
    def params(self):
        return ()

    @cached_property
    def args(self):
        return ()

    @cached_property
    def func(self):
        return self

    def __call__(self, *args, **kwargs):
        if not licensed:
            raise LicenseException('call a q function in a Python process')
        args = {i: K(x) for i, x in enumerate(args)}
        if kwargs:
            for param_name, param_value in kwargs.items():
                try:
                    i = self.params.index(param_name)
                except ValueError:
                    raise TypeError(
                        f"Function {self!r} got an unexpected keyword argument '{param_name}'"
                    ) from None
                else:
                    if i in args:
                        raise TypeError(
                            f"Function {self!r} got multiple values for parameter '{param_name}'"
                        ) from None
                args[i] = K(param_value)
        if args: # Avoid calling `max` on an empty sequence
            args = {**{x: ... for x in range(max(args))}, **args}
        args = K([x for _, x in sorted(args.items())]) if len(args) else q('enlist(::)')
        return _factory(<uintptr_t>core.ee(core.dot(_k(self), _k(args))), False)

    @cached_property
    def each(self):
        return q("{x'}", self)

    @cached_property
    def over(self):
        return q('{x/}', self)

    @cached_property
    def scan(self):
        return q('{x\\}', self)

    @cached_property
    def each_prior(self):
        return q("{x':}", self)

    prior = each_prior

    @cached_property
    def each_right(self):
        return q('{x/:}', self)

    sv = each_right

    @cached_property
    def each_left(self):
        return q('{x\\:}', self)

    vs = each_left


class Lambda(Function):
    t = 100

    @cached_property
    def params(self):
        # Strip "PyKXParam" from all param names if it is a prefix for all
        return tuple(q('k){x:.:x;$[min (x:x[1]) like "PyKXParam*"; `$9_\'$x; x]}', self))


class UnaryPrimative(Function):
    t = 101

    def __call__(self, *args, **kwargs):
        if kwargs:
            raise TypeError('Cannot use kwargs on a unary primative')
        return super().__call__(*args, **kwargs)


class Identity(UnaryPrimative):
    def __repr__(self):
        # for clarity, represents Identity objects with ::
        return f"pykx.{type(self).__name__}(pykx.q('::'))"

    def py(self, *, raw: bool = False, has_nulls: Optional[bool] = None, stdlib: bool = True):
        return None

    def __bool__(self):
        return False

    def __str__(self):
        return '::'

    @property
    def is_null(self) -> bool:
        return True


class ProjectionNull(Identity):
    def py(self, *, raw: bool = False, has_nulls: Optional[bool] = None, stdlib: bool = True):
        return ...

    def __lt__(self, other):
        return q('0b')

    def __gt__(self, other):
        return q('0b')

    def __eq__(self, other):
        try:
            return q('104h ~ type {1b}@', other)
        except TypeError:
            return q('0b')

    def __ne__(self, other):
        return K(not self.__eq__(other))

    __ge__ = __eq__
    __le__ = __eq__


class Operator(Function):
    t = 102

    def __call__(self, *args, **kwargs):
        if kwargs:
            raise TypeError('Cannot use kwargs on an operator')
        return super().__call__(*args, **kwargs)


class Iterator(Function):
    t = 103

    def __call__(self, *args, **kwargs):
        if kwargs:
            raise TypeError('Cannot use kwargs on an iterator')
        return super().__call__(*args, **kwargs)


class Projection(Function):
    t = 104

    @cached_property
    def params(self):
        """The param names from the base function that have not been set."""
        if not self.func.params:
            return self.func.params
        # Find the projection nulls
        xi = q('{(104h~type {1b}@) each value x}', self)[1:]
        return tuple(x for i, x in enumerate(self.func.params) if i >= len(xi) or xi[i])

    @cached_property
    def args(self):
        return tuple(q('value', self)[1:])

    @cached_property
    def func(self):
        return q('{value[x]0}', self)

    def __call__(self, *args, **kwargs):
        if kwargs and not self.params:
            raise TypeError(
                'This projection does not support kwargs, as it is a '
                f'projection of a {type(self.func).__name__}')
        return super().__call__(*args, **kwargs)


class Composition(Function):
    t = 105

    @cached_property
    def params(self):
        return self.func.params

    @cached_property
    def args(self):
        return tuple(q("k){$[(@)~*v:. x;*|v;x]}'{,/({::;$[105h~@y;x y;y]}.z.s)@'. x }@", self))

    @cached_property
    def func(self):
        return self.args[len(self.args) - 1]

    def __call__(self, *args, **kwargs):
        if kwargs and not self.params:
            raise TypeError(
                'This composition does not support kwargs, as the first '
                f'function applied is a {type(self.func).__name__}')
        return super().__call__(*args, **kwargs)


class AppliedIterator(Function):
    @cached_property
    def params(self):
        return self.func.params

    @cached_property
    def args(self):
        return self.func.args

    @cached_property
    def func(self):
        return q('value', self)

    def __call__(self, *args, **kwargs):
        if kwargs and not self.params:
            raise TypeError(
                'This applied iterator does not support kwargs, as the base'
                f'function is a {type(self.func).__name__}'
            )
        return super().__call__(*args, **kwargs)


class Each(AppliedIterator):
    t = 106


class Over(AppliedIterator):
    t = 107


class Scan(AppliedIterator):
    t = 108


class EachPrior(AppliedIterator):
    t = 109


class EachRight(AppliedIterator):
    t = 110


class EachLeft(AppliedIterator):
    t = 111


class Foreign(Function):
    t = 112


vector_to_atom = {
    BooleanVector:   BooleanAtom,
    GUIDVector:      GUIDAtom,
    ByteVector:      ByteAtom,
    ShortVector:     ShortAtom,
    IntVector:       IntAtom,
    LongVector:      LongAtom,
    RealVector:      RealAtom,
    FloatVector:     FloatAtom,
    CharVector:      CharAtom,
    SymbolVector:    SymbolAtom,
    TimestampVector: TimestampAtom,
    MonthVector:     MonthAtom,
    DateVector:      DateAtom,
    DatetimeVector:  DatetimeAtom,
    TimespanVector:  TimespanAtom,
    MinuteVector:    MinuteAtom,
    SecondVector:    SecondAtom,
    TimeVector:      TimeAtom,
    EnumVector:      EnumAtom,
}


atom_to_vector = {v: k for k, v in vector_to_atom.items()}


# tags for types that need special handling
_k_table_type = object()
_k_dictionary_type = object()
_k_unary_primative = object()


type_number_to_pykx_k_type = {
    -128: QError,
    -20: EnumAtom,
    -19: TimeAtom,
    -18: SecondAtom,
    -17: MinuteAtom,
    -16: TimespanAtom,
    -15: DatetimeAtom,
    -14: DateAtom,
    -13: MonthAtom,
    -12: TimestampAtom,
    -11: SymbolAtom,
    -10: CharAtom,
     -9: FloatAtom,                                                     # noqa
     -8: RealAtom,                                                      # noqa
     -7: LongAtom,                                                      # noqa
     -6: IntAtom,                                                       # noqa
     -5: ShortAtom,                                                     # noqa
     -4: ByteAtom,                                                      # noqa
     -2: GUIDAtom,                                                      # noqa
     -1: BooleanAtom,                                                   # noqa
      0: List,                                                          # noqa
      1: BooleanVector,                                                 # noqa
      2: GUIDVector,                                                    # noqa
      4: ByteVector,                                                    # noqa
      5: ShortVector,                                                   # noqa
      6: IntVector,                                                     # noqa
      7: LongVector,                                                    # noqa
      8: RealVector,                                                    # noqa
      9: FloatVector,                                                   # noqa
     10: CharVector,                                                    # noqa
     11: SymbolVector,                                                  # noqa
     12: TimestampVector,                                               # noqa
     13: MonthVector,                                                   # noqa
     14: DateVector,                                                    # noqa
     15: DatetimeVector,                                                # noqa
     16: TimespanVector,                                                # noqa
     17: MinuteVector,                                                  # noqa
     18: SecondVector,                                                  # noqa
     19: TimeVector,                                                    # noqa
     20: EnumVector,                                                    # noqa
     98: _k_table_type,                                                 # noqa
     99: _k_dictionary_type,                                            # noqa
    100: Lambda,
    101: _k_unary_primative,
    102: Operator,
    103: Iterator,
    104: Projection,
    105: Composition,
    106: Each,
    107: Over,
    108: Scan,
    109: EachPrior,
    110: EachRight,
    111: EachLeft,
    112: Foreign
}


cdef object q_table_type(core.K val):
    cdef core.K x
    x = core.k(0, <char* const>'.Q.qp', core.r1(val), NULL)
    if x.t == -1 and x.g == 1:      # x == 1b
        wrapper = PartitionedTable
    elif x.t == -1 and x.g == 0:    # x == 0b
        wrapper = SplayedTable
    else:                           # x == 0
        wrapper = Table
    core.r0(x)
    return wrapper


cdef bint is_pnull(core.K val):
    cdef core.K x
    x = core.k(0, <char* const>'104h ~ type {1b}@', core.r1(val), NULL)
    cdef bint r = x.g
    core.r0(x)
    return r


cdef inline object select_wrapper(core.K k):
    cdef signed char ktype = k.t
    cdef signed char key_ktype
    wrapper = type_number_to_pykx_k_type.get(ktype, K)
    if wrapper is _k_unary_primative:
        if k.g:
            wrapper = UnaryPrimative
        elif licensed and is_pnull(k):
            wrapper = ProjectionNull
        else:
            wrapper = Identity
    elif wrapper is _k_table_type:
        wrapper = q_table_type(k) if licensed else Table
    elif wrapper is _k_dictionary_type:
        # XXX: it's possible to have a dictionary that is not a keyed table, but still uses tables
        # as its keys
        key_ktype = (<core.K*>k.G0)[0].t
        wrapper = KeyedTable if key_ktype == 98 else Dictionary
    return wrapper


cdef inline _factory(uintptr_t addr, bint incref):
    wrapper = select_wrapper(<core.K>addr)
    if wrapper is QError:
        q_exception = wrapper(str((<core.K>addr).s, 'utf-8'))
        # `pop` the exception object out to prevent it from being handled multiple times.
        _current_exception_in_thread = _current_exception.pop(threading.get_ident(), None)
        if _current_exception_in_thread is None:
            # `raise ex from None` is different from `raise ex`, so we can't write this line as
            # `raise q_exception from _current_exception_in_thread`.
            raise q_exception
        else:
            raise q_exception from _current_exception_in_thread
    return wrapper._from_addr(addr, incref)


def factory(uintptr_t addr, bint incref=True):
    return _factory(addr, incref)


__all__ = sorted([
    *(() if pa is None else ('ArrowUUIDType',)),
    'PandasUUIDArray',
    'PandasUUIDType',
    'atom_to_vector',
    'vector_to_atom',
    *(x.__name__ for x in subclasses(K)),
])


def __dir__():
    return __all__
